
//**************************************************************************
//    Connection.h
//
//    The Connection Class - Used to hold all information about a 
//    connection.
//**************************************************************************

#ifndef _Connection_H_
#define _Connection_H_

#include "User.h"
#include "ParseStack.h"


//**************************************************************************

                      // The number of logical users to 
                      // allow on a single physical connection
                      // (Actually, one less "real" user
                      // that this number is allowed to
                      // give room for a default user "nobody"
#define  MAX_USRS_PER_CONN         17

                      // A special user that always exists if there is
                      // a connection.  This is needed so that an exEnv
                      // exists so that you can enter commands (like login()).
                      // This value should not be changed, it is here only
                      // to clarify the code.
#define  NOBODY    0

                      // Max length of string in cwd variable
#define  MAX_PATH_LEN  256
//**************************************************************************

class corConnection {

friend class corConnectArray;

protected:

  int         num_users;    // A count of the number of logical users

  int         nxt_user;     // A spot "likely" to be free in the userArray

  int         cur_user;     // The index of the current user

public:  

  int         conn_id;      // The index of this connection

                            // An array of users
  corUser       *userArray[MAX_USRS_PER_CONN];

  corParseStack  pStack;       // One ParseStack needed per socket connection

  CoralSock  *conn_sock;    // The socket for this connection

  char        cwd[MAX_PATH_LEN];  // Where this connection is at.

  corConnection(CoralSock *serv_sock, int put_on_stack = 1);
 ~corConnection();

    // Switch to this connection
  int useConnection();

    // Switch to this user
  int useUser(int user_id);

    // Generates a new user
  int  NewUser(char *name, char *passwd);

    // Deletes a user from the array of users
  void DeleteUser(int user_id);

    // Find the user with this name, returning the first match
  int FindUser(char *name);

    // Return a pointer to the current user
  inline corUser *currentUser() {
      return(userArray[cur_user]);
  }

    // Returns a pointer to the user specified (by user_id) 
  inline corUser *indxUser(int user_id) {
      return(userArray[user_id]);
  }

    // List all users for this connection, return the number found
  int ListWho(FILE *fp);
};

//**************************************************************************

#endif  // _Connection_H_
