/*****************************************************************************/
//	FuncArg.h
/*****************************************************************************/

#ifndef _FuncArg_H_
#define _FuncArg_H_

#include "Config.h"
#include "Arg.h"
#include "ArgList.h"
#include "SymbolArg.h"

#include "ObjectCount.h"	// REMOVE WHEN FIXED

/*****************************************************************************/
// Arguments that are functions. eg : f(X,Y) in pred(f(X,Y), Z).
// WARNING : FuncArg is variable size. Use FuncArg::New as constructor
/*****************************************************************************/

class corFuncArg : public corArg {
protected:
  inline corFuncArg(Name op, int argcount) : _functor(op), args(argcount)
   {
		_hash = UnknownHashValue;
		//args.set_count(argcount);
		_kindof = COR_FUNCTOR;
	}
	virtual ~corFuncArg() {};
	HashVal _hash;		
	// The field _hash of a FuncArg must be set to UnknownHashValue if the
	// 	hash value has not been computed.	It must be set to 
	//	VarHashValue if FuncArg is found to have variables in it.
	//		It must be set to other values if FuncArg is found to be 
	//	variable free.	Thus looking at the _hash value is a 
	// 	quick sufficient test for the FuncArg being constant.
	//
public:
	Name _functor;				 /* functor info if kind==COR_FUNCTOR */
	corArgList	args;				 /* args of functor */

	static corFuncArg * New(Name op, int arg_count);

	virtual int isConstant();
	virtual int equals(corArg *arg2);

	int count() const { return args.count(); }
	inline Name functor() const { return _functor; }
	
	inline int arity() { return args.count(); }

	virtual void print(corBindEnv*, FILE *file) const;
	virtual void print(corBindEnv*, FILE *file, char *fmt_str) const;
	virtual void sprint(char *str, int *pos, corBindEnv *context = NULL) const;

	virtual corVarLink *var_list(corVarLink* rest);
	virtual HashVal hash(corBindEnv *context = 0);
	virtual int scan_vars(void *data, VarFunc *func);
	virtual corCArg simplify(corBindEnv *context, corTermLink *&renamed_vars,
		corBindEnv *const_env, corBindEnv *rename_only_env);
};

/*****************************************************************************/

#ifdef USE_HASHCONSING
void corcreateFunctorTable();
#endif USE_HASHCONSING

/*****************************************************************************/

#endif _FuncArg_H_

