// *************************************************************************
//    GenericIndex.h
// *************************************************************************

#ifndef _GenericIndex_H_
#define _GenericIndex_H_

#include "ArgList.h"
#include "StorageRelation.h"
#include "BindEnv.h"
#include "BitVector.h"
#include "Tuple.h"
#include "TupleIterator.h"

class corGenericIndexSet; 

// *************************************************************************

class corGenericIndex {
    friend class corGenericIndexSet;
protected:
    corStorageRelation* relation;        // relation that is being indexed

    // for argument form indexing only next field is relevant
    corBitVector *argset; 

    // for pattern form indexing only next three fields are relevant
    corArgList *pattern;                // pattern to be used for indexing
    int num_vars_pat;                // number of variables in pattern
    corArgList *var_names;                // subset of variables in pattern
                                    // that are used for indexing
    
public: 
    corGenericIndex(corStorageRelation* rel, corBitVector* aset) {
        relation = rel;
            // LEAK
        argset = new corBitVector(*aset);
        pattern = NULL;
        num_vars_pat = 0;
        var_names = NULL;
     }

    corGenericIndex(corStorageRelation* rel,corArgList* pat,int n_var,corArgList* v_nm) {
        relation = rel; 
        pattern = pat;
        num_vars_pat = n_var;
        var_names = v_nm;
        argset = NULL;
     }
    /*
     *    this assumes that ArgList *pat and ArgList *v_nm
     *    have been created with the same variable numbers assigned 
     *    to the same variable names, and that the variable numbers are
     *    sequentially assigned from 0, with no number being skipped 
     *    in between.
     */
    // NEW_CHNG:wX
    // required for this abstract class
    virtual ~corGenericIndex() {};

    inline int add_index_to_reln(corStorageRelation *reln) {
        if (argset == NULL) { // add pattern form index
            return reln->add_index(pattern, num_vars_pat, var_names);
        }
        else { // add argument form index
            return reln->add_index(argset);
        }
     }

    virtual void initialize_iterator( corTupleIterator& ) = 0;

    inline int is_argument_form()        { return (argset != NULL); }

    virtual corTuple *get_next_tuple(corTupleIterator& ) = 0;
    virtual corBindEnv *get_next( corTupleIterator& ) = 0;
    int matches(corGenericIndex *ind);
    virtual int index_type() = 0;
    virtual int is_subsumed(corTuple *tuple)=0;
    
    virtual void insert_tuple( corTuple *tuple )    = 0;
    virtual void delete_all_tuples();
    virtual void free_all_entries() = 0;
    virtual void move_all_entries(corGenericIndex *ind) = 0;
    virtual void reset_index() = 0;
        // Empty the index, but do NOT delete tuples (these are shared)
        // Retain the index, for reuse, and reset its size
};


typedef corGenericIndex *GenericIndexPtr;

// *************************************************************************

enum IndexType { COR_HASH_INDEX };

// *************************************************************************

char *corsimplify_index(corArgList *pat, int n_var, corArgList *v_nm);

// *************************************************************************

#endif _GenericIndex_H_
