// *************************************************************************
//    GenericIndexSet.h
// *************************************************************************

#ifndef _GenericIndexSet_H_
#define _GenericIndexSet_H_

#include "Config.h"

#include "Context.h"
#include "BitVector.h"
#include "TupleIterator.h"
//#include "StorageRelation.h"
#include "GenericIndex.h"

// *************************************************************************

class corStorageRelation;

// *************************************************************************

class corGenericIndexSet {
protected:
    corRelation *rel;
    int num_in_array;
    int max_in_array;
    GenericIndexPtr *Indices;
public:
    corGenericIndexSet(corStorageRelation *reln) {
        rel = reln;
        num_in_array = 0;  // no indices available at initialization 
        max_in_array = COR_MAX_INDEXES;
        Indices = new GenericIndexPtr [COR_MAX_INDEXES];
    }

   ~corGenericIndexSet() {
        // Magic doesn't work with this in.
        //delete [] Indices;
    }

    int add_all_indices(corStorageRelation *reln);
    // transfers its indices to reln.
    virtual int add_all_tuples(corStorageRelation *, int, int, int ) { return 0;}
    virtual int add_all_goals(corStorageRelation*, corContext*, corStorageRelation*) {
        return 0;
    }

    inline int num_indices() { return num_in_array; }
    virtual corGenericIndex *find_index( corTupleIterator& ) = 0;
        // finds the index that best "matches" the given argument set in
        // the iterator
    virtual corGenericIndex *find_index( corTuple*, int *tryindexp= NULL ) = 0;
    virtual corGenericIndex *find_index( corArgList&, corBindEnv*,
        int *tryindexp=NULL) = 0;

    inline GenericIndexPtr& operator[](int i) {return Indices[i];}

    virtual int insert_index( corGenericIndex* Index ) = 0;
        // adds an index to the set of known indices

    virtual void fill_index(corGenericIndex* Index) = 0;
        // fills in the index using default index in posn. 0

    virtual void print_adornments(FILE *outf);
};

// *************************************************************************

#define COR_I_FAIL 0        // failure for indices
#define COR_I_SUCCESS 1        // success for indices
#define COR_I_PARTIAL_SUCCESS  2

// *************************************************************************

#endif _GenericIndexSet_H_

