// *************************************************************************
//    HashSimpleRelation.h
// *************************************************************************

#ifndef _HashSimpleRelation_H_
#define _HashSimpleRelation_H_

#include "HeapNode.h"
#include "StorageRelation.h"
#include "RelationList.h"
#include "HashIndex.h"
#include "HashIndexSet.h"
#include "NumArg.h"

// *************************************************************************
/*
    Tuples which are newer than the oldest RMark are in the "chain" list
    inherited from LinkedRelation.  The others are accessed through the
    Hash Indexes.
*/
// *************************************************************************

class corHashSimpleRelation : public corStorageRelation {
public:
    void initialize_iterator(corTupleIterator&);

    corHeapNode *heap;                    // For prioritize annotation

    corHashSimpleRelation( int arity, int delta = -1, int buckets = -1);

    virtual void freeMark(corRMark *, int hide_em=0);
    virtual void free_allMarks();

    //    Indexing related methods
    corHashIndexSet index_set;
    virtual corGenericIndexSet *get_index_set() { return &index_set ;}
    virtual corGenericIndex *create_index_object(corBitVector *bv) {
        return new corHashIndex(this, bv, -1);
    }
    virtual corGenericIndex *create_index_object(corArgList *pat, int n_var,
        corArgList *v_nm)
    {
        return new corHashIndex(this, pat, n_var, v_nm);
    }

    //    Some priority related methods
    virtual int add_hidden_relation(corPrioritizeInfo *info);
    virtual int insert_in_hidden_relation(corTuple *tuple);

    // Pop a tuple from the hidden relation and add it
    //    to visible part.
    virtual int pop_from_hidden_relation();

    // Pop all tuples from the hidden relation, and
    // add them to visible part.
    virtual int pop_all_from_hidden_relation();

    virtual int insert_tuple(corTuple *tuple);
    virtual int insert_raw( corTuple *tuple );

    int insert_in_heap(corTuple *tuple);
    virtual corTuple *get_next_tuple( corTupleIterator& it) ;
    virtual void release(corTupleIterator& it) ;
    virtual corBindEnv *get_next( corTupleIterator& );
    virtual int is_subsumed(corTuple *tuple, corRMark *startmark = NULL,
        corRMark *endmark = NULL);

    virtual void free_all();
    virtual void free_indices();
    virtual void free_delta_relations();
    int move_all_entries(corStorageRelation *rel);     
            // Move all tuples/index entries into rel.  
            // Assumes that rel has the same set of indices, 
            // and all corresponding indices and relations are
            // of exactly the same type.

    virtual void empty_relation(int deleteTuples=0);
    virtual void print_name(FILE *outf);

    virtual ~corHashSimpleRelation();
};

// *************************************************************************

#endif _HashSimpleRelation_H_

