/*****************************************************************************/
//    PersistentRelation.h
/*****************************************************************************/

#ifndef _PersistentRelation_H_
#define _PersistentRelation_H_

#include "Globals.h"
#include "CoralError.h"
#include "SMUtil.h"
#include "sm_client.h"
#include "Config.h"
#include "Schema.h"
#include "IndexStruct.h"
#include "BPlusTreeIndex.h"
#include "FileDescStruct.h"
#include "BufferGroup.h"
#include "Transaction.h"
#include "Catalog.h"
#include "StorageManager.h"

#include "RelationMisc.h"
#include "Relation.h"
#include "RelationList.h"

#include "PersistentIndex.h"
#include "PersistentIndexSet.h"


#define Scanning_Persistent 1

/*****************************************************************************/

class corPersistentIndexSet;

class corPersistentRel : public corStorageRelation {
    friend class corPersistentIndex;
    friend class corPersistentIndexSet;
private:
    OID oid;                    // OID of FileDescStruct of the relation
    corFileDescStruct fd;            // File descriptor

    LOADDESC* lddesc;            // bulk load descriptor

    corPersistentIndexSet index_set;

    //    Need representation of field types

    void insertObject(corArgList &arglist);
    corTuple *fileScanGetNext(corTupleIterator &, int delete_flag = 0);
    corTuple *indexScanGetNext(corTupleIterator&, int, int delete_flag = 0);

    // Returns position where an index entry can be made
    //int canAddIndex() { return fd.canAddIndex() ;} ;
    corTuple *checkTuple(USERDESC *, corTupleIterator&);

public:
    corPersistentRel(Name name, int arity, int& error);
    corPersistentRel(Name name, int arity, int& error, OID*);
    corPersistentRel(Name name, int arity, int& error, corFileDescStruct *);

    //virtual RelationKind find_r_kind() { return COR_R_EXODUS; }
    //NEW_CHNG:wX

    virtual void release(corTupleIterator& iterator);
    virtual void print_name(FILE *file);
    virtual void print_facts(FILE *file, corArgList *arglist = NULL) const;

    virtual void freeMark(corRMark *, int hide_em=0);
    virtual void free_allMarks();

    virtual corBindEnv* get_next(corTupleIterator&);
    virtual corTuple* get_next_tuple(corTupleIterator&) ;
    virtual int insert_raw(corTuple *tuple);

    virtual int insert_tuple(corTuple *tuple);
    // maybe this can use the ::insert_new defined in generic-rel ????
    // virtual int insert_new(ArgList& args, BindEnv* env,
    //    BindEnv *dont_rename_env = NULL, Tuple *parent_goal = NULL,
    //    int dont_rename = 0); 

    virtual corGenericIndexSet *get_index_set() { return &index_set; }
    virtual corGenericIndex *create_index_object(corBitVector *bv) {
        return new corPersistentIndex(this, bv);
    }
    virtual corGenericIndex *create_index_object(corArgList *, int, corArgList *) {
        return NULL;
    }

    virtual int tuple_delete(corTuple *tuple);
    // Empty tuples out of the relation
    virtual void empty_relation(int delete_tuples = 0); 
    void initialize_iterator( corTupleIterator& );
    virtual int is_subsumed(corTuple *tuple, corRMark *startmark= NULL,
        corRMark *endmark = NULL);

    int start_bulk_load();
    int end_bulk_load();
    void flush_to_disk();
};

/*****************************************************************************/

#endif _PersistentRelation_H_


