# line 1 "RdbIngresDatabase.scc"	

#include <strings.h>
#include "RdbIngresDatabase.h"
#include "Log.h"

# include "/ingres/r6/ingres/files/eqsqlca.h"
    extern IISQLCA sqlca;   /* SQL Communications Area */
#include "IngresProtos.h"


int RdbIngresDatabase::nextSessionId = 1; // if this rolls back around,
                                          // could be big trouble!!
                                          // Note: id of 0 is not valid
                                          // for multiple connections,
					  // it means single connect session.


RdbError::ErrorCode RdbIngresDatabase::VConnect()
{
  char *db = dbName;
  int session = sessionId;

    if( connected )
      return NO_ERROR;

    LOG("Connecting to "<<db<<" session: "<<session<<'\n');
# line 28 "RdbIngresDatabase.scc"	/* connect */
  {
    IIsqInit(&sqlca);
    IILQsidSessID((int)session);
    IIsqConnect((int)0,db,(char *)0, (char *)0, (char *)0, (char *)0, 
    (char *)0, (char *)0, (char *)0, (char *)0, (char *)0, (char *)0, 
    (char *)0, (char *)0, (char *)0);
  }
# line 29 "RdbIngresDatabase.scc"	/* host code */
    if( sqlca.sqlcode )
      return Error(SQL_ERROR);
    connected = TRUE;
    return NO_ERROR;
}


RdbError::ErrorCode RdbIngresDatabase::VDisconnect()
{
    if( !connected )
      return NO_ERROR;
    InvalidateCompilableDMLs();
    InvalidateCursors();
# line 42 "RdbIngresDatabase.scc"	/* disconnect */
  {
    IIsqInit(&sqlca);
    IIsqDisconnect();
  }
# line 43 "RdbIngresDatabase.scc"	/* host code */
    connected = FALSE;		// even if error, set it to disconnected
    if( sqlca.sqlcode )
      return Error(SQL_ERROR);
    return NO_ERROR;
}


String RdbIngresDatabase::VGetErrorMessage() 
{
  char errmsg[512];

    // could use sqlca.sqlerrm, but that is not always the full message
    LOG(sqlca.sqlerrm.sqlerrmc << '\n');
# line 58 "RdbIngresDatabase.scc"	/* inquire_ingres */
  {
    IILQisInqSqlio((short *)0,(int)1,(int)32,(int)511,errmsg,(int)63);
  }
# line 59 "RdbIngresDatabase.scc"	/* host code */
    LOG(sqlca.sqlerrm.sqlerrmc << '\n' << errmsg << '\n');
    if( sqlca.sqlcode ) {
	Error(SQL_ERROR);
	return "Could not get error message!";
    }
    return errmsg;
}


RdbError::ErrorCode RdbIngresDatabase::VImmediateSql(const String& sql)
{
  char *stmt = sql;

    if( !strcasecmp( stmt, "commit" ) )
      return Commit();
    if( !strcasecmp( stmt, "rollback" ) )
      return Rollback();

# line 79 "RdbIngresDatabase.scc"	/* execute */
  {
    IIsqInit(&sqlca);
    IIsqExImmed(stmt);
    IIsyncup("RdbIngresDatabase.scc",(int)79);
  }
# line 80 "RdbIngresDatabase.scc"	/* host code */
    if( sqlca.sqlcode )
      return Error(SQL_ERROR);
    return NO_ERROR;
}


RdbError::ErrorCode RdbIngresDatabase::VCommit()
{
# line 88 "RdbIngresDatabase.scc"	/* commit */
  {
    IIsqInit(&sqlca);
    IIxact((int)3);
  }
# line 89 "RdbIngresDatabase.scc"	/* host code */
    InvalidateCompilableDMLs();
    InvalidateCursors();
    if( sqlca.sqlcode )
      return Error(SQL_ERROR);
    return NO_ERROR;
}


RdbError::ErrorCode RdbIngresDatabase::VRollback()
{
# line 99 "RdbIngresDatabase.scc"	/* rollback */
  {
    IIsqInit(&sqlca);
    IIxact((int)2);
  }
# line 100 "RdbIngresDatabase.scc"	/* host code */
    InvalidateCompilableDMLs();
    InvalidateCursors();
    if( sqlca.sqlcode )
      return Error(SQL_ERROR);
    return NO_ERROR;
}


RdbError::ErrorCode RdbIngresDatabase::VSetFocus()
{
  int session = sessionId;
    LOG("Switching to session "<<session<<" ("<<dbName<<")\n");
# line 114 "RdbIngresDatabase.scc"	/* set_sql */
  {
    IILQssSetSqlio((int)11,(short *)0,(int)1,(int)30,(int)4,&session);
  }
# line 115 "RdbIngresDatabase.scc"	/* host code */
    if( sqlca.sqlcode )
      return Error(SQL_ERROR);
    return NO_ERROR;
}


RdbError::ErrorCode RdbIngresDatabase::VRemoveFocus()
{
    return NO_ERROR;
}


RdbError::ErrorCode RdbIngresDatabase::VPrepare(const String& id,
						const String& statement)
{
  char *qname = id;
  char *stmt = statement;
    LOG("Preparing "<<qname<<" from "<<stmt<<'\n');
# line 135 "RdbIngresDatabase.scc"	/* prepare */
  {
    IIsqInit(&sqlca);
    IIsqPrepare((int)0,qname,(char *)0,(int)0,stmt);
  }
# line 136 "RdbIngresDatabase.scc"	/* host code */
    if( sqlca.sqlcode )
      return Error(SQL_ERROR);
    return NO_ERROR;
}
