#include <iostream.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include "String.h"



String::Rep::Rep()
{
    Init();
    Resize(0);
}


String::Rep::Rep(const String::Rep& other) // make a duplicate copy
{
    Init();
    Resize(strlen(other.str));
    strcpy(str, other.str);
}


String::Rep::Rep(const char* newStr)
{
    Init();
    Resize(strlen(newStr));
    strcpy(str,newStr);
}


String::Rep::Rep(char** strpp)
{
    Init();
    str = *strpp;
    *strpp = NULL;		// to insure no one else delete's the buffer
    size = strlen(str);
}


String::Rep::~Rep()
{
    assert(ref_count == 0);
    free(str);
}


void String::Rep::Init()
{
    ref_count = 1;
    size = 0;
    str = NULL;
}


void String::Rep::Resize(int newLen)
{
    assert(newLen >= 0);
    if( newLen >= size || newLen < (size>>1) ) {
	// increase buffer when needed,
	// shrink buffer when over 50% is waste.
	size = newLen + 1; // dont forget that null char
	if( str == NULL )
	  str = (char*)malloc(size);
	else
	  str = (char*)realloc( str, size );
	assert(str);
    }
    str[newLen] = '\0';
}


//***************************************************************************


static const int NUM_STR_SIZE = 50; // max chars for converted number (double)


String::String(ostrstream& s)
{
    s << '\0';
    char* cp = s.str();
    rep = new Rep(&cp);
}


String::String(int convert)
{
    char buf[NUM_STR_SIZE];
    sprintf( buf, "%d", convert );
    rep = new Rep(buf);
}


String::String(long convert)
{
    char buf[NUM_STR_SIZE];
    sprintf( buf, "%ld", convert );
    rep = new Rep(buf);
}


String::String( double convert )
{
    char buf[NUM_STR_SIZE];
    sprintf( buf, "%f", convert );
    rep = new Rep(buf);
}


String& String::operator =(const String& other)
{
    if( this != &other ) {	// be careful of s = s
	DeleteRep();
	rep = other.rep->AddReference();
    }
    return *this;
}


String& String::operator =(const char* newStr)
{
    DeleteRep();
    rep = new Rep(newStr);
    return *this;
}


String& String::operator =(char** newStr)
{
    DeleteRep();
    rep = new Rep(newStr);
    return *this;
}


String& String::operator =(ostrstream& s)
{
    DeleteRep();
    s << '\0';
    char* cp = s.str();
    rep = new Rep(&cp);
    return *this;
}


String& String::Concat(const String& other)
{
    UniqueRep();
    int len = Length();
    rep->Resize(len + other.Length());
    strcpy(rep->str + len, other.rep->str); // faster strcat
    return *this;
}


void String::LTrim()
{
    UniqueRep();
    char* cp = rep->str;
    while( *cp && isspace(*cp) ) // find first non-ws
      cp++;
    char* dp = rep->str;
    while( *cp )		// copy rest
      *(dp++) = *(cp++);
}


void String::RTrim()
{
    UniqueRep();
    int l = Length();
    char* cp = rep->str + l -1;
    for( ; l > 0 && isspace(*cp); l-- )
      *(cp--) = '\0';
}


void String::Trim()
{
    LTrim();
    RTrim();
}


ostream& operator <<(ostream& outs, const String& item)
{
    return outs << item.rep->str;
}



