#include <sys/types.h>
#include <sys/time.h>
#include "sybase.h"
#include "RdbSybaseDatabase.h"
#include "Log.h"


RdbSybaseDatabase::RdbSybaseDatabase(RdbSybaseSystem& sybaseSystem,
				     const String newDbName,
				     const String userid,
				     const String password,
				     const String newServerName)
: RdbDatabase(newDbName, sybaseSystem, newServerName),
  system(sybaseSystem)
{
    processPool = new RdbSybaseProcessPool(newDbName, userid, password,
					   newServerName);
}


RdbSybaseDatabase::~RdbSybaseDatabase()
{
    delete processPool;		// in Disconnect??
    Disconnect();
}


RdbError::ErrorCode RdbSybaseDatabase::VConnect()
{
    // nothing happens for sybase
    // each (concurrent) cursor that opens must create a new session.
    connected = TRUE;
    return NO_ERROR;
}


RdbError::ErrorCode RdbSybaseDatabase::VDisconnect()
{
    if( !connected )
      return NO_ERROR;

    // should these be done??
    InvalidateCompilableDMLs();
    InvalidateCursors();

    connected = FALSE;
    return NO_ERROR;
}


String RdbSybaseDatabase::VGetErrorMessage() 
{
    // messages are on a per session (cursor) basis, 
    // or use the global message handler.  Hmm... What to do??
    return "Error messages not yet implemented!\n";
}


RdbError::ErrorCode RdbSybaseDatabase::VImmediateSql(const String& sql)
{
    RETCODE	rc;

    // note: command should not do a commit/rollback (directly or indirectly).
    // This is not currently checked.
    // It should also not attempt to change databases (this will not work)
    
    RDBLOG(DBLog(), "processing immediate command:\n\t" << sql << "\n");
    RdbSybaseProcess dbproc(*this);
    if( dbproc == NULL ) {
	return Error(SQL_ERROR);
    }
    rc = dbcmd(dbproc, sql);
    if( rc == FAIL ) {
	RDBLOG(DBLog(), "failed to create dbcmd\n");
	return Error(SQL_ERROR);
    }
    rc = dbsqlexec(dbproc);
    if( rc == FAIL ) {
	RDBLOG(DBLog(), "failed to send command to sybase\n");
	return Error(SQL_ERROR);
    }
    while( (rc = dbresults(dbproc)) != NO_MORE_RESULTS) {
	if( rc == FAIL ) {
	    RDBLOG(DBLog(), "could not obtain results\n");
	    return Error(SQL_ERROR);
	}
	if( DBROWS(dbproc) == SUCCEED ) { // command wants to return rows
	    RDBLOG(DBLog(), "commands that return results are not really supported...\n");
	    dbprhead(dbproc);	// print result headings on stdout
	    dbprrow(dbproc);	// print all result rows on stdout
	    // dbcanquery(dbproc);
	    // return Error(SQL_ERROR);
	}
    }
    return NO_ERROR;
}


RdbError::ErrorCode RdbSybaseDatabase::VCommit()
{
    RDBLOG(DBLog(), "Sybase commit not yet implemented...\n");
    return Error(SQL_ERROR);

#if NYI
    // Do this??
    InvalidateCompilableDMLs();
    InvalidateCursors();
    ImmediateSql("commit transaction");
    return NO_ERROR;
#endif
}


RdbError::ErrorCode RdbSybaseDatabase::VRollback()
{
    RDBLOG(DBLog(), "Sybase rollback not yet implemented...\n");
    return Error(SQL_ERROR);

#if NYI
    // Do this??
    InvalidateCompilableDMLs();
    InvalidateCursors();
    ImmediateSql("rollback transaction");
    return NO_ERROR;
#endif
}


RdbError::ErrorCode RdbSybaseDatabase::VSetFocus()
{
    return NO_ERROR;
}


RdbError::ErrorCode RdbSybaseDatabase::VRemoveFocus()
{
    return NO_ERROR;
}


RdbError::ErrorCode RdbSybaseDatabase::VPrepare(const String& id,
						const String& statement)
{
    assert(0);			// prepare is not currently used.
    return NO_ERROR;
}
