// This may look like C code, but it is really -*- C++ -*-

//**************************************************************************
// Coral_Utils.S - Utilities for Explain Client
// 
//**************************************************************************
//  File Name:  Coral_Utils
//  File:       /p/coral/dev/src/clients/tree/coral/Coral_Utils.S
//**************************************************************************

#include <stdio.h>
extern  "C" {
#include <tcl.h>
}  // extern "C"

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "Coral.h"
#include "Coral_Utils.h"
#include <new.h>
#include <assert.h>
#include <ctype.h>
#include <time.h>
#include <sys/param.h>
#include <sys/types.h>
#include "Interface.h"

#ifndef    TRUE
#    define    TRUE    (1)
#    define    FALSE    (0)
#endif

/*  #include <sys/select.h>  */

extern int errno;

#include "CoralInit.h"
#include "Globals.h"
#include "ParseStack.h"
#include "CoralError.h"
#include "Util.h"
#include "BuiltinUtil.h"
#include "Database.h"


/**********************************************************************
 * Inputs: A relation name and a predicate
 * Outputs: The arity of that relation
 * Comments: Returns 0 if arity cannot be found.
 **********************************************************************/

int Arity(char *relname, char *pred)
{
    corRelation* rel = find_relation("arity",3);

    if (!rel)
        return(0);

    corTuple tup(make_arglist3(make_arg(relname),
                make_arg(pred), make_var("XX",0)));

    C_ScanDesc scan(rel, &tup);

    corTuple *tuple = scan.next_tuple();

    if (scan.no_match()) {
        fprintf(stderr, "No match from arity\n");
        return(0);
    }

    if (!is_int((*tuple)[2])) {
        fprintf(stderr, "Non-interger answer from arity\n");
        return(0);
    }

    return(make_int((*tuple)[2]));
}

//**************************************************************************

int RegExScan(Tcl_Interp *interp, char *relname, char *pattern )
{
    corRelation *arel;
    FILE *fp;
    char file[64];
    C_ScanDesc *scan;

    strcpy(file,"/tmp/__explain_pattern.F");

    if ((fp = fopen(file, "w")) == NULL){
        Tcl_AppendResult(interp,"coral: ex_open: cannot open file ",file,0);
        return(TCL_ERROR);
    }

    fprintf(fp,"explain__foo(%s).\n", pattern);
    fclose(fp);

    corTuple *tuple;


    \[
        consult("/tmp/__explain_pattern.F").
    \]


    unlink(file);

    corRelation *foo_rel = find_relation("explain__foo", 1);
    C_ScanDesc foo_scan(foo_rel);
    corTuple *foo_t = foo_scan.next_tuple();
    if (foo_scan.no_match()) {
        Tcl_AppendResult(interp,"No value in explain__foo relation!",0);
        return(TCL_ERROR);
    }

    int max_vars = (*foo_t)[0].max_vars();
    corTuple stup(make_arglist4(make_arg(relname),(*foo_t)[0].FuncArgP(),
            make_var("X1",max_vars), make_var("X2",max_vars+1)));

    foo_t->do_delete();

    corRelation* rel = find_relation("instances",4);   // a loaded func

    scan = new C_ScanDesc(rel, &stup);

    Tcl_AppendResult(interp,create_entry("scan",(void *) scan),0);
   return(TCL_OK);
}

//**************************************************************************

int EmptyRel(Tcl_Interp *interp, char *relname, char *str_arity )
{
    corRelation *rel;
    int arity;

   if (Tcl_GetInt(interp,str_arity,&arity) != TCL_OK)
      return(TCL_ERROR);

        /*  If relation does not yet exist this is a NOP  */
    if ((rel = find_relation(relname,arity)) == NULL) {
        return(TCL_OK);
    }

    rel->empty_relation();

    return(TCL_OK);
}

//**************************************************************************

int DumpRel(Tcl_Interp *interp, char *dest_rel, char *src_rel,
                                char *str_arity )
{
    corRelation *dstrel, *srcrel;
    C_ScanDesc *scan;
    corTuple *tuple;
    int arity;

   if (Tcl_GetInt(interp,str_arity,&arity) != TCL_OK)
      return(TCL_ERROR);

        /*    If source relation does not exist this is a NOP */
    if ((srcrel = find_relation(src_rel,arity)) == NULL) {
        return(TCL_OK);
    }

    if ((dstrel = find_relation(dest_rel,arity)) == NULL) {
        dstrel = make_rel(dest_rel,arity);
    }

    scan = new C_ScanDesc(srcrel);

    for (tuple = scan->next_tuple(); !(scan->no_match());
                                        tuple = scan->next_tuple()) {

        dstrel->tuple_insert(tuple);
    } 

   return(TCL_OK);
}

//**************************************************************************
