/*****************************************************************************/
//    AggregateArg.h
/*****************************************************************************/

#ifndef _AggregateArg_H_
#define _AggregateArg_H_

#include "NumArg.h"
#include "StorageRelation.h"

/*****************************************************************************/
//    AvgArg: specialization class used only by Avg aggregate to store running 
//    count.

class corAvgArg : public corNumArg {
public:
    inline corAvgArg(double s) : corNumArg(s)            
     { a_count = 1; 
       _num_kindof = COR_AVG_ARG; 
      }
    inline corAvgArg(int c, double s) : corNumArg(s)
     { a_count = c; 
       _num_kindof = COR_AVG_ARG; 
      }
    inline int count()
     { return a_count; }
protected:
    int  a_count;
};

/*****************************************************************************/
//    DistArg: specialization class used for count_distinct, sum_distinct,
//    prod_distinct

class corDistArg : public corNumArg {
public:
    inline corDistArg(double x, corStorageRelation *r) : corNumArg(x)
     { distrel = r;
       _num_kindof = COR_DIST_ARG; 
      }
    inline corDistArg(int x, corStorageRelation *r) : corNumArg(x)
     { distrel = r; 
       _num_kindof = COR_DIST_ARG; 
     }
    inline corStorageRelation  *rel() { return distrel; }
protected:
    corStorageRelation  *distrel;
};

/*****************************************************************************/
//    ADistArg: specialization class used for avg_distinct

class corADistArg : public corNumArg {
public:
    inline corADistArg(int c, double x, corStorageRelation *r) : corNumArg(x) {
        a_count = c;
        distrel = r; 
	_num_kindof = COR_ADIST_ARG; 
       }
    inline corADistArg(double x, corStorageRelation *r) : corNumArg(x) {
        a_count = 1;
        distrel = r;
	_num_kindof = COR_ADIST_ARG; 
       }
    inline int count() { return a_count; }
    inline corStorageRelation  *rel() { return distrel; }
protected:
    int a_count;
    corStorageRelation *distrel;
};

/*****************************************************************************/

#endif _AggregateArg_H_

