/*****************************************************************************/
//	FuncArg.C
/*****************************************************************************/

#include "FuncArg.h"

#include "Config.h"
#include "BuiltinSymbols.h"
#include "CoreHashTable.h"
#include "Assert.h"
#include "Term.h"

/*****************************************************************************/

#define NilArg corNilSymbol

/*****************************************************************************/

#ifdef USE_HASHCONSING

// WARNING: only FuncArg addresses may be entered into hashcons_table
//			inserting anything else will cause the system to crash! 
//			Further, the addresses inserted should not be equal to 
//			VarHashValue or UnknownHashValue!!

CoreHashTable *corhashcons_table;

int corequal_funcargs(void *a, void*corb);

#endif USE_HASHCONSING

/*****************************************************************************/
corFuncArg * corFuncArg::New(Name op, int arg_count)
{
/*
	FuncArg dummy(op, arg_count);
	struct dummy_class {
		char c[sizeof(FuncArg)];
	};
	int size = sizeof(FuncArg) + arg_count * sizeof(CArg);
	FuncArg *arg = (FuncArg*)new char[size];
	*(dummy_class*)arg = *(dummy_class*)&dummy;
*/
#ifdef WEI_DEBUG
	corC_NewFunctorCount++;
	corC_NewFunctorSize+=size;
#endif
	corFuncArg *arg = new corFuncArg(op, arg_count);
	return arg;
}

/*****************************************************************************/
void corFuncArg::print(corBindEnv* env, FILE *file) const
{
	if (functor() == corListSymbol) {
		fputs("[",file);
		args.print(env, file);
		fputs("]",file);
	}
	else if (functor() == corConsSymbol) {
		fputs("[",file);
		const corFuncArg *cdr = this;
		for (;;) {
			//cdr->args[0]->print(env, file);
			(cdr->args[0]).print(env, file);
			//Arg *cdr_arg = cdr->args[1];
			//cdr = (FuncArg*)cdr_arg;
			corCArg cdr_arg = cdr->args[1];
			cdr = (cdr_arg.FuncArgP());
			if ((corArg*)cdr == (corArg*)NilArg) break;
			if (cdr->kindof() != COR_FUNCTOR
				|| cdr->functor() != corConsSymbol) break;
			fputs(",",file);
		}
		if ((corArg*)cdr != (corArg*)NilArg) {
			fputs("|",file);
			cdr->print(env, file);
		}
		fputs("]",file);
	}
	else {
		fprintf(file,"%s(", corSymbolString(functor()));
		args.print(env, file);
		fputs(")",file);
	}
}

/*****************************************************************************/
void corFuncArg::print(corBindEnv* env, FILE *file, char *) const
{
	print(env, file);
}

/*****************************************************************************/
void corFuncArg::sprint(char* str, int* pos, corBindEnv *env) const
{
	if (functor() == corListSymbol) {
		sprintf(str + *pos, "[");
		(*pos)++ ;
		args.sprint(str, pos, env);
		sprintf(str + *pos, "]");
		(*pos)++ ;
	}
	else if (functor() == corConsSymbol) {
		sprintf(str + *pos, "[");
		(*pos)++ ;
		const corFuncArg *cdr = this;
		for (;;) {
// 			cdr->args[0]->sprint(str, pos, env);
// 			Arg *cdr_arg = cdr->args[1];
// 			cdr = (FuncArg*)cdr_arg;
			(cdr->args[0]).sprint(str, pos, env);
			corCArg cdr_arg = cdr->args[1];
			cdr = (cdr_arg.FuncArgP());
			if ((corArg*)cdr == (corArg*)NilArg) break;
			if (cdr->kindof() != COR_FUNCTOR
				|| cdr->functor() != corConsSymbol) break;
			sprintf(str + *pos, ",");
			(*pos)++ ;
		}
		if ((corArg*)cdr != (corArg*)NilArg) {
			sprintf(str + *pos, "|");
			(*pos)++ ;
			cdr->sprint(str, pos, env);
		}
		sprintf(str + *pos, "]");
		(*pos)++ ;
	}
	else {
		sprintf(str + *pos, "%s(", corSymbolString(functor()));
		*pos += SymbolLength(functor()) + 1;
		args.sprint(str, pos, env);
		sprintf(str + *pos, ")");
		(*pos)++ ;
	}
}

/*****************************************************************************/
int corFuncArg::isConstant() 
{
	if (corisGroundHashValue(_hash))
		return 1;
	if (_hash == VarHashValue)
		return 0;		
	FOR_EACH_ARG(arg, args) {
		if (!arg->isConstant()) {
			_hash = VarHashValue;
			return 0;
		}
	} END_EACH_ARG;
	return 1;
}

/*****************************************************************************/
corCArg corFuncArg::simplify(corBindEnv *env, corTermLink*& renamed_vars, 
				corBindEnv *const_env, corBindEnv *rename_only_env) 
{
    corCArg tmp_carg(this);

	int changes = 0;

	if (isConstant())
		return tmp_carg;

	if (rename_only_env) {
		if (env != const_env && env != rename_only_env){
		  tmp_carg.init();
		  return tmp_carg;
		 }

		// WARNING:
		//	this assumes that if rename_only_var is non-null, the function 
		//	has been called recursively from backpatch_trail.
		//	In this case, there is no need to follow terms that are
		//	interpreted in const_env.

		if (env == const_env)
			return tmp_carg;
	}

	/*************************
	//  Used to do the following as an efficiency hack, but it isn't fast
	//  any more.
	// if ((env==const_env) &&  (env->all_free()) )
	//		return this;
	*************************/

	corFuncArg *this_copy = corFuncArg::New(functor(), count());
	//const Arg** dstPtr = this_copy->args.first();
	int i= 0; 
	FOR_EACH_ARG(arg, args) {
		this_copy->args[i] = arg->simplify(env, renamed_vars, const_env, rename_only_env);
		//if (this_copy->args[i] == NULL) {  // simplify failed due to rename_only_env
		if ((this_copy->args[i]).IsEmpty()) { 
			::delete this_copy;
			tmp_carg.init();
			return tmp_carg;
		}
		//if (this_copy->args[i] != arg) changes = 1;
		if (!(this_copy->args[i]).IsPtrEqual(*arg)) changes = 1;
		i++;
	} END_EACH_ARG;
	if (changes) {
			this_copy->hash(); /* computes, and if the term is ground, stores
								the hash value */
			tmp_carg= this_copy; // Wx:mem leak?
			return tmp_carg;
	}
	else {
		::delete this_copy; /* So that no recursive delete calls are made. */
		return tmp_carg;
	}
}

/*****************************************************************************/
corVarLink * corFuncArg::var_list(corVarLink *rest)
{
	FOR_EACH_ARG(farg, args) {
		rest = farg->var_list(rest);
	} END_EACH_ARG;
	return rest;
}

/*****************************************************************************/
int corFuncArg::equals(ArgPtr arg2)
{
 	if ((arg2)->kindof() != COR_FUNCTOR)
		return (0);
	if ( !(functor()->equals(((corFuncArg*)arg2)->functor()) ))
		return (0);
	if ( arity() != ((corFuncArg*)arg2)->arity() ) 
		return (0);
	if ( _hash != arg2->hash() && corisGroundHashValue(_hash)
				&& corisGroundHashValue(arg2->hash()) )
		return(0);
#if USE_HASHCONSING
	if ( hash() == arg2->hash() && corisGroundHashValue(hash()) ) 
		return (1);
#endif
	/* When called on ground terms, and with hashconsing in use, this
		function will never reach this point */
	for (int i=0; i< arity(); i++)
//		if ( ! args[i]->equals( ((FuncArg *)arg2)->args[i]) )
		if ( ! args[i].equals( ((corFuncArg *)arg2)->args[i]) )
			return(0);
	return (1);
}

/*****************************************************************************/
HashVal corFuncArg::hash(corBindEnv *env /*= 0*/) 
{
	if ( corisGroundHashValue(_hash) )
		return _hash;
	int i;
	int is_const = 1;
	if ( _hash == VarHashValue && env == NULL ) // FuncArg has variables,
						// and they are definitely not bound.
		return VarHashValue;


#define MAX_HASH_ARITY  8
	HashVal *hash_vals;
	HashVal hashval_tab[MAX_HASH_ARITY];
		/** Optimization to reduce calls to new and delete **/
	if (arity() <= MAX_HASH_ARITY)
		hash_vals = hashval_tab;
	else hash_vals = new HashVal[arity()];

	HashVal ihash = SymbolHash(_functor);
	_hash = VarHashValue;
	i=-1;
	FOR_EACH_ARG(farg, args) {
		hash_vals[++i] = farg->hash(env);
		if (hash_vals[i] == VarHashValue) {
			if ( arity() > MAX_HASH_ARITY) 
				delete hash_vals;
			return VarHashValue;
		}
		if ( ! farg->isConstant() ) /* Guaranteed to take constant time for
										all system defined types */
			is_const = 0;
		ihash = (ihash * 3) ^ hash_vals[i];
	} END_EACH_ARG;

	if (is_const) {
#if USE_HASHCONSING
		_hash = (HashVal) corhashcons_table->insert(this, corIntToHash(ihash));
#else
		_hash = corIntToHash(ihash);
#endif
		if ( arity() > MAX_HASH_ARITY) 
			delete hash_vals;
		return (_hash);
	}

#if USE_HASHCONSING

	// Otherwise, a new functor must be created
	corFuncArg *fa = corFuncArg::New(functor(), arity());
	for(i=0; i<arity(); i++) {
	        if ( args[i].isConstant() )
			fa->args[i] = args[i];
		else if (args[i].kindof()==COR_VARIABLE) {
			ASSERT(env != NULL);
			corTerm t = args[i].VarArgP()->dereference(env);
			ASSERT(!(t.expr.IsEmpty()) && t.expr.kindof() != COR_VARIABLE);
			if (t.expr.isConstant())
				fa->args[i] = t.expr;
			else {
				ASSERT(t.expr.kindof()==COR_FUNCTOR);
				fa->args[i] = (corArg *) (hash_vals[i]);  
			}
		}
		else {
			ASSERT(args[i].kindof()==COR_FUNCTOR);
			fa->args[i] = (corArg *) (hash_vals[i]);  
                        // arg[i] is a ground FuncArg. Since hashconsing 
                        // is used, the hash value of each such arg 
                        // is a pointer to a constant term that is 
                        // equal to the arg
		}
	}
	HashVal tmp_hash ; int to_delete = 1 ;
	tmp_hash = (HashVal) corhashcons_table->insert(fa, corIntToHash(ihash), to_delete);

	// Make sure that fa wasn't deleted inside insert()
	if (!to_delete) fa->_hash = tmp_hash ;
#ifdef DEBUG
	else {
	 int size = sizeof(corFuncArg) + arity() * sizeof(ArgPtr);
	 corC_DeleteFunctorCount++;
	 corC_DeleteFunctorSize+=size;
	}
#endif

	if ( arity() > MAX_HASH_ARITY) 
		delete hash_vals;
	return tmp_hash;
#else
	if ( arity() > MAX_HASH_ARITY) 
		delete hash_vals;
	return corIntToHash(ihash);
#endif
}

/*****************************************************************************/
int corFuncArg::scan_vars(void *data, VarFunc *func)
{
	FOR_EACH_ARG(arg, this->args) {
		int code = arg->scan_vars(data, func);
		if (code) return code;
	} END_EACH_ARG
	return 0;
}

/*****************************************************************************/
#ifdef USE_HASHCONSING

void corcreateFunctorTable()
{
	// this is called from init_coral (util.C) so that defaults from exEnv
	// can be used
	corhashcons_table = new CoreHashTable(corequal_funcargs, 1000);
}

#endif USE_HASHCONSING

/*****************************************************************************/
#ifdef USE_HASHCONSING

int corequal_funcargs(void *a, void*corb)
{
  corFuncArg *fa = (corFuncArg *)a;
  corFuncArg *fb = (corFuncArg *)corb;
  int i;
  
  if (fa->functor() != fb->functor() || fa->arity() != fb->arity())
    return(0);
  
  /** Each of the tests in the following loop should be constant time,
   * since the args are ground, and hashconsing is used,
   * and the hash values of the args should have been computed
   * already.
   **/
  for(i=0; i < fa->arity(); i++)
//    if ( ! fa->args[i]->equals(fb->args[i]) ) 
    if ( ! (fa->args[i]).equals(fb->args[i]) ) 
      return(0);
  return(1);
}

#endif USE_HASHCONSING

/*****************************************************************************/

