/*****************************************************************************/
//	GroupingArg.C
/*****************************************************************************/

#include <stdlib.h>

#include "GroupingArg.h"
#include "Globals.h"
#include "AggregateArg.h"
#include "BuiltinSymbols.h"
#include "HashSimpleRelation.h"

/*****************************************************************************/

extern corArrayBindEnv corUnusedEnv;

/*****************************************************************************/
corGrouping::corGrouping(int argcount)
{
	nestedArgs = corFuncArg::New(corNestedSymbol, argcount);
	index = 0;
	kind = NoAggregation;
	_kindof = COR_GROUPING;
}

/*****************************************************************************/
corGrouping::corGrouping(corFuncArg *func, int _index)
{
	nestedArgs = func;
	index = _index;
	kind = NoAggregation;
	_kindof = COR_GROUPING;
}

/*****************************************************************************/
corCArg corGrouping::simplify(corBindEnv *context, corTermLink*& renamed_vars,
	corBindEnv *const_env, corBindEnv *rename_only_env) 
{
 corCArg tmp_carg(this);

	switch (kind) {
	case SetAggregation:
	case NoAggregation:
	{
		corStorageRelation *rel = new corHashSimpleRelation(nested_set_arity, 0,
											   COR_DEFAULT_SET_BUCKETS);
		rel->name = corNestedSymbol;
		//rel->set_subsumption( this->check_subsum );
		corArgList *copy_args = new corArgList(arity());
		// WARNING - the following may not be quite right
		int env_size = corCopyArgs(copy_args, &nested_args(),
								context, arity(), const_env,
								rename_only_env);
		corTuple *tuple1 = new corTuple(copy_args);
		tuple1->env_size = env_size;
		int inserted = 0;
		if (kind == SetAggregation) {
			if (!rel->is_subsumed(tuple1))
				inserted = rel->insert_tuple(tuple1);
		}
		else
			inserted = rel->insert_tuple(tuple1);

		if (!inserted) { delete tuple1; delete copy_args; }
		tmp_carg = rel;
		return tmp_carg;
	}
	case CountAggregation:
	        tmp_carg = 1;
		return tmp_carg;

	case ProductAggregation:
	case SumAggregation:
	case MinAggregation:
	case MaxAggregation:
	case AnyAggregation:
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file, "Grouping::simplify: bad arity\n");
			tmp_carg = 0;
			return tmp_carg;
		}
		return nested_args()[0].simplify(context, renamed_vars, const_env, 
						rename_only_env);
	case AvgAggregation:
	{ 
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file, "Grouping::simplify: bad arity\n");
			tmp_carg =0;
			return tmp_carg;
		}

		corCArg new_val = nested_args()[0].simplify(context, renamed_vars, const_env,
			rename_only_env);

		if (new_val.kindof() != COR_NUM_CONST) { // ERROR!
			 fprintf(corexEnv->error_file,
				"Avg aggregation not a numeric constant!\n");
			 tmp_carg = 0;
			 return tmp_carg;
		}

		corAvgArg *avg = new corAvgArg((new_val.NumDval()));
		tmp_carg = (corArg *)avg;
		return tmp_carg;
	}
	case CountDAggregation:
	{
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file, "Grouping::simplify: bad arity\n");
			 tmp_carg = 0;
			 return tmp_carg;
		}

		corCArg new_valc = nested_args()[0].simplify(context, renamed_vars, const_env, rename_only_env);

		corStorageRelation *new_relc = new corHashSimpleRelation(nested_set_arity, 0,
			COR_DEFAULT_SET_BUCKETS);
		new_relc->name = corNestedSymbol;
		new_relc->set_subsumption(1);
		corArgList *copy_argsc = new corArgList(1);
		//(((Arg**)copy_argsc)+1) = new_valc;
		(*copy_argsc)[0] = new_valc;
		corTuple *new_tuplec = new corTuple(copy_argsc, NULL);
		new_relc->insert_tuple(new_tuplec);

		corDistArg *new_argc = new corDistArg((int)(1),new_relc);
		tmp_carg = (corArg *)new_argc;
		return tmp_carg;
	}

	case AvgDAggregation:
	{
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file, "Grouping::simplify: bad arity\n");
			tmp_carg = 0;
			return tmp_carg;
		}

		corCArg new_vala = nested_args()[0].simplify(context, renamed_vars, const_env,
			rename_only_env);
		
		if (new_vala.kindof() != COR_NUM_CONST) { // ERROR!
			fprintf(corexEnv->error_file,"Aggregation on non-numeric constant!\n");
			tmp_carg = 0;
			return tmp_carg;
		       }

		corStorageRelation *new_rela = new corHashSimpleRelation(nested_set_arity, 0,
											   COR_DEFAULT_SET_BUCKETS);
		new_rela->name = corNestedSymbol;
		new_rela->set_subsumption(1);
		corArgList *copy_argsa = new corArgList(1);
		//(((Arg**)copy_argsa)+1) = new_vala;
		(*copy_argsa)[0] = new_vala;
		corTuple *new_tuplea = new corTuple(copy_argsa, NULL);
		new_rela->insert_tuple(new_tuplea);

		corADistArg *new_arga =
			  new corADistArg((new_vala.NumDval()),new_rela);

		tmp_carg =(corArg *) new_arga;
		return tmp_carg;
	}
	case ProdDAggregation:
	case SumDAggregation:
	{ 
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file, "Grouping::simplify: bad arity\n");
			tmp_carg = 0;
			return tmp_carg;
		}

		corCArg new_val2 = nested_args()[0].simplify(context, renamed_vars, const_env,
			rename_only_env);
		
		if (new_val2.kindof() != COR_NUM_CONST) { // ERROR!
			fprintf(corexEnv->error_file, "Aggregation not a numeric constant!\n");
			tmp_carg = 0;
			return tmp_carg;
		}

		corStorageRelation *new_rel = new corHashSimpleRelation(nested_set_arity, 0,
			COR_DEFAULT_SET_BUCKETS);
		new_rel->name = corNestedSymbol;
		new_rel->set_subsumption(1);
		corArgList *copy_args = new corArgList(1);
		//(((Arg**)copy_args)+1) = new_val2;
		(*copy_args)[0] = new_val2;
		corTuple *new_tuple = new corTuple(copy_args, NULL);
		new_rel->insert_tuple(new_tuple);

		corDistArg *new_arg = new corDistArg((new_val2.NumDval()),new_rel);

		tmp_carg =(corArg *) new_arg;
		return tmp_carg;
	}
	default:
	        
		tmp_carg = 0;
	        return tmp_carg;
	}
}

/*****************************************************************************/
// Count and Sum:   These currently work ONLY if they are called in a 
//				 stratified fashion.  That is, if a relation is defined using
//				 count or sum, it better not be recursive.
// Max and Min:			These currently work in some non-stratified cases.
//				 If a new tuple is found to be better than an old one,
//				 the old one is deleted and the do_insert flag is set so that
//				 the new tuple is inserted later on.
/*****************************************************************************/

/*****************************************************************************/
inline void coralloc_tuple(corTuple *t_old, corTuple *&t_new)
{
	if (!t_new) {
		int i = 0;

		t_new = new corTuple();
		*t_new = *t_old;
		t_new->_args = new corArgList(t_old->arity());
		FOR_EACH_ARG(a, *(t_old->_args)) {
			(*(t_new->_args))[i++] = *a;
		} END_EACH_ARG
	}
}

/*****************************************************************************/
int corupdate_grouping(corRelation *rel, corTuple *old_tuple, corArgList *args,
	corBindEnv *bindings)
{
	corTermLink *renamed_vars = NULL;
	//ASSERT(argnum < old_tuple->arity() );

	corTuple *new_tuple = NULL;
	int grouping_arg_num = -1;
	int stat = 1;

	for (int i=0; i < args->count(); i++) {
		if ( ((*args)[i]).kindof() == COR_GROUPING) {

			grouping_arg_num = i;

			/* This code will not work if val has variables in it */
    // old             Arg *new_val = ((Grouping *)(*args)[i])->first()[0]->

			corCArg new_val = (((*args)[i]).GroupingP()->nested_args())[0].simplify(bindings, renamed_vars, NULL,0);

			if (!new_val.isConstant()) {
				fprintf(corexEnv->error_file,
					"Grouping::update_grouping: value is not a constant.\n");
				return 0;
			}

			stat *= ((*args)[i].GroupingP())->update_tuple(old_tuple,
				new_tuple, new_val, grouping_arg_num, bindings);
		}
	}

	if (grouping_arg_num < 0) fprintf(corexEnv->error_file,
		"Grouping::update_grouping: No grouping operator found.\n");


	if (new_tuple) {
		old_tuple->do_delete();
		rel->insert_tuple(new_tuple);
	}

	return stat;
}

/*****************************************************************************/
int corGrouping::update_tuple(corTuple *old_tuple, corTuple *&new_tuple, corCArg val,
	int argnum, corBindEnv *context)
{
	switch (kind) {
	case SetAggregation:
	case NoAggregation:
	{
 		if (old_tuple->args()[argnum].kindof() != COR_RELATION) {
			fprintf(corexEnv->error_file, 
				"Error: grouping - expected relation argument to insert into");
			return(0);
		}
		// this creates a copy of the original tuple and updates
		// the value in the grouping argument.  the original tuple is left
		// where it was, and it is marked deleted -- so as to not
		// affect indices, etc.

		coralloc_tuple(old_tuple,new_tuple);   // allocate a new tuple if one
										   //  hasn't already been allocated.

		corStorageRelation *rel = new_tuple->args()[argnum].StorageRelP();

		rel->_hash = UnknownHashValue; 
				// so that a new hash value will be computed for the set
		corArgList *copy_args = new corArgList(arity());

		/*********************************
		// WARNING::  ReturnUnify optimizations needed here!!!
		....
		.... See how this is affected by and will affect return unify 
		.... optimizations
		....
		**********************************/

		int env_size = corCopyArgs(copy_args, & nested_args(), context, arity(),
			&corUnusedEnv);
		corTuple *tuple1 = new corTuple(copy_args);
		tuple1->env_size = env_size;
		if (kind == SetAggregation) {
			if (!rel->is_subsumed(tuple1))
				rel->insert_tuple(tuple1);
		}
		else
			rel->insert_tuple(tuple1);

		// NOTE: add rel->env_size and update it here.
		// also add rel->hash_val and rel->id and update them
		// also do this whenever nested relations are created.
		// (see gram.y)

		return 0;
	}
	case CountAggregation:
	{
	 if (arity() != 1) { // ERROR!
	   fprintf(corexEnv->error_file,
		   "Grouping::update_tuple: CountAggregation: bad arity\n");
	   return 0;
	  }
	 
		// this creates a copy of the original tuple with the updated
		// value in the grouping argument.  the original tuple is left
		// where it was, and it is marked deleted -- so as to not
		// affect indices, etc.

		coralloc_tuple(old_tuple,new_tuple);
		new_tuple->args()[argnum]=old_tuple->args()[argnum].NumIval() + 1;
		return 1;
	       }
	case AnyAggregation:
	{
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file,
				"Grouping::update_tuple: AnyAggregation: bad arity\n");
			return 0;
		}
		return 1;
	}
	case SumAggregation:
	{
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file,
				"Grouping::update_tuple: SumAggregation: bad arity\n");
			return 0;
		}
		if (old_tuple->args()[argnum].kindof() != COR_NUM_CONST ||
			val.kindof() != COR_NUM_CONST)
		{ // ERROR!
			fprintf(corexEnv->error_file,
				"Grouping::update_tuple: SumAggregation: not an integer!\n");
			return 0;
		}
				
		// this creates a copy of the original tuple with the updated
		// value in the grouping argument.  the original tuple is left
		// where it was, and it is marked deleted -- so as to not
		// affect indices, etc.

		coralloc_tuple(old_tuple,new_tuple);

		new_tuple->args()[argnum] = 
		  old_tuple->args()[argnum].NumDval() + val.NumDval();
		return 1;
	}
	case ProductAggregation:
	{
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file,
				"Grouping::update_tuple: ProductAggregation: bad arity\n");
			return 0;
		}
		if (old_tuple->args()[argnum].kindof() != COR_NUM_CONST ||
			val.kindof() != COR_NUM_CONST)
		{
			fprintf(corexEnv->error_file,
			   "Grouping::update_tuple: ProductAggregation: not an integer!\n");
   			return 0;
		}
				

		// this creates a copy of the original tuple with the updated
		// value in the grouping argument.  the original tuple is left
		// where it was, and it is marked deleted -- so as to not
		// affect indices, etc.

		coralloc_tuple(old_tuple,new_tuple);

		new_tuple->args()[argnum] =
		  (old_tuple->args()[argnum].NumDval()) * val.NumDval();
		return 1;
	}
	case MinAggregation:
	case MaxAggregation:
	{
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file,
				"Grouping::update_tuple: Min/MaxAggregation: bad arity\n");
			return 0;
		}
		if (old_tuple->args()[argnum].kindof() != COR_NUM_CONST) { // ERROR!
			fprintf(corexEnv->error_file,
				"Grouping::update_tuple: Min/MaxAggregation: not a number\n");
			return 0;
		}

		int c_stat = old_tuple->args()[argnum].compare_num(val);
		if (kind == MaxAggregation)
			c_stat = -c_stat;
		if (c_stat <= 0) { 		// No replacement to be done.
			return 0;
		}

		#ifdef DO_TRACE
		if (corexEnv->dbg_aggregates)
			fprintf(corexEnv->trace_file,
				"Min/Max aggregation - new extreme value found\n");
		#endif

		coralloc_tuple(old_tuple,new_tuple);

		if ( val.cNum_kindof() == COR_INTEGER) 
			new_tuple->args()[argnum] =  val.NumIval();
		else
			new_tuple->args()[argnum] = val.NumDval();

		return 1;
	}
	case AvgAggregation:
	{
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file,
				"Grouping::update_tuple: AvgAggregation: bad arity\n");
			return 0;
		}

		if ((old_tuple->args()[argnum]).AvgArgP()->kindof() != COR_NUM_CONST ||
			val.kindof() != COR_NUM_CONST)
		{ // ERROR!
			fprintf(corexEnv->error_file,
				"Grouping::update_tuple: AvgAggregation: not a numeric constant!\n");
			return 0;
		}
				
		// this creates a copy of the original tuple with the updated
		// value in the grouping argument.  the original tuple is left
		// where it was, and it is marked deleted -- so as to not
		// affect indices, etc.

		coralloc_tuple(old_tuple,new_tuple);

		int cnt = ((old_tuple->args()[argnum]).AvgArgP())->count();
		int new_cnt = cnt + 1;

		double av =  ( (double)(cnt)
			* ((old_tuple->args()[argnum]).AvgArgP())->doubval()
			+ (val.NumDval()) ) / (double)(new_cnt);

		 new_tuple->args()[argnum] = new corAvgArg(new_cnt, av);

		 return 1;
	}
	case AvgDAggregation:
	{
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file, "Grouping operator bad arity\n");
			return 0;
		}

		if ((old_tuple->args()[argnum]).DistArgP()->kindof() != COR_NUM_CONST ||
			val.kindof() != COR_NUM_CONST)
		{ // ERROR!
			fprintf(corexEnv->error_file,
				"Numeric Aggregation on non-numeric constant\n");
			return 0;
		}
				
		corArgList *val_lista = new corArgList(1);
		// (((Arg**)val_lista)+1) = val;
		(*val_lista)[0] = val;
		corTuple *val_tuplea = new corTuple(val_lista, NULL);

		corStorageRelation *drela =
			((old_tuple->args()[argnum]).ADistArgP())->rel();

		if (drela->is_subsumed(val_tuplea)) {
		   // then not a distinct element, so don't need to do anything.
		   return 1;
		}

		// otherwise, need to update the grouping value, and the set

		drela->insert_tuple(val_tuplea);
		
		coralloc_tuple(old_tuple,new_tuple);

		int countd = ((old_tuple->args()[argnum]).ADistArgP())->count();
		int new_countd = countd + 1;

		double avd = ( (double)(countd)
			* ((old_tuple->args()[argnum]).ADistArgP())->doubval()
			+ (val.NumDval()) ) / (double)(new_countd);

		 new_tuple->args()[argnum] = new corADistArg(new_countd, avd, drela);

		 return 1;
	}
	case CountDAggregation:
	{
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file, "Grouping operator bad arity\n");
			return 0;
		}

		corArgList *val_listc = new corArgList(1);
		//(((Arg**)val_listc)+1) = val;
		(*val_listc)[0] = val;
		corTuple *val_tuplec = new corTuple(val_listc, NULL);

		corStorageRelation *drelc = 
			   ((old_tuple->args()[argnum]).DistArgP())->rel();

		if (drelc->is_subsumed(val_tuplec)) {
		   // then not a distinct element, so don't need to do anything.
		   return 1;
		}

		// otherwise, need to update the grouping value, and the set

		drelc->insert_tuple(val_tuplec);
		
		coralloc_tuple(old_tuple,new_tuple);

		new_tuple->args()[argnum] = 
		  new corDistArg(((old_tuple->args()[argnum]).DistArgP())->i_val+1,drelc);

		return 1;
	}
	case ProdDAggregation:
	case SumDAggregation:
	{
		if (arity() != 1) { // ERROR!
			fprintf(corexEnv->error_file, "Grouping operator bad arity\n");
			return 0;
		}

		if ((old_tuple->args()[argnum]).DistArgP()->kindof() != COR_NUM_CONST ||
			val.kindof() != COR_NUM_CONST)
		{ // ERROR!
			fprintf(corexEnv->error_file,
				"Numeric Aggregation on non-numeric constant\n");
			return 0;
		}
				
		corArgList *val_list = new corArgList(1);
		//(((Arg**)val_list)+1) = val;
		(*val_list)[0] = val;
		corTuple *val_tuple = new corTuple(val_list, NULL);

		corStorageRelation *drel=((old_tuple->args()[argnum]).DistArgP())->rel();
		if (drel->is_subsumed(val_tuple)) {
		   // then not a distinct element, so don't need to do anything.
		   return 1;
		}

		// otherwise, need to update the grouping value, and the set

		drel->insert_tuple(val_tuple);
		
		coralloc_tuple(old_tuple,new_tuple);

		if (kind == SumDAggregation) {
			new_tuple->args()[argnum] =
				new corDistArg(((old_tuple->args()[argnum]).DistArgP())->doubval()
				+ (val.NumDval()),drel);
		}
		else {
			new_tuple->args()[argnum] =
				new corDistArg(((old_tuple->args()[argnum]).DistArgP())->doubval()
				* (val.NumDval()),drel);
		}
		return 1;
	}
	default:
		return 0;
	}
}

/*****************************************************************************/
void corGrouping::print(corBindEnv* env, FILE *file, char *) const
{
	this->print(env, file);
}

/*****************************************************************************/
void corGrouping::print_dump(corBindEnv *env, FILE *file) const
{
	fprintf(file, "\"");
	print(env, file);
	fprintf(file, "\"");
}

/*****************************************************************************/
void corGrouping::print(corBindEnv* env, FILE *file) const
{
	switch (kind) {
	case NoAggregation:
		fputs("<", file); break;
	case SetAggregation:
		fputs("distinct(<", file); break;
	case SumAggregation:
		fputs("sum(<", file); break;
	case ProductAggregation:
		fputs("prod(<", file); break;
	case CountAggregation:
		fputs("count(<", file); break;
	case MinAggregation:
		fputs("min(<", file); break;
	case MaxAggregation:
		fputs("max(<", file); break;
	case AvgAggregation:
		fputs("avg(<", file); break;
	case SumDAggregation:
		fputs("sum_distinct(<", file); break;
	case ProdDAggregation:
		fputs("prod_distinct(<", file); break;
	case CountDAggregation:
		fputs("count_distinct(<", file); break;
	case AvgDAggregation:
		fputs("avg_distinct(<", file); break;
	case AnyAggregation:
		fputs("any(<", file); break;
	default:
		break;
	}
	corArgList arg_ptr = nested_args();
	int i = arity();
	for (;;) {
	  i--;
		(arg_ptr[i]).print(env, file);
		if (i == 0)
			break;
		fputs(", ", file);
	};
	fprintf(file, kind == NoAggregation ? ">" : ">)");
}

/*****************************************************************************/
void corGrouping::sprint(char *str, int *pos, corBindEnv *env) const
{
	switch (kind) {
	case NoAggregation:
		sprintf(str + *pos, "<");
		(*pos)++ ;
		break;
	case SumAggregation:
		sprintf(str + *pos, "sum(<");
		*pos += 5;
		break;
	case ProductAggregation:
		sprintf(str + *pos, "prod(<");
		*pos += 6;
		break;
	case CountAggregation:
		sprintf(str + *pos, "count(<");
		*pos += 7 ;
		break;
	case MinAggregation:
		sprintf(str + *pos, "min(<");
		*pos += 5;
		break;
	case MaxAggregation:
		sprintf(str + *pos, "max(<");
		*pos += 5;
		break;
	case AvgAggregation:
		sprintf(str + *pos, "avg(<");
		*pos += 5;
		break;
	case SumDAggregation:
		sprintf(str + *pos, "sum_distinct(<");
		*pos += 14;
		break;
	case ProdDAggregation:
		sprintf(str + *pos, "prod_distinct(<");
		*pos += 15;
		break;
	case CountDAggregation:
		sprintf(str + *pos, "count_distinct(<");
		*pos += 16;
		break;
	case AvgDAggregation:
		sprintf(str + *pos, "avg_distinct(<");
		*pos += 14;
		break;
	case SetAggregation:
		sprintf(str + *pos, "distinct(<");
		*pos += 10;
		break;
	case AnyAggregation:
		sprintf(str + *pos, "any(<");
		*pos += 5;
		break;
	default :
		break;
	}
	corArgList arg_ptr = nested_args();
	int i = arity();
	for (;;) {
	  i--;
		(nested_args()[i]).sprint(str, pos, env);
		if (i == 0)
			break;
		sprintf(str + *pos, ", ");
		*pos += 2 ;
	};
	sprintf(str + *pos, kind == NoAggregation ? "> " : ">)");
	*pos += 2 ;
}

/*****************************************************************************/
int corGrouping::equals(corArg *arg2)
{
	if (arg2->kindof() != COR_GROUPING)
	return (0);
	fprintf(corexEnv->error_file, "Grouping::equals not yet implemented \n");
	return(0);
}

/*****************************************************************************/

