/*****************************************************************************/
//	GroupingArg.h
/*****************************************************************************/

#ifndef _GroupingArg_H_
#define _GroupingArg_H_

#include "Arg.h"
#include "FuncArg.h"

/*****************************************************************************/
// WARNING : Use Grouping::New to get a new Grouping (it is of variable size)
/*****************************************************************************/

class corGrouping : public corArg {
public:
	corFuncArg * nestedArgs;
	int index;							 // Sequence number within current clause.

	short check_subsum;

	enum AggregateKind kind;
	// If kind==NoAggregation, then 'this' is the parse of '<ARGS>',
	// where 'nestedArgs' represents 'ARGS'.
	// Otherwise, if kind is (say) SummAggregation,
	// then 'this' is the parse of 'sum<ARGS>'.

	corGrouping(int argcount);
	corGrouping(corFuncArg *func, int index);
	corGrouping(corFuncArg *func, int _index, int k) {
		nestedArgs = func;
		index = _index;
		kind = (enum AggregateKind)k;
		_kindof = COR_GROUPING;
	}
	static corGrouping* New(int arity) { return new corGrouping(arity); }
	virtual ~corGrouping() {}

	virtual int equals(corArg *arg2);
	virtual int isConstant() { return 1; }
	virtual void set_subsumption(short subsum_flag) {
		check_subsum = subsum_flag;
	}

	virtual void print_dump(corBindEnv * env, FILE *file) const;
	virtual void print(corBindEnv* env, FILE *file) const;
	virtual void print(corBindEnv* env, FILE *file, char *fmt_str) const;
	virtual void sprint(char *str, int *pos, corBindEnv *context = NULL) const;
	// a grouping has an arglist inside:
	inline corArgList & nested_args() const { return nestedArgs->args; }

	int arity() const { return nestedArgs->count(); }
	virtual corCArg simplify(corBindEnv *context, corTermLink *&renamed_vars,
		corBindEnv *const_env, corBindEnv *rename_only_env);
	// WARNING:: This assumes that nested sets are ground!! This ought to
	// 		be the case, but we should make sure using the 
	//	 	hash value.
	int update_tuple(corTuple *old_tuple, corTuple *&new_tuple, corCArg val,
		int argnum, corBindEnv *context);
};

/*****************************************************************************/

int corupdate_grouping(corRelation *rel, corTuple *old_tuple, corArgList *args,
	corBindEnv *bindings);

/*****************************************************************************/

#endif _GroupingArg_H_

