/*****************************************************************************/
//    ArrayBindEnv.C
/*****************************************************************************/

#include <string.h>

#include "ArrayBindEnv.h"
#include "VersionedBindEnv.h"
#include "ObjectCount.h"

/*****************************************************************************/
corArrayBindEnv::corArrayBindEnv(int size, int )
{
    _limit = size;
    bindarray = new corTerm [size];
    memset(bindarray, 0, size * sizeof(corTerm));
    _is_versioned = 0;

#    ifdef DEBUG
        corC_ConstructEnvCount++;
        corC_ConstructEnvSize += _limit;
#    endif
}

/*****************************************************************************/
corArrayBindEnv::corArrayBindEnv()
{
    _limit = 0;
    bindarray = NULL;
    _is_versioned = 0;

#    ifdef DEBUG
        corC_ConstructEnvCount++;
#    endif
}

/*****************************************************************************/
corArrayBindEnv::~corArrayBindEnv()
{
    delete [] bindarray;
#    ifdef DEBUG
        corC_DestructEnvCount++;
        corC_DestructEnvSize += _limit;
#    endif
}

/*****************************************************************************/
int corArrayBindEnv::all_free()
{
    for(register int i=0; i < _limit; i++) {
        if (bindarray[i].expr.IsEmpty())
            return 0;
    }
    return 1;
}

/*****************************************************************************/
// WARNING - got to change this once versioned bindenvs are implemented
//       properly.

corBindEnv* corArrayBindEnv::make_version()
{
    corVersionedBindEnv *newenv = new corVersionedBindEnv(_limit);

    for (register int i=0; i< _limit; i++) {
#        ifdef USE_TRIES
            newenv->forcebind(i, bindarray[i]);
#        else
            newenv->bind(i, bindarray[i]);
#        endif
    }

    return ((corBindEnv*) newenv);
}

/*****************************************************************************/
void corArrayBindEnv::resize(int newsize)
{
    corTerm *newbindarray = new corTerm [newsize];

    for (int i=0; i < _limit; i++)
       newbindarray[i] = bindarray[i];

    memset(&(newbindarray[_limit]), 0, (newsize-_limit) * sizeof(corTerm));
    corTerm *oldbindarray = bindarray;
    bindarray = newbindarray;
    _limit = newsize;
    delete oldbindarray;
}

/*****************************************************************************/
// This is a hack --- see note on versioning-hack for more information

corBindEnv* corArrayBindEnv::make_version_and_swap()
{
    corArrayBindEnv *newenv = new corArrayBindEnv(_limit);

    for (register int i=0; i< _limit; i++)
        newenv->bindarray[i] = bindarray[i];

    // Swap bindarrays, so that backpatching trashes the new version.
    //  the old version is still available for reuse.

    corTerm *temparray = bindarray;
    bindarray = newenv->bindarray;
    newenv->bindarray = temparray;
    return ((corBindEnv*) newenv);
}

/*****************************************************************************/
// This is a hack --- see note on versioning-hack for more information
// env has to be an array bindenv

void corArrayBindEnv::restore(corBindEnv *env)
{
    corArrayBindEnv *newenv = (corArrayBindEnv *)env;

    bindarray = newenv->bindarray;
    _limit = newenv->_limit;
}

/*****************************************************************************/
// This is a hack -- see note on versioning-hack for more information
// Copy out the structure, but not the contents
// Use this very carefully --- deleting bindenvs will delete the
// bindarray too, so better use ::delete on the original bindenv
// rather than delete.

corBindEnv* corArrayBindEnv::copy_shell()
{
    corArrayBindEnv *newenv = new corArrayBindEnv();
    newenv->_limit = _limit;
    newenv->bindarray = bindarray;
    return newenv;
}

/*****************************************************************************/

