/*****************************************************************************/
//    VersionedBindEnv.h
/*****************************************************************************/

#ifndef _VersionedBindEnv_H_
#define _VersionedBindEnv_H_

#include "Term.h"
#include "BindEnv.h"

/*****************************************************************************/

class corTrieNode {
public:
    int height;

    // NOTE: The following two should be a union, but C++ doesn't like it
    corTrieNode *pointers[2];
    corTerm      binding;

    corTrieNode* copy(); 
};

/*****************************************************************************/

class corVersionedBindEnv: public corBindEnv {
    friend class corBindEnv;
    friend class corArrayBindEnv;
#ifdef USE_TRIES
    corTrieNode* trie;
#else
    corTerm* bindarray;
#endif
    corVersionedBindEnv();

public:
    corVersionedBindEnv(int size, int incr=-1);
#ifdef NOBINDENVHACK
    virtual corTerm lookup(Variable var) const;
    virtual int bind(Variable var, corTerm binding);
#else
    virtual corTerm lookupV(Variable var) const;
    virtual int bindV(Variable var, corTerm binding);
#endif
    virtual int forcebind(Variable var, corTerm binding);
    virtual void resize(int newsize);  // Aborts
    virtual int all_free();
    virtual corBindEnv *make_version();  // returns a new version of the bindenv
    virtual corBindEnv *make_version_and_swap();
    virtual corBindEnv *copy_shell();
    virtual void restore(corBindEnv *env);
    virtual ~corVersionedBindEnv();    
};

/*****************************************************************************/

#endif _VersionedBindEnv_H_

