/*****************************************************************************/
//    BuiltinRelation.C
/*****************************************************************************/

#include "BuiltinRelation.h"

#include "Config.h"
#include "Globals.h"
#include "CoralError.h"
#include "InitBuiltins.h"
#include "Interface.h"        // add builtin
#include "StackMark.h"
#include "TupleIterator.h"
#include "CArg.h"

/*****************************************************************************/
int corBuiltinRelation::insert_tuple(corTuple *tuple)
{
    corArgList& args = tuple->args();
    int env_size = tuple->env_size;

    corStackMark stackmark0;
    corArrayBindEnv *env = new corArrayBindEnv (env_size);
    corTupleIterator it(this, args, env, NULL, NULL);
    it.get_next();
    int stat =    !(it.no_match());
    stackmark0.pop_to();
    delete env;
    //return stat;
    return 0;
}

/*****************************************************************************/
int corBuiltinRelation::insert_new(corArgList &args, corBindEnv *env, corBindEnv *,
    corTuple *, int )
{
    corBindEnv *new_env;
    int env_size;

    if (env==NULL) {
          env_size = args.max_vars();
          if (env_size != 0)
            new_env = new corArrayBindEnv(args.max_vars());
        else
            new_env = NULL;
    }
    else 
          new_env = env;

    corStackMark stackmark0;
    corTupleIterator it(this, args, new_env, NULL, NULL);
    it.get_next();
    int stat = !(it.no_match());
    stackmark0.pop_to();
    if (new_env && new_env != env) 
        delete new_env;
    return stat;
}

/*****************************************************************************/
corBuiltinRelation::corBuiltinRelation(int a, Name n, SolveProc solve_proc):
corRelation()
{
    count = -1;  // NOTE: this is a kludge to easily identify a builtin relation
    single_success = 1;
    _arity = a;
    name = n;
    solver = solve_proc;
    ref_count = 1;
    corInsertBuiltin(n, this);
}

/*****************************************************************************/
corBuiltinRelation::corBuiltinRelation(char *predname, int arity, UserRelProc solver)
{
    add_builtin(predname, arity, solver);
}

/*****************************************************************************/
corBuiltinRelation::corBuiltinRelation(char *predname, int arity,
    UserTupleProc solver)
{
    add_builtin(predname, arity, solver);
}

/*****************************************************************************/
void corBuiltinRelation::print_name(FILE *outf)
{
    if (is_printable_builtin()) {
        if (arity() >= 0) {
            fprintf(outf,"%15s/%d\t : (builtin)\n",corSymbolString(name),arity());
        }
        else {
            fprintf(outf,"%15s/-\t : (builtin)\n",corSymbolString(name));
        }
    }
}

/*****************************************************************************/
corBindEnv * corBuiltinRelation::get_next(corTupleIterator &iterator)
{
    if (iterator.ipos > 0) { 
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
#        ifdef DO_TRACE
            if (corexEnv->dbg_get_next) {
                fprintf(corexEnv->trace_file,
                    "Builtin Relation: %s - no more answers\n",
                    name->string());
            }
#        endif
        return iterator.bindenv;
    }
#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file,
                "Builtin Relation: Starting on %s ( ", name->string());
            iterator.arg_list.print(iterator.bindenv, corexEnv->trace_file);
            fprintf(corexEnv->trace_file,")\n");
        }
#    endif
    iterator.stack_mark.get_mark();

    iterator.ipos++;
    return (solver(*this, iterator));
}

/*****************************************************************************/
void corBuiltinRelation::release(corTupleIterator &)
{
    //    Do nothing.  Subclasses of builtin relation that store structures in
    //    TupleIterator may redefine this function to release allocated space.
}

/*****************************************************************************/

