/*****************************************************************************/
//    BuiltinUnaryRelation.C
/*****************************************************************************/

#include "BuiltinUnaryRelation.h"

#include "InitBuiltins.h"
#include "TupleIterator.h"
#include "Globals.h"
#include "CoralError.h"

/*****************************************************************************/
corBuiltinUnaryRelation::corBuiltinUnaryRelation(int a, Name n, AggProc agg_op)
    : corBuiltinRelation(a, n, NULL)
{
    op = agg_op;
}

/*****************************************************************************/
corBindEnv * corBuiltinUnaryRelation::get_next(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return env;
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if ((iterator.arg_list).count() < 2) {
        fprintf(corexEnv->error_file, "Error: Not enough arguments provided.\n");
        return NULL;
    }

   // Term terms[2];
    for (int i = 0; i < 2; i++) {
        corstatic_terms[i].expr = iterator.arg_list[i];
        corstatic_terms[i].bindenv = env;
        FULL_DEREFERENCE_TERM(corstatic_terms[i]);
    }

    corRelation *rel;
    switch (corstatic_terms[0].kindof()) {
        case COR_RELATION:
            rel = (corRelation*)corstatic_terms[0].expr.RelationP();
            break;
        case COR_FUNCTOR:
            rel = find_relation((corstatic_terms[0].expr.FuncArgP())->_functor,
                (corstatic_terms[0].expr.FuncArgP())->args.count());
            if (rel == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return NULL;
            }
            break;
         case COR_SYMBOL_CONST:
            rel = find_relation(corstatic_terms[0].expr.SymbolP(), -1);
            if (rel == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, "get_next");
                iterator.stack_mark.pop_to();
                return NULL;
            }
            break;
        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "get_next");
            iterator.stack_mark.pop_to();
            return NULL;
    }

    //Term result_term;
    corusable_terms[0].expr = (*op)(rel);
    corusable_terms[0].bindenv = NULL;

    if (corusable_terms[0].expr.IsEmpty()) {
        iterator.set_no_match();
        iterator.stack_mark.pop_to();
        return env;
    }

    if ( ! corunify_args (corusable_terms[0], corstatic_terms[1])) {
        iterator.set_no_match();
        iterator.stack_mark.pop_to();
    }

    return(env);
}

/*****************************************************************************/

