/*****************************************************************************/
//    BuiltinUserRelRelation.C
/*****************************************************************************/

#include "BuiltinUserRelRelation.h"

#include "TupleIterator.h"

/*****************************************************************************/
corBuiltinUserRelRelation::corBuiltinUserRelRelation(int arity, Name n,
    UserRelProc user_op)
    : corBuiltinRelation(arity, n, NULL)
{
    op = user_op;
}

/*****************************************************************************/
/*
    BuiltinUserRelRelation::get_next  is called only the first time.
    This computes the desired relation, and then destructively
    updates the tuple iterator, so that the `relation' field points
    to the user returned relation.  It then directly calls the 
    get_next on the tuple iterator, which should work fine since
    the built-in relation has been replaced by a fixed relation.
    Subsequent calls to get_next, also work with the modified 
    relation field.
*/

corBindEnv* corBuiltinUserRelRelation::get_next(corTupleIterator& iterator)
{
    corTuple *tuple = get_next_tuple(iterator);
    if (iterator.no_match())
        return NULL;
    return(iterator.bindenv);
}

/*****************************************************************************/
corTuple* corBuiltinUserRelRelation::get_next_tuple(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;
    corBindEnv *queryenv = NULL;

    if (iterator.arg_list.arity() != arity()) {
      iterator.set_no_match(); 
      return NULL;
    }


    // first create the query tuple from the bindenv
    corArgList *alist = new corArgList(arity());

    for (int i = 0; i < arity(); i++) {
        corTerm term(iterator.arg_list[i], env);
        FULL_DEREFERENCE_TERM(term);
        if (!(((term.kindof() == COR_VARIABLE) && (term.bindenv == env))
            || (term.isConstant())))
        {
            term.expr = iterator.arg_list[i];  // Dereferencing lead to a
            // different bindenv -- cannot use dereferenced term.expr
            // with iterator.bindenv, so revert to original arg_list[i]
            queryenv = iterator.bindenv;      
        }
        (*alist)[i] = term.dereference().expr;
    }

    corTuple query(alist, queryenv); 

    // now call the user routine to compute the (single) answer to the query
    corRelation *result = (corRelation *) (*op)(&query); 

    if (result == NULL) {
        // no answer to query
        iterator.set_no_match();
        return NULL;
    }

    // now destructively modify the tuple iterator to point to the 
    // computed relation, and call the iterator's get_next_tuple which
    // will now work on the computed relation.

    iterator.relation = result;
    return iterator.get_next_tuple();
}

/*****************************************************************************/

