//**************************************************************************
//    GlobalBuiltinVars.C
//**************************************************************************

#include "GlobalBuiltinVars.h"

#include "Globals.h"
#include "CoralError.h"

//**************************************************************************

    // To speed things up a little
#define STRCMP(X,Y) \
    ((*(X) < *(Y)) ? (-1) : ((*(X) > *(Y)) ? (1) : strcmp((X),(Y))))

//**************************************************************************

  // PFI - Pointer to function returning int
typedef int    (*PFI)(int    val);

  // PFD - Pointer to function returning double
typedef double (*PFD)(double val);

//**************************************************************************
// These one-liners are used in an array of function pointers at the end
// of this section.
//**************************************************************************

static int set_answer_style( int value )
     { return(corexEnv->C_answer_style_default = value); }

static int set_check_subsumption( int value )
     { return(corexEnv->C_check_subsum_default = value); }

static int set_convert_functions( int value )
     { return(corexEnv->C_convert_functions_default = value); }

static int set_create_magic( int value )
     { return(corexEnv->C_create_magic_default = value); }

static int set_debug_aggregates( int value )
     { return(corexEnv->dbg_aggregates = value); }

static int set_debug_backtrack( int value )
     { return(corexEnv->dbg_backtrack = value); }

static int set_debug_get_next( int value )
     { return(corexEnv->dbg_get_next = value); }

static int set_debug_indexing( int value )
     { return(corexEnv->dbg_indexing = value); }

static int set_debug_ordered_search( int value )
     { return(corexEnv->dbg_ordsearch = value); }

static int set_debug_pipelining( int value )
     { return(corexEnv->dbg_pipelining = value); }

static int set_debug_return_unify( int value )
     { return(corexEnv->dbg_return_unify = value); }

static int set_debug_sets( int value )
     { return(corexEnv->dbg_sets = value); }

static int set_debug_subsumes( int value )
     { return(corexEnv->dbg_subsumes = value); }

static int set_debug_unify( int value )
     { return(corexEnv->dbg_unify = value); }

static int set_delta_buckets( int value )
     { return(corexEnv->C_delta_buckets_default = value); }

static int set_existential( int value )
     { return(corexEnv->C_use_exist_opt_default = value); }

static int set_factor_magic( int value )
     { return(corexEnv->C_use_factor_magic_default = value); }

static int set_factoring( int value )
     { return(corexEnv->C_use_factoring_default = value); }

static int set_hash_buckets( int value )
     { return(corexEnv->C_hash_buckets_default = value); }

static int set_history( int value )
     { return(corexEnv->C_do_history_default = value); }

static double set_incr_ratio( double value )
     { return(corexEnv->C_incr_ratio_default = value); }

static int set_index_deltas( int value )
     { return(corexEnv->C_index_deltas_default = value); }

static int set_insert_mode( int value )
     { return(corexEnv->C_insert_mode_default = value); }

static int set_interactive_mode( int value )
     { return(corexEnv->C_interactive_mode_default = value); }

static int set_lazy_eval( int value )
     { return(corexEnv->C_use_eager_eval_default = value); }

static int set_magic( int value )
     { return(corexEnv->C_use_magic_default = value); }

static double set_max_occupancy( double value )
     { return(corexEnv->C_max_occupancy_default = value); }

static int set_monotonic( int value )
     { return(corexEnv->C_use_monotonic_default = value); }

static int set_multiset( int value )
     { return(corexEnv->C_multisets_default = value); }

static int set_nested_scc_eval( int value )
     { return(corexEnv->C_nested_scc_eval_default = value); }

static int set_non_ground_facts( int value )
     { return(corexEnv->C_non_ground_facts_default = value); }

static int set_ordered_search( int value )
     { return(corexEnv->C_use_ordsearch_default = value); }

static int set_pipelining( int value )
     { return(corexEnv->C_use_pipelining_default = value); }

static int set_preprocessing( int value )
     { return(corexEnv->C_preprocessing_default = value); }

static int set_print_prompt( int value )
     { return(corexEnv->C_print_prompt_default = value); }

static int set_profile_module( int value )
     { return(corexEnv->profile_module = value); }

static int set_profile_scc( int value )
     { return(corexEnv->profile_scc = value); }

static int set_profile_sn( int value )
     { return(corexEnv->profile_sn = value); }

static int set_psn( int value )
     { return(corexEnv->C_use_predicate_sn_default = value); }

static int set_quiet_mode( int value )
     { return(corexEnv->C_quiet_mode_default = value); }

static int set_read_coralrc( int value )
     { return(corexEnv->C_read_coralrc_default = value); }

static int set_relation_type( int value )
     { return(corexEnv->C_relation_type_default = (RelationTypeDefault) value); }

static int set_return_unify( int value )
     { return(corexEnv->C_use_return_unify_default = value); }

static int set_rewriting( int value )
     { return(corexEnv->C_rewriting_default = value); }

static int set_save_module( int value )
     { return(corexEnv->C_save_module_default = value); }

static int set_single_answer( int value )
     { return(corexEnv->C_single_answer_default = value); }

static int set_single_scc( int value )
     { return(corexEnv->C_single_scc_default = value); }

static int set_sup_magic( int value )
     { return(corexEnv->C_use_supplementary_magic_default = value); }

static int set_sup_magic_indexing( int value )
     { return(corexEnv->C_sup_magic_indexing_default = value); }

static int set_use_quotes( int value )
     { return(corexEnv->C_use_quotes = value); }

static int set_verbose_answer( int value )
     { return(corexEnv->C_verbose_answer_default = value); }

    // This can't be done in one line, (oh well, there goes the symmetry)
static int set_recursion_depth(int value)
    {
        // Do some appropriate action here !! -- PRAVEEN
        if (value <= corT_Stack.count) {
            fprintf(corexEnv->error_file,
                 "CORAL Error:: New value should be greater"
                 " than total stack depth %d !\n", corT_Stack.count);
            return(corexEnv->C_max_recursion_depth_default);
        }
        corexEnv->C_max_recursion_depth_default = value;
        corTable **tmp = corT_Stack.tables;
        corT_Stack.tables    = new corTable*[value];
        for (int i = 0; i < corT_Stack.count; i++)
            corT_Stack.tables[i] = tmp[i];
        delete tmp;
        return(value);
    }

//**************************************************************************
//  This struct is used for all global variables that the user might want
//  to set, clear or assign from the CORAL prompt.
//
//  NOTE: The variable names should be listed in the array alphabetically.
//
//**************************************************************************

// change this when changing the GlobalVarTable !!
#define COR_GLOBAL_VAR_TABLE_SIZE    52

struct {
    char    *var_name;
    PFI      func_ptr;

} corGlobalVarTable[COR_GLOBAL_VAR_TABLE_SIZE] = {

  // The variables must be listed in ALPHABETICAL Order

    {"answer_style", set_answer_style},

    {"check_subsumption", set_check_subsumption},
    {"convert_functions", set_convert_functions},
    {"create_magic", set_create_magic},

    {"debug_aggregates", set_debug_aggregates},
    {"debug_backtrack", set_debug_backtrack},
    {"debug_get_next", set_debug_get_next},
    {"debug_indexing", set_debug_indexing},
    {"debug_ordered_search", set_debug_ordered_search},
    {"debug_pipelining", set_debug_pipelining},
    {"debug_return_unify", set_debug_return_unify},
    {"debug_sets", set_debug_sets},
    {"debug_subsumes", set_debug_subsumes},
    {"debug_unify", set_debug_unify},
    {"delta_buckets", set_delta_buckets},

    {"existential", set_existential},

    {"factor_magic", set_factor_magic},
    {"factoring", set_factoring},

    {"hash_buckets", set_hash_buckets},
    {"history", set_history},

    {"incr_ratio", (PFI) set_incr_ratio},
    {"index_deltas", set_index_deltas},
    {"insert_mode", set_insert_mode},
    {"interactive_mode", set_interactive_mode},

    {"lazy_eval", set_lazy_eval},

  // The variables must be listed in ALPHABETICAL Order

    {"magic", set_magic},
    {"max_occupancy", (PFI) set_max_occupancy},
    {"monotonic", set_monotonic},
    {"multiset", set_multiset},

    {"nested_scc_eval", set_nested_scc_eval},
    {"non_ground_facts", set_non_ground_facts},

    {"ordered_search", set_ordered_search},

    {"pipelining", set_pipelining},
    {"preprocessing", set_preprocessing},
    {"print_prompt", set_print_prompt},
    {"profile_module", set_profile_module},
    {"profile_scc", set_profile_scc},
    {"profile_sn", set_profile_sn},
    {"psn", set_psn},

    {"quiet_mode", set_quiet_mode},

    {"read_coralrc", set_read_coralrc},
    {"recursion_depth", set_recursion_depth},
    {"relation_type", set_relation_type},
    {"return_unify", set_return_unify},
    {"rewriting", set_rewriting},

    {"save_module", set_save_module},
    {"single_answer", set_single_answer},
    {"single_scc", set_single_scc},
    {"sup_magic", set_sup_magic},
    {"sup_magic_indexing", set_sup_magic_indexing},

    {"use_quotes", set_use_quotes},
    {"verbose_answer", set_verbose_answer}

  // The variables must be listed in ALPHABETICAL Order
};

//**************************************************************************
//  AssignValueToGlobal() -
//  This function is used to assign values to global variables that are
//  accessed using the 'set','clear' and 'assign' builtin debug relations.
//
//  This function should really be using a binary search, but there arent
//  really that many values yet, and this is not really a common operation.
//
//**************************************************************************

void corAssignValueToGlobal(corArg *arg, int value)
{
    int index, cmp = -1;
    char *sym = corSymbolString((corSymbol *) arg);

    for (index=0; ((cmp=STRCMP(sym,corGlobalVarTable[index].var_name))>0)
                && (index < COR_GLOBAL_VAR_TABLE_SIZE); index++)
        ;    // This for intentionally left blank.

    if (!cmp){

        int tmp = (corGlobalVarTable[index].func_ptr)(value);

        if (!corexEnv->C_quiet_mode_default)
            fprintf(corexEnv->print_file, "Value of %s is %d\n", sym, tmp);

    } else {
        CORAL_error(COR_BAD_GLOBAL_VAR, NULL, "AssignValueToGlobal");
    }
}

//**************************************************************************
void corAssignValueToGlobal(corArg *arg, double value)
{
    int index;
    char *sym = corSymbolString((corSymbol *) arg);

    int cmp = -1;

    for (index=0; ((cmp=STRCMP(sym,corGlobalVarTable[index].var_name))>0)
                && (index < COR_GLOBAL_VAR_TABLE_SIZE); index++)
        ;    // This for intentionally left blank.

    if (!cmp){

        double tmp = ((PFD) corGlobalVarTable[index].func_ptr)(value);

        if (!corexEnv->C_quiet_mode_default)
            fprintf(corexEnv->print_file, "Value of %s is %f\n", sym, tmp);

    } else {
        CORAL_error(COR_BAD_GLOBAL_VAR, NULL, "AssignValueToGlobal");
    }
}

//**************************************************************************
