/*****************************************************************************/
//    Operators.C
/*****************************************************************************/

#include <stdio.h>

#include "Operators.h"

#include "Globals.h"
#include "Interface.h"
#include "Tuple.h"
#include "TupleIterator.h"
#include "VersionedBindEnv.h"
#include "CoralError.h"

/*****************************************************************************/
corTuple* corFuncOp(corTuple* query)
{
    if (query->arity() != 3) {
        fprintf(corexEnv->error_file, "Three arguments required for Functor.\n");
        return NULL;
    }

    corCArg arg0 = (query->args())[0];
    corCArg arg1 = (query->args())[1];
    corCArg arg2 = (query->args())[2];

    if ( (arg0.kindof() != COR_VARIABLE) && (arg0.kindof() != COR_FUNCTOR)
        && (arg0.kindof() != COR_SYMBOL_CONST) )
    {
        fprintf(corexEnv->error_file, "First argument should be a variable or functor\n");
        return NULL;
    }
    else if ((arg0.kindof() == COR_FUNCTOR)
        || (arg0.kindof() == COR_SYMBOL_CONST))
    {
        Name func_name;
        corCArg num;

        if (arg0.kindof() == COR_SYMBOL_CONST) {
            // char* _name = ((Name)arg0)->string();
            // func_name = AllocSymbol(_name, -1);
            func_name = arg0.SymbolP();
            num = 0;
        }
        else {
            // func_name = AllocSymbol(functor_name(arg0), -1);
            func_name = arg0.FuncArgP()->functor();
            corArgList* func_list = functor_args(arg0.FuncArgP());
            num = func_list->arity();
        }
        corArgList* alist = new corArgList(3);
        (*alist)[0] = arg0;
        (*alist)[1] = func_name;
        (*alist)[2] = num;

        corTuple* result = new corTuple(alist);
        return result;
    }
    else if ((arg0.kindof() == COR_VARIABLE)
        && (arg1.kindof() == COR_SYMBOL_CONST)
        && (arg2.kindof() == COR_NUM_CONST))
    {
        if (!arg2.is_int()) {
            fprintf(corexEnv->error_file,
                "Third argument should be of integer type.\n");
            return NULL;
        }

        corFuncArg * new_arg = corFuncArg::New(arg1.SymbolP(), make_int(arg2));
        int i;
        int index = make_int(arg2);
        for (i = 0;i < index; i++)
        new_arg->args[i] = i;
   
        corArgList* alist = new corArgList(3);
        (*alist)[0] = new_arg;
        (*alist)[1] = arg1;
        (*alist)[2] = arg2;

        corTuple * result = new corTuple(alist);
        result->bindenv = new corArrayBindEnv(index, -1);
        return result;
    }

    fprintf(corexEnv->error_file, "Wrong argument types supplied to Functor\n");
    return NULL;
}

/*****************************************************************************/
corTuple * ArgOp(corTuple * query)
{
    if (query->arity() != 3) {
        fprintf(corexEnv->error_file, "Three arguments required for Arg.\n");
        return NULL;
    }

    corCArg arg0 = (query->args())[0];
    corCArg arg1 = (query->args())[1];
    corCArg arg2 = (query->args())[2];

    if ((arg0.kindof() == COR_NUM_CONST) && (arg1.kindof() == COR_FUNCTOR)) {
        if (!arg0.is_int()) {
                fprintf(corexEnv->error_file,
                    "First argument should be of integer type.\n");
                return NULL;
        }
        corArgList * func_list = functor_args(arg1);
        int num = make_int(arg0);
        if (num > func_list->arity()) {
            fprintf(corexEnv->error_file, "Not enough arguments in arg.\n");
            return NULL;
        }
        corCArg new_arg = func_list->arg(num-1);
        corArgList * alist = make_arglist3(arg0, arg1, new_arg);

        corTuple * result = new corTuple(alist);
        result->bindenv = query->bindenv;
        return result;
    }

    fprintf(corexEnv->error_file, "Wrong argument types supplied to Arg\n");
    return NULL;
}

const corCArg corAddOp(const corCArg arg0,const  corCArg arg1)
{
    if ((arg0.kindof() != COR_NUM_CONST) ||
        (arg1.kindof() != COR_NUM_CONST))
    {
        fprintf(corexEnv->error_file,
            "One of the arguments to AddOp is not a number.\n");
        return corNullCArg;
    }

    if (arg0.is_int() && arg1.is_int())
    {
        return arg0.NumIval() + arg1.NumIval();
    }

    return arg0.NumDval() + arg1.NumDval();
}

/*****************************************************************************/
const corCArg corSubOp(const corCArg arg0, const corCArg arg1)
{
    if ((arg0.kindof() != COR_NUM_CONST) || (arg1.kindof() != COR_NUM_CONST))
    {
        fprintf(corexEnv->error_file,
            "One of the arguments to SubOp is not a number.\n");
        return corNullCArg;
    }

    if (arg0.is_int() && arg1.is_int())
    {
        return arg0.NumIval() - arg1.NumIval();
    }

    return arg0.NumDval() - arg1.NumDval();

}

/*****************************************************************************/
const corCArg  corMulOp(const corCArg arg0,const corCArg arg1)
{
    if ((arg0.kindof() != COR_NUM_CONST) ||
        (arg1.kindof() != COR_NUM_CONST))
    {
        fprintf(corexEnv->error_file,
            "One of the arguments to MulOp is not a number.\n");
        return corNullCArg;
    }
    if (arg0.is_int() && arg1.is_int())
    {
        return arg0.NumIval() * arg1.NumIval();
    }

    return arg0.NumDval() * arg1.NumDval();

}

/*****************************************************************************/
const corCArg corDivOp(const corCArg arg0,const corCArg arg1)
{
    if ((arg0.kindof() != COR_NUM_CONST) ||
        (arg1.kindof() != COR_NUM_CONST))
    {
        fprintf(corexEnv->error_file,
            "One of the arguments to DivOp is not a number.\n");
        return corNullCArg;
    }
    if (!arg1.compare_num(corCArg(0))) 
    {
        fprintf(corexEnv->error_file,
            "Divided by Zero!\n");
        return corNullCArg;
    }
    if (arg0.is_int() && arg1.is_int())
    {
        return arg0.NumIval() / arg1.NumIval();
    }

    return arg0.NumDval() / arg1.NumDval();

}

/*****************************************************************************/
const corCArg corSumOp(corRelation * rel)
{
    int total_i = 0;
    double total_d = 0.0;
    int isInt = 0;
    int isDoub = 0;

    corCArg tmp_carg;

    corStackMark stack_mark0;
    corArrayBindEnv * env0 = new corArrayBindEnv(rel->arity());
    corArgList * args = cormake_vararglist(rel->arity());
    corTupleIterator * iter = new corTupleIterator(rel, *args, env0);
    for(;;) {
        corTuple * ret_tuple = iter->get_next_tuple();
        
        if (iter->no_match())
            break;
        
        if ((ret_tuple->arity() != 1)
            || ( (ret_tuple->arg(0)).kindof() != COR_NUM_CONST))
        {
            fprintf(corexEnv->error_file,
                "Error: Sum: Set must contain numeric constants as elements.\n"
                );
            return(tmp_carg);
        }
        
        if (ret_tuple->arg(0).is_int()) {
            isInt = 1;
            total_i += ret_tuple->arg(0).NumIval();
        }
        else if (ret_tuple->arg(0).is_double()) {
            isDoub = 1;
            total_d += ret_tuple->arg(0).NumDval();
        }
        else {
            fprintf(corexEnv->error_file,
              "Error: Sum: Set must contain numeric constants as elements.\n");
            return(tmp_carg);
        }
    }
    stack_mark0.pop_to();
    if (isInt && isDoub) {
        fprintf(corexEnv->error_file,
            "Error: Sum: Set must contain either integers or doubles\n");
        return(tmp_carg);
    }
    if (isInt)  tmp_carg = total_i;
    if (isDoub) tmp_carg = (total_d);

    return(tmp_carg);
}

/*****************************************************************************/
const corCArg corProdOp(corRelation * rel)
{
    int total_i = 1;
    double total_d = 1.0;
    int is_int = 0;
    int is_doub = 0;
    corCArg tmp_carg;

    corStackMark stack_mark0;
    corArrayBindEnv * env0 = new corArrayBindEnv(rel->arity());
    corArgList * args = cormake_vararglist(rel->arity());
    corTupleIterator * iter = new corTupleIterator(rel, *args, env0);
    for(;;) {
        corTuple * ret_tuple = iter->get_next_tuple();
        
        if (iter->no_match())
            break;
        
        if ((ret_tuple->arity() != 1)
            || ( (ret_tuple->arg(0)).kindof() != COR_NUM_CONST))
        {
            fprintf(corexEnv->error_file,
              "Error: Prod: Set must contain numeric constants as elements.\n");
            return(tmp_carg);
        }
        
        if (ret_tuple->arg(0).is_int()) {
            is_int = 1;
            total_i *= ret_tuple->arg(0).NumIval();
        }
        else if (ret_tuple->arg(0).is_double()) {
            is_doub = 1;
            total_d *= ret_tuple->arg(0).NumDval();
        }
        else {
            fprintf(corexEnv->error_file,
              "Error: Sum: Set must contain numeric constants as elements.\n");
            return(tmp_carg);
        }
    }
    stack_mark0.pop_to();
    if (is_int && is_doub) {
        fprintf(corexEnv->error_file,
            "Error: Sum: Set must contain either integers or doubles\n");
        return(tmp_carg);
    }
    if (is_int)  tmp_carg = (total_i);
    if (is_doub) tmp_carg = (total_d);

    return(tmp_carg);
}

/*****************************************************************************/
const corCArg corAvgOp(corRelation* rel)
{
    int total_i = 0;
    double total_d = 0.0;
    int is_int = 0;
    int is_doub = 0;
    int count = 0;

    corCArg tmp_carg;
    corStackMark stack_mark0;
    corArrayBindEnv * env0 = new corArrayBindEnv(rel->arity());
    corArgList * args = cormake_vararglist(rel->arity());
    corTupleIterator * iter = new corTupleIterator(rel, *args, env0);
    for(;;) {
        corTuple * ret_tuple = iter->get_next_tuple();
        
        if (iter->no_match())
            break;
        
        if ((ret_tuple->arity() != 1)
            || ( (ret_tuple->arg(0)).kindof() != COR_NUM_CONST))
        {
            fprintf(corexEnv->error_file,
              "Error: Sum: Set must contain numeric constants as elements.\n");
            return(tmp_carg);
        }
        
        count++;
        if (ret_tuple->arg(0).is_int()) {
            is_int = 1;
            total_i *= ret_tuple->arg(0).NumIval();
        }
        else if (ret_tuple->arg(0).is_double()) {
            is_doub = 1;
            total_d *= ret_tuple->arg(0).NumDval();
        }
        else {
            fprintf(corexEnv->error_file,
              "Error: Sum: Set must contain numeric constants as elements.\n");
            return(tmp_carg);
        }
    }
    stack_mark0.pop_to();
    if (is_int && is_doub) {
        fprintf(corexEnv->error_file,
            "Error: Sum: Set must contain either integers or doubles\n");
        return(tmp_carg);
    }
    if (is_int)  tmp_carg = (total_i);
    if (is_doub) tmp_carg = (total_d);

    return(tmp_carg);
}

/*****************************************************************************/
const corCArg corCountOp(corRelation* rel)
{
    corCArg tmp(rel->count);
    return(tmp);
}

/*****************************************************************************/
const corCArg corMinOp(corRelation* rel)
{
    int first_tuple_flag = 1;
    corTuple * ret_tuple;
    corCArg min;
    corStackMark stack_mark0;
    corArrayBindEnv * env0 = new corArrayBindEnv(rel->arity());
    corArgList * args = cormake_vararglist(rel->arity());
    corTupleIterator * iter = new corTupleIterator(rel, *args, env0);

    for(;;) {
        if (first_tuple_flag) {
            ret_tuple = iter->get_next_tuple();
            if (iter->no_match()) {
                fprintf(corexEnv->error_file,
                    "Error: Min: No elements in the set.\n");
                return corNullCArg;
            }
            min = ret_tuple->arg(0);
            first_tuple_flag = 0;
        }
        else
            ret_tuple = iter->get_next_tuple();

        if (iter->no_match())
            break;

        if ((ret_tuple->arity() != 1)
            || ( (ret_tuple->arg(0)).kindof() != COR_NUM_CONST))
        {
            fprintf(corexEnv->error_file,
              "Error: Min: Set must contain numeric constants as elements.\n");
            return corNullCArg;
        }

        if (!first_tuple_flag) {
            corTerm *num0 = new corTerm(min) ;
            corTerm *num1 = new corTerm(ret_tuple->arg(0));
            if (!corLssOp(*num0, *num1))
                min = ret_tuple->arg(0);
        }
    }
    stack_mark0.pop_to();
    return min;
}

/*****************************************************************************/
const corCArg corMaxOp(corRelation* rel)
{
    int first_tuple_flag = 1;
    corTuple * ret_tuple;
    corCArg max;
    corStackMark stack_mark0;
    corArrayBindEnv * env0 = new corArrayBindEnv(rel->arity());
    corArgList * args = cormake_vararglist(rel->arity());
    corTupleIterator * iter = new corTupleIterator(rel, *args, env0);

    for(;;) {
        if (first_tuple_flag) {
            ret_tuple = iter->get_next_tuple();
            if (iter->no_match()) {
                fprintf(corexEnv->error_file,
                    "Error: Max: No elements in the set.\n");
                return corNullCArg;
            }
            max = ret_tuple->arg(0);
            first_tuple_flag = 0;
        }
        else
            ret_tuple = iter->get_next_tuple();

        if (iter->no_match())
            break;

        if ( (ret_tuple->arity() != 1)
            || ( (ret_tuple->arg(0)).kindof() != COR_NUM_CONST))
        {
            fprintf(corexEnv->error_file,
              "Error: Max: Set must contain numeric constants as elements.\n");
            return corNullCArg;
        }

        if (!first_tuple_flag) {
            corTerm *num0 = new corTerm(max) ;
            corTerm *num1 = new corTerm(ret_tuple->arg(0));
            if (!corGrtOp(*num0, *num1))
                max = ret_tuple->arg(0);
        }
    }
    stack_mark0.pop_to();
    return max;
}

/*****************************************************************************/
int corEquOp(corTerm& term0, corTerm& term1)
{
    return corunify_args(term0, term1);
}

/*****************************************************************************/
int corNeqOp(corTerm& term0, corTerm& term1)
{
    return !corunify_args(term0, term1);
}

/*****************************************************************************/
int corLssOp(corTerm& term0, corTerm& term1)
{
    return (term0.expr.compare_num(term1.expr) < 0);
}

/*****************************************************************************/
int corGrtOp(corTerm& term0, corTerm& term1)
{
    return (term0.expr.compare_num(term1.expr) > 0);
}

/*****************************************************************************/
int corLeqOp(corTerm& term0, corTerm& term1)
{
    return (term0.expr.compare_num(term1.expr) <= 0);
}

/*****************************************************************************/
int corGeqOp(corTerm& term0, corTerm& term1)
{
    return (term0.expr.compare_num(term1.expr) >= 0);
}

/*****************************************************************************/
int corCopyTermOp(corTerm& term0, corTerm& term1) 
{
    corTerm newterm;
    int env_size = corCopyArg( newterm.expr, term0.expr, term0.bindenv) ;
    if (env_size > 0) newterm.bindenv = new corVersionedBindEnv(env_size);
    return corunify_args(term1, newterm);
}

/*****************************************************************************/
const corCArg corLShiftOp(const corCArg arg0, const corCArg arg1)
{

    if (arg0.is_int() && arg1.is_int())
        return arg0.NumIval() << arg1.NumIval();
    else{
        fprintf(corexEnv->error_file,
            "One of the arguments to LShiftOp is not an integer.\n");
        return corNullCArg;
    }

}

/*****************************************************************************/
const corCArg corRShiftOp(const corCArg arg0, const corCArg arg1)
{
    if (arg0.is_int() && arg1.is_int())
        return arg0.NumIval() >> arg1.NumIval();
    else{
        fprintf(corexEnv->error_file,
            "One of the arguments to LShiftOp is not an integer.\n");
        return corNullCArg;
    }
}

/*****************************************************************************/
const corCArg corBitorOp(const corCArg arg0, const corCArg arg1)
{
    if (arg0.is_int() && arg1.is_int())
        return arg0.NumIval() | arg1.NumIval();
    else{
        fprintf(corexEnv->error_file,
            "One of the arguments to LShiftOp is not an integer.\n");
        return corNullCArg;
    }
}

/*****************************************************************************/
const corCArg corBitandOp(const corCArg arg0, const corCArg arg1)
{
    if (arg0.is_int() && arg1.is_int())
        return arg0.NumIval() & arg1.NumIval();
    else{
        fprintf(corexEnv->error_file,
            "One of the arguments to LShiftOp is not an integer.\n");
        return corNullCArg;
    }
}

/*****************************************************************************/
const corCArg corBitxorOp(const corCArg arg0, const corCArg arg1)
{
    if (arg0.is_int() && arg1.is_int())
        return arg0.NumIval() ^ arg1.NumIval();
    else{
        fprintf(corexEnv->error_file,
            "One of the arguments to LShiftOp is not an integer.\n");
        return corNullCArg;
    }
}

/*****************************************************************************/
const corCArg corModOp(const corCArg arg0, const corCArg arg1)
{
    if (arg0.is_int() && arg1.is_int())
        return arg0.NumIval() % arg1.NumIval();
    else{
        fprintf(corexEnv->error_file,
            "One of the arguments to LShiftOp is not an integer.\n");
        return corNullCArg;
    }
}

/*****************************************************************************/

static GoalIdCount = 0;

corBindEnv* corGoalIdOp(corBuiltinRelation& , corTupleIterator& iter)
{
    corStackMark stack_mark0;

    if (iter.arg_list.arity() != 1) {
        // should not happen
        CORAL_error(COR_BAD_ARITY, "of GoalID", "GoalIdOp");
        stack_mark0.pop_to();
        iter.set_no_match();
        return NULL;
    }

    corTerm term(iter.arg_list.arg(0), iter.bindenv);
    corfull_dereference_term(term);

    corCArg query_arg = term.expr;
    corBindEnv *query_env = term.bindenv;

    if (!is_functor(query_arg)) {
        CORAL_error(COR_BAD_ARG_TYPE, "arg of GoalID", "GoalIdOp") ;
        stack_mark0.pop_to();
        iter.set_no_match();
        return NULL;
    }

    // find out all info about the magic predicate in GoalId
    Name mpred_name = query_arg.FuncArgP()->functor();
    corArgList *mpred_args = functor_args(query_arg);
    int mpred_arity = mpred_args->arity();

    // check if id is indeed a variable
    corCArg mpred_id = mpred_args->arg(0); 
    if (!is_var(mpred_id)) {
        // should not happen
        fprintf(corexEnv->error_file, "Id of GoalId should not be bound\n");
        stack_mark0.pop_to();
        iter.set_no_match();
        return NULL;
    }

    // we need to access the relation off the frame because of recursive calls
    corRelation *mreln = find_relation(mpred_name,mpred_arity); 
    if (!mreln) mreln = cormake_relation(mpred_name, mpred_arity);

#    ifdef MAKE_SUPP_INDEX
        if (mreln->for_supp_index == NULL) {
            // make a new relation and add all appropriate indices to it
            mreln->for_supp_index = new corHashSimpleRelation(mreln->arity());
            mreln->for_supp_index->check_subsum = mreln->check_subsum;
            mreln->add_all_indices(mreln->for_supp_index);
            // mreln->for_supp_index inherits all indices from mreln
        }
#    endif

    if (mreln->check_subsum==1) {
        // set up a tuple iterator on relation and find matching tuple
        corStackMark stack_mark1;
        
        // searching main relation
        corTupleIterator gid_iter(mreln, *mpred_args, query_env);
    
        for (;;) {
            corStackMark stack_mark2;
            corTuple *tuple1 = gid_iter.get_next_tuple();
            stack_mark2.pop_to(); // undo bindings to query_env
            if (tuple1 == NULL) break; // search for_supp_index now

            // found a matching tuple; check for subsumption

            corCArg tuple1_id = tuple1->arg(0);

            // this will work only if the tuple in the relation has
            // no bindings.

            corTerm term;
            term.expr = tuple1_id;
            corunify_binding(query_env, mpred_id.VarArgP()->var, term);
            corArrayBindEnv tuple1_env(tuple1->env_size);
            if (corsubsumes_arg_lists(tuple1->args(), *mpred_args,
                &tuple1_env, query_env))
            {
                // found the right id value too
                stack_mark1.pop_to(); // undo bindings because of subsumes
                corunify_binding(query_env, mpred_id.VarArgP()->var, term);
                    // get back the id into the BindEnv
                iter.reset_no_match();
                return query_env;
            }
            stack_mark2.pop_to();
        }

        // now searching for_supp_index relation
        stack_mark1.pop_to();

#        ifdef MAKE_SUPP_INDEX
            corTupleIterator gid_iter1(mreln->for_supp_index, *mpred_args,
                query_env);
            for (;;) { 
                corStackMark stack_mark3;
                corTuple *tuple2 = gid_iter1.get_next_tuple();
                stack_mark3.pop_to(); // undo bindings
                if (tuple2 == NULL) break; // need to get brand new id value 

                // found a matching tuple; check for subsumption

                corArg *tuple2_id = tuple2->arg(0);

                // this will work only if the tuple in the relation has
                // no bindings.

                //    why is this part necessary ? isnt the fact that the
                //    iterator returned a tuple sufficient to show that
                //    there is already a subsuming tuple  -- PRAVEEN ???

                corTerm term;
                term.expr = tuple2_id;
                corunify_binding(query_env, mpred_id.VarArgP()->var, term);
                corArrayBindEnv tuple2_env(tuple2->env_size);
                if (corsubsumes_arg_lists(tuple2->args(), *mpred_args,
                    &tuple2_env, query_env))
                {
                    // found the right id value too
                    stack_mark3.pop_to(); // undo bindings because of subsumes
                    corunify_binding(query_env, mpred_id.VarArgP()->var, term);
                        // get back the id into the BindEnv
                    iter.reset_no_match();
                    return query_env;
                }
                stack_mark3.pop_to();
            }
#        endif
    }

    // else can immediately insert it in mreln->for_supp_index with NEW id
    corTerm term1;
    corCArg new_id = (GoalIdCount++);
    term1.expr = new_id;
    corunify_binding(query_env, mpred_id.VarArgP()->var, term1);

    //    WARNING::  ReturnUnify optimizations needed here!!!

    /*
        If #define INSERT_GOALID is included, the magic relation gets
        inserted into, without a parent goal being added to the magic fact.
        This results in the return_unify optimization getting screwed up,
        since that expects that every magic fact has a pointer to a
        supplementary magic fact as a parent goal. : Praveen.
    */

#    ifdef INSERT_GOALID
#        ifdef MAKE_SUPP_INDEX
            mreln->for_supp_index->insert_new(*mpred_args, query_env);
#        else
            mreln->insert_new(*mpred_args, query_env);
#        endif
#    endif

    iter.reset_no_match();
    return query_env;
}

/*****************************************************************************/

