#include "CRdbManager.h"

#ifdef DO_RDB

#include "CRdbSolvers.h"
#include "SymbolTable.h"
#include "BuiltinRelation.h"
#include "Log.h"
#include "Rdb.h"



//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

// Global Coral-Rdb handler -- set by coral_init, destroyed by exit_coral
corCRdbManager* pCRdbManager = NULL;


//---------------------------------------------------------------------------

corCRdbManager::corCRdbManager()
{
    LOG( "Allocating rdb builtins" << endl );
    corBuiltinRelation* temp;
    temp =
      new corBuiltinRelation(-1, corEnterSymbol("rdb_open_db"),  RdbOpenDBSolver);
    temp =
      new corBuiltinRelation(-1, corEnterSymbol("rdb_map"),      RdbMapSolver);
    temp =
      new corBuiltinRelation(-1, corEnterSymbol("rdb_join"),     RdbJoinSolver);
    temp =
      new corBuiltinRelation(-1, corEnterSymbol("rdb_commit"),   RdbCommitSolver);
    temp =
      new corBuiltinRelation(-1, corEnterSymbol("rdb_rollback"), RdbRollbackSolver);
    temp = 
      new corBuiltinRelation(-1, corEnterSymbol("rdb_execute"),  RdbExecuteSolver);
    temp = 
      new corBuiltinRelation(-1, corEnterSymbol("rdb_logging"),  RdbLoggingSolver);
}


//---------------------------------------------------------------------------

corCRdbManager::~corCRdbManager()
{
    // remove any crdb relations still hanging around, before deleting db's
//     for( int i = crelList.Size() -1 ; i >= 0 ; i-- ) {
// 	delete crelList[i];	// warning: this will call us back
// 				// to remove itself from the list
//     }
//     assert(crelList.Size() == 0 );
}


//---------------------------------------------------------------------------

corCRdbManager::corDBEntry::corDBEntry(const String newName, RdbDatabase* newDb)
{
    nameList.Append(newName);
    assert(newDb);
    db = newDb;
}


//---------------------------------------------------------------------------

corCRdbManager::corDBEntry::~corDBEntry()
{
    delete db;
}


//---------------------------------------------------------------------------

RdbDatabase* corCRdbManager::LookupDatabase(const String name)
{
    int size = dbList.Size();
    for( int i = 0 ; i < size ; i++ ) {
	corDBEntry& dbe = *dbList[i];
	int numNames = dbe.nameList.Size();
	for( int j = 0 ; j < numNames ; j++ ) {
	    if( name == dbe.nameList[j] )
	      return dbe.db;
	}
    }
    return NULL;
}


//---------------------------------------------------------------------------

void corCRdbManager::StoreDatabase(const String name,
				RdbDatabase* db)
{
    int size = dbList.Size();
    for( int i = 0 ; i < size ; i++ ) {	// check for duplicates
	corDBEntry& dbe = *dbList[i];
	if( dbe.db == db ) {
	    dbe.nameList.Append(name);
	    return;
	}
    }
    // db has not been opened under another name.
    dbList.Append( new corDBEntry(name, db) );
}


// inline void CRdbManager::StoreCRdbRelation(CRdbRelation* crel)
// {
//     crelList.Append(crel);
// }



//---------------------------------------------------------------------------

// void CRdbManager::RemoveCRdbRelation(CRdbRelation* crel)
// {
//     int s = crelList.Size();
//     for( int i = 0 ; i < s ; i++ ) {
// 	if( crelList[i] == crel )
// 	  crelList.Delete(i);
//     }
// }

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#endif // DO_RDB
