/*****************************************************************************/
//    PersistentIndex.C
/*****************************************************************************/

#ifdef WITH_PERSISTENCE


#include "PersistentRelation.h"


#include "Globals.h"
#include "CoralError.h"
#include "SMUtil.h"
#include "sm_client.h"
#include "Config.h"
#include "Schema.h"
#include "IndexStruct.h"
#include "BPlusTreeIndex.h"
#include "FileDescStruct.h"
#include "BufferGroup.h"
#include "Transaction.h"
#include "Catalog.h"
#include "StorageManager.h"


#include "PersistentIndex.h"

/*****************************************************************************/

extern int corunify_literal(corTupleIterator&, corTuple *, int) ;
extern int SM_initialized;

/*****************************************************************************/

#define DELETE_INDEX_ENTRY 1
#define DELETE_OBJ_AND_INDEX_ENTRY 2

// temporary hack to avoid changing generic-index.h
// also see persistent-index.C  :: PRAVEEN
#define COR_I_PARTIAL_SUCCESS  2

/*****************************************************************************/
/*
    Constructor that initializes the persistent-index structure from the
    catalog if it exists in the catalog.  Otherwise, an error is raised.

    Note that the file names are only guarenteed to be unique upto
    COR_MAX_FILE_NAME_LEN
*/

corPersistentIndex::corPersistentIndex(corPersistentRel *rel, corBitVector *bv)
    : corGenericIndex((corStorageRelation *)rel, bv)
{
}

/*****************************************************************************/
corTuple* corPersistentIndex::get_next_tuple (corTupleIterator& iterator)
{
  if (!SM_initialized) return NULL;

  corTuple *tuple;

  return p_rel()->indexScanGetNext(iterator, chosen_index);

}

/*****************************************************************************/
corBindEnv* corPersistentIndex::get_next(corTupleIterator& iterator)
{
    if (!SM_initialized) return NULL;

    corTuple* tuple = get_next_tuple(iterator);
    if (iterator.no_match())
        return NULL;
    else
        return iterator.bindenv;
}

/*****************************************************************************/

#endif WITH_PERSISTENCE

