
%{
#include <stdio.h>

#include "Globals.h"
#include "Scanner.h"
#include "Profile.h"
#include "Annotations.h"
#include "Literal.h"
#include "Interface.h"
#include "ParseActions.h"
#include "CoralError.h"
#include "Execute.h"
#include "Database.h"
#include "Sips.h"
#include "Rule.h"

#include "ParseUtils.h"
#include "BuiltinSymbols.h"
#include "ExprArg.h"
#include "NumArg.h"
#include "VarArg.h"
#include "FuncArg.h"
#include "DontCareArg.h"
#include "ArgList.h"
#include "CArg.h"
#include "ClientProtocol.h"

#define NilArg corNilSymbol	/* symbol arg "nil" */

void yyerror(char *s);

EXTERN int yylex(YYSTYPE *lvalp);
EXTERN int yyparse();

// No. of evaluable functors in the lhs of a clause
// that were converted into literals by create_functor.
// (We want to evaluate these *after* the true rhs literals.)
static char *fname = "yyparse" ;

extern void COR_set_single_answer(corModuleInfo& Module, int val);

//DontCareArg TheParameterArg;
corArg *corTheParameterArg;
corCArg* cortmpCArg;

    // Redefine the parser to use this "lexer" so that we can switch
    // scanners (aka lexers) depending who is one the top of the stack.
    // The function will actually be called as:
    //     (parseEnv->lexer)(&yylval)
    // assuming YYPURE is defined and YYLSP_NEEDED is not.
    // This is done here to get around bison.simple's definition of YYLEX.
#undef    yylex
#define   yylex    (corparseEnv->lexer)


corSymbol ** corbuiltin_ops[] = {
    // Aggregation operators: These must be FIRST, and in the same
    // ORDER as in the enum AggregateKind (defined in Arg.h). 
    &corSumOpSymbol,
    &corProdOpSymbol,
    &corCountOpSymbol,
    &corMinOpSymbol,
    &corMaxOpSymbol,
    &corAvgOpSymbol,
    &corSDstOpSymbol,
    &corPDstOpSymbol,
    &corCDstOpSymbol,
    &corADstOpSymbol,
    &corDstOpSymbol,
    &corAnyOpSymbol,

    /** Normal (not aggregation) operators: **/
    /** Builtin_Arith_Start = 12 **/
    &corLssOpSymbol,
    &corGrtOpSymbol,
    &corLeqOpSymbol,
    &corGeqOpSymbol,
    /** Builtin_Convertible_Start = 16 **/
    &corAddOpSymbol,
    &corSubOpSymbol,
    &corMulOpSymbol,
    &corDivOpSymbol,
    &corIDivOpSymbol,

    &corLShiftOpSymbol,
    &corRShiftOpSymbol,
    &corBitxorOpSymbol,
    &corBitandOpSymbol,
    &corBitorOpSymbol,
    &corBitnotOpSymbol,
    &corModOpSymbol,

    0
};

struct corSymbolTable * corBuiltinRelTable ;
int corBuiltin_Arith_Start = 12 ;
int corBuiltin_Convertible_Start = 16 ;

%}

// This next line makes the parser re-entrant.
%pure_parser

%type   <pred>   pred clause_lhs clause_left clause_lhs_list clause_lhs_element
%type   <carg>   arg arg_fact arg_primary var_arg simple_arg
%type   <sip>    arc sip
%type   <sips>   sip_list sip_list_entry
%type   <carg>   id const anno_name adorn_string opt_adorn_string
%type   <NameArg>    relop
%type   <alias_list> literal_alias
%type   <rule_anno> rule_anno rule_anno_list opt_annotes sips_or_rule_anno
%type   <anno_case> anno_case anno_case_list
%type   <anno_case_name> anno_case_name
%type   <anno_case_details> anno_case_details
%type   <integer> plus_or_minus
%token  APPEND UPDATE ASSIGN DELETE
%token  MARK LPAREN RPAREN COMMA COLON SLASHSLASH PERIOD LBRACE RBRACE ARROW 
%token  SIPS LBRACKET RBRACKET UNDERSCORE VBAR PERCENT NOT QUESTION DOLLAR TRUE
%token  SUM PROD COUNT MIN MAX AVG ANY DISTINCT COUNT_D SUM_D PROD_D AVG_D
%token  DEL START_C_MODULE END_C_MODULE MODULE EXPORT ALIAS END_MODULE 
%token  PLUS MINUS STAR SLASH EQUAL NOTEQUAL LESS GREATER LESSEQUAL GREATEREQUAL
%token  LSHIFT RSHIFT BITXOR BITAND BITNOT BITOR MOD
%token  AT CUT
%token  <carg> SYMBOL VAR OP RELOP 
%token  <carg> NUM
%left   BITOR
%left   BITXOR
%left   BITAND
%left   LSHIFT RSHIFT
%left   PLUS MINUS
%left   STAR SLASH SLASHSLASH MOD
%right  BITNOT

%%
prog:
        {
            corexEnv->C_interrupt_raised = 0;
            corparserStruct.init() ;  
            corparserStruct.rules->chain = NULL;
            if (!coralConfig.server_mode)
                corparseEnv->print_prompt();
        }
        clause_list
        {
            corparserStruct.rule_list = (corClause*) corparserStruct.rules->chain;
        }
;

clause_list:    clause_list
        {
            corparserStruct.rule_var_names.length = 0;
            corparserStruct.rhspreds.length = 0;
            corparserStruct.nested_index = 0;
            if (corexEnv->C_at_end_module) {
                      // IS THIS NECESSARY ?? - PRAVEEN
                corparserStruct.init();
                corinit_collection(corparserStruct.rules, 0);
            }
        }
        clause
        {
            if (corexEnv->C_interrupt_raised != 0) {
                if (corexEnv->C_interrupt_raised == 2) {

                    /* Special Handling Interrupt */
                    /* used for quit() and yield() */
                    /* which require returning from yyparse() */

                    if (corexEnv->quit){
                          /* quit(). */

                          /* Inform the client of quit() if in server mode. */
                        if (coralConfig.server_mode)
                            fprintf(corexEnv->print_file,"%s\n",CP_EXIT);

                        corparseStack->abort_file();
                        while (! corparseStack->is_empty())
                            corparseStack->pop();

                    } else if (corexEnv->yield){
                          /* yield(). */
                        corexEnv->yield = 0;
                    }
                    corexEnv->C_interrupt_raised = 0;

                      /* Return from yyparse() */
                    YYACCEPT;
                }

                if (corexEnv->C_interrupt_raised > 0) {
                   fprintf(corexEnv->error_file,
                            "CORAL:: Interrupt occurred during execution!:");
                   fprintf(corexEnv->error_file,"resetting state !\n");
                   fflush(corexEnv->error_file);
                    //abort a consult that is halfway thru
                   corparseStack->abort_file();
                }
                   // else, the user does not want to see more answers
                   // to the query, and the value is -1
                corexEnv->C_interrupt_raised = 0;
            }

            if (corexEnv->C_do_history_default) corparseEnv->handle_history();

            if (corparseEnv->show_prompt) corparseEnv->print_prompt();
#ifdef DEBUG
            if (corT_Stack.count) {
                fprintf(corexEnv->error_file,
                    "Warning : non-zero tablestack = %d!\n", corT_Stack.count);
                corT_Stack.count = 0;
            }
#endif
        }
    |
;
clause_lhs:    id COLON clause_lhs
        {

            struct corClauseName *clname = new corClauseName();
            clname->name = ($1)->SymbolP();
            clname->negated = 0;
            clname->next = corparseEnv->cur_rule->clause_names;
            corparseEnv->cur_rule->clause_names = clname;
	    delete $1;
        }
    |    id MINUS COLON clause_lhs
        {
            struct corClauseName *clname = new corClauseName();
            clname->name = ($1)->SymbolP();
            clname->negated = 1;
            clname->next = corparseEnv->cur_rule->clause_names;
            corparseEnv->cur_rule->clause_names = clname;
	    delete $1;
        }
    |     clause_lhs_list
        {
        }
;

clause_lhs_list: clause_lhs_list COMMA clause_lhs_element
        {
        }
    |       clause_lhs_element
        {
        }
;

clause_lhs_element: const LPAREN  optarglist RPAREN
        {
            corparseEnv->cur_rule->head = new corLiteral(corparserStruct, $1,
                            corparserStruct.cur_arg_chain->count);
            if (corparseEnv->complexRelation)
                corparseEnv->complexRelation--;
            corpop_arg_chain(corparserStruct, corparseEnv->cur_rule->head->args);
            corparseEnv->cur_rule->clause_names = NULL;
	    delete $1;
        }
    |
        DEL const LPAREN optarglist RPAREN
        {
            corClause *cur_rule = corparseEnv->cur_rule;
            corLiteral *lit = new corLiteral(corparserStruct, $2,
                       corparserStruct.cur_arg_chain->count);
            if (corparseEnv->complexRelation)
                corparseEnv->complexRelation--;
            lit->negated = 1;
            corpop_arg_chain(corparserStruct, lit->args);
            cur_rule->clause_names = NULL;

            if (corexEnv->in_query_loop && cur_rule->head == NULL) {
                cur_rule->head = lit;
            } else {
                if (cur_rule->head_deletes) {
                    cur_rule->head_deletes =
                      (corLiteral **) realloc((char *)cur_rule->head_deletes,
                         sizeof(corLiteral*) *(++cur_rule->head_deletes_count) );
                } else {
                    cur_rule->head_deletes = new (corLiteral*);
                        //(Literal **)malloc(sizeof(Literal*));
                    cur_rule->head_deletes_count++;
                }
                cur_rule->head_deletes[cur_rule->head_deletes_count-1]=lit;
            }
            COR_set_eager_eval(corparserStruct.CurModule, 0);
	    delete $2;
        }
    |
        SYMBOL
        {
            corparseEnv->cur_rule->head = new corLiteral(corparserStruct, $1,0);
            if (corparseEnv->complexRelation)
                corparseEnv->complexRelation--;
            corparseEnv->cur_rule->clause_names = NULL;
	    delete $1;
        }
;
/* export pred1(adorn11, adorn12, ...), pred2(adorn21, ..), ... */
export_clause:    EXPORT arg_fact
        {
            corAddExportInfo(corparserStruct, $2);
	    delete $2;
        }
    |    export_clause COMMA arg_fact
        {
            corAddExportInfo(corparserStruct, $3);
	    delete $3;
        }
    |    EXPORT const LBRACKET optarglist RBRACKET
        {
                  corAddExportInfo(corparserStruct, 
                  (corFuncArg*) corcreate_functor(corparserStruct, $2));
	    delete $2;		
        }
    |    export_clause COMMA const LBRACKET optarglist RBRACKET
        {
                  corAddExportInfo(corparserStruct, 
                (corFuncArg*) (corcreate_functor(corparserStruct, $3)));
	    delete $3;
        }
;

clause_left:
        {
           corparserStruct.lhs_flag = 1; 
           corparseEnv->cur_rule = corcopy_rule(NULL);
           corparseEnv->cur_rule->line_number = corparseEnv->cur_line_num;
        }
             clause_lhs
        {
           corparserStruct.lhs_flag = 0;
           corparseEnv->C_num_head_arith_lits = corparserStruct.rhspreds.length;
        }
;

clause:       clause_left clause_rhs
        {
            corClause *cur_rule = corparseEnv->cur_rule;
            if (corexEnv->in_query_loop) {
                corparserStruct.assign_func(corFactFunction);
                corparserStruct.cur_rule = cur_rule;
                if (corexEnv->C_exec_mode == COR_DOING_IMPERATIVE)
                    YYACCEPT;
                corparserStruct.execute();
                corparserStruct.reset(0);
            } else {

              // assign a label to the rule if it has no label
              if (cur_rule->clause_names == NULL) {
                  char label_str[5];
                  struct corClauseName *clname1 = new corClauseName();

                  sprintf(label_str, "%d", corparserStruct.label_counter); 
                  clname1->name = corEnterSymbol(label_str);
                  clname1->negated = 0;
                  clname1->next = NULL;
                  cur_rule->clause_names = clname1;
                  corparserStruct.label_counter++;
              } 
            
              // initialize the sn_rule_label_counter of the rule
              cur_rule->sn_rule_label_counter = 0;


              /* transfer information from parserStruct to cur_rule
               * and insert cur_rule into the parserStruct's rulelist
               */
              corcomplete_rule(cur_rule, corparserStruct);
            }
        }

    |    clause_left APPEND rhs_pred_list opt_annotes PERIOD
        {
              corparserStruct.assign_func(corAppendFunction);
              corparserStruct.cur_rule = corparseEnv->cur_rule;

              if (corexEnv->C_exec_mode == COR_DOING_IMPERATIVE)
                  YYACCEPT;
              corparserStruct.execute();
              delete corparserStruct.cur_rule; 
              // Better not delete the head predicate.
              corparserStruct.cur_rule = NULL;
        }

    |       clause_left DELETE rhs_pred_list opt_annotes PERIOD
        {
              corparserStruct.assign_func(corDeleteFunction);
              corparserStruct.cur_rule = corparseEnv->cur_rule;
              if (corexEnv->C_exec_mode == COR_DOING_IMPERATIVE)
                  YYACCEPT;
              corparserStruct.execute();
              delete corparserStruct.cur_rule; 

              // Better not delete the head predicate.
              corparserStruct.cur_rule = NULL;
        }

    |    clause_left UPDATE rhs_pred_list opt_annotes PERIOD
        {
          fprintf(corexEnv->error_file, "Sorry : Updates not implemented yet\n");
        }

    |    clause_left ASSIGN rhs_pred_list opt_annotes PERIOD
        {
            corparserStruct.assign_func(corAssignFunction);
            corparserStruct.cur_rule = corparseEnv->cur_rule;
            if (corexEnv->C_exec_mode == COR_DOING_IMPERATIVE)
                YYACCEPT;
            corparserStruct.execute();
            delete corparserStruct.cur_rule; 

            // Better not delete the head predicate.
            corparserStruct.cur_rule = NULL;
        }

    |    QUESTION 
        {
            corparseEnv->cur_rule = corcopy_rule(NULL);
            corparseEnv->cur_rule->line_number = corparseEnv->cur_line_num;
            corparseEnv->cur_rule->clause_names = NULL;
        }
                rhs_pred_list PERIOD
        {
              corparserStruct.assign_func(corQueryFunction);
              corparserStruct.cur_rule = corparseEnv->cur_rule;
              if (corexEnv->C_exec_mode == COR_DOING_IMPERATIVE)
                  YYACCEPT;
              corparserStruct.execute();
              corparserStruct.reset(1);
        }

        |       START_C_MODULE PERIOD
        {
              corexEnv->c_module_separator = 0;
        }

        |       END_C_MODULE PERIOD
        {
              corexEnv->c_module_separator = 1;
              if (corexEnv->C_exec_mode == COR_DOING_IMPERATIVE)
                  YYACCEPT;
        }

    |    MODULE id PERIOD
        {
            corparserStruct.init() ;
            corinit_collection(corparserStruct.rules, 1);
            corexEnv->C_at_end_module = 0;
            corexEnv->in_query_loop = 0 ;
            corparserStruct.CurModule.name = ($2)->SymbolP();
            strcat(corparseEnv->prompt, "*");
	    delete $2;
        }

        |       END_MODULE PERIOD
        {
            corexEnv->C_at_end_module = 1;

            corparserStruct.rule_list = (corClause*)
                      corparserStruct.rules->chain;
            corparserStruct.assign_func(corEndModuleFunction);
            corparserStruct.cur_rule = corparseEnv->cur_rule;

            if (corexEnv->C_exec_mode == COR_DOING_INTERPRET) {
                corparserStruct.execute();
                corinit_collection(corparserStruct.rules, 0);
                // reinitialize CurrModuleInfo 
                corCurrModuleInfo->init_flags();
                corexEnv->in_query_loop = 1 ;
    
                if (!strcmp(corSymbolString(corCurDB->name), "default_db"))
                    corparseEnv->prompt = coralConfig.prompt ;
                else {
                  strcpy(corparseEnv->prompt_name, corSymbolString(corCurDB->name));
                  strcat(corparseEnv->prompt_name, coralConfig.prompt);
                }
                corparseEnv->print_prompt();

            } else if (corexEnv->C_exec_mode == COR_DOING_IMPERATIVE) {
                corexEnv->in_query_loop = 1;
                YYACCEPT;

            } else { 
                // magic or factoring or existential optimization
                YYACCEPT;
            }
        }

    |    export_clause PERIOD

    |    AT SYMBOL SYMBOL opt_adorn_string plus_or_minus PERIOD
        {
            corArgList *al1 = new corArgList(1);
            if (!($4)->IsEmpty())
                (*al1)[0] = *($4);
            else al1->set_count(0);

            if (*($2) == corAllowedAdornSymbol) {
                coradd_allowed_adornment(corparserStruct, $3, al1 );
            }

            // NOTE :: Currently this does not give multiset
            // semantics. It merely acts as do_not_check_subsum.
            // The user should not use this annotation in its 
            // current state.

            else if (*($2) == corMultisetSymbol) {
                coradd_multiset_annotation(corparserStruct, $3, al1, 0);
            }
            else if (*($2) == corCheckSubsumSymbol) {
                coradd_multiset_annotation(corparserStruct, $3, al1, $5);
            }
/***
            else if (*($2) == DiskResidentSymbol) {
                add_diskrel_annotation(parserStruct, $3, al1);
            }
***/
            else if (*($2) == corIndexDeltaSymbol) {
                coradd_index_delta_annotation(corparserStruct, $3, al1 );
            }
            else CORAL_error(COR_BAD_ANNO, corSymbolString($2->SymbolP()),
                     "yyparse:anno3", __FILE__, __LINE__);
	    delete $2; delete $3; delete $4;
        }

    |    AT SYMBOL SYMBOL opt_adorn_string LPAREN optarglist RPAREN LPAREN optarglist RPAREN plus_or_minus PERIOD
        {
            corArgList *al2 = new corArgList(corparserStruct.cur_arg_chain->count);
            corpop_arg_chain(corparserStruct, *al2);
            corArgList *al1 = new corArgList(corparserStruct.cur_arg_chain->count);
            corpop_arg_chain(corparserStruct, *al1);
            if ( *($2) == corMakeIndexSymbol) {
                coradd_make_index_annotation(
                corparserStruct.make_index_annotations, $3, $4, al1, al2);

            } else if ( *($2) == corChoiceSymbol) {
                fprintf(corexEnv->error_file,
                     "ERROR:  choice no longer supported %s\n",
                     " -- use aggregate selection `any' instead");

            } else
                CORAL_error( COR_BAD_ANNO,corSymbolString($2->SymbolP()),"yyparse:anno1",
                    __FILE__, __LINE__);
	    delete $2; delete $3; delete $4;
        }

    |    AT SYMBOL SYMBOL opt_adorn_string LPAREN optarglist RPAREN plus_or_minus PERIOD
        {
            corArgList *al1 = new corArgList(corparserStruct.cur_arg_chain->count);
            corpop_arg_chain(corparserStruct, *al1);
            if ( *($2) == corMakeIndexSymbol) {
              coradd_make_index_annotation(
                corparserStruct.make_index_annotations, $3, $4, al1, NULL);

            } else if (*($2) == corDiskResidentSymbol) {
                coradd_diskrel_annotation(corparserStruct, $3, $4, al1);

            } else CORAL_error(COR_BAD_ANNO, 
			       corSymbolString($2->SymbolP()),"yyparse:anno2", __FILE__, __LINE__);
	    delete $2; delete $3; delete $4;
        }

    |    AT SYMBOL SYMBOL opt_adorn_string LPAREN optarglist RPAREN
                LPAREN optarglist RPAREN simple_arg plus_or_minus PERIOD
        {
            corCArg* op_arg = $11;
            corArgList *al2 = new corArgList(corparserStruct.cur_arg_chain->
                        count);
            corpop_arg_chain(corparserStruct, *al2);

            corArgList *al1 = new corArgList(corparserStruct.cur_arg_chain->
                        count);
            corpop_arg_chain(corparserStruct, *al1);

            /* Should have some error checking to see if 
                predicate name is correct! */

            if ( *($2) == corAggSelectionSymbol) {
            coradd_agg_sel_annotation(corparserStruct,
                    $3, $4, al1, al2, op_arg);
            COR_set_eager_eval(corparserStruct.CurModule, 0);
            }
            else CORAL_error(COR_BAD_ANNO, corSymbolString($2),
                     "yyparse:anno4", __FILE__, __LINE__);
	    delete $2; delete $3; delete $4; delete $11;
        }

    |    AT SYMBOL SYMBOL opt_adorn_string LPAREN optarglist RPAREN
                simple_arg  plus_or_minus PERIOD
        {
                  corArgList *al1 = new corArgList(corparserStruct.cur_arg_chain->count);
                  corpop_arg_chain(corparserStruct, *al1);
                  if ( *($2) == corPrioritizeSymbol) {
                      coradd_prioritize_annotation(corparserStruct, $3, $4,
                              al1, $8);
                  }
                  else CORAL_error(COR_BAD_ANNO, 
				   corSymbolString($2->SymbolP()), fname,
				   __FILE__, __LINE__);
		  delete $2; delete $3; delete $4; delete $8; 
	}

    |       AT SYMBOL plus_or_minus PERIOD
        {
            if ( *($2) == corSaveModuleSymbol)
                COR_set_save_module(corparserStruct.CurModule, $3);
            else if ( *($2) == corSupplementaryMagicSymbol)
                COR_set_supplementary_magic(corparserStruct.CurModule, $3);
            else if ( *($2) == corOrdinaryMagicSymbol)
                COR_set_ordinary_magic(corparserStruct.CurModule, $3);
            else if ( *($2) == corSupMagicIndexingSymbol)
                COR_set_sup_magic_indexing(corparserStruct.CurModule, $3);
            else if ( (*($2) == corNoMagicSymbol) && $3) 
                COR_set_no_magic(corparserStruct.CurModule);
            else if (*($2) == corExistOptSymbol)
                COR_set_exist_opt(corparserStruct.CurModule, $3);
            else if (*($2) == corUseFactoringSymbol)
                COR_set_use_factoring(corparserStruct.CurModule, $3);
            else if (*($2) == corUseOrdSearchSymbol)
                COR_set_use_ordsearch(corparserStruct.CurModule, $3);
            else if (*($2) == corEagerEvalSymbol)
                COR_set_eager_eval(corparserStruct.CurModule, $3);
            else if (*($2) == corUsePipeliningSymbol)
                COR_set_pipelining(corparserStruct.CurModule, $3);
            else if ( *($2) == corIndexDeltaSymbol)
                COR_set_index_deltas(corparserStruct.CurModule, $3);
            else if ( *($2) == corPredicateSNSymbol)
                COR_set_predicate_seminaive(corparserStruct.CurModule, $3);
            else if (*($2) == corMultisetSymbol)
                COR_set_multisets(corparserStruct.CurModule, $3);
            else if (*($2) == corCheckSubsumSymbol)
                COR_set_check_subsum(corparserStruct.CurModule, $3);
            else if (*($2) == corSingleSccSymbol)
                COR_set_single_scc(corparserStruct.CurModule, $3);
            else if (*($2) == corSingleAnswerSymbol)
                COR_set_single_answer(corparserStruct.CurModule, $3);
            else if (*($2) == corConvertFunctionsSymbol)
                COR_set_convert_functions(corparserStruct.CurModule, $3);
            else if (*($2) == corMonotonicSymbol)
                COR_set_monotonic(corparserStruct.CurModule, $3);
            else if (*($2) == corReturnUnifySymbol)
                COR_set_return_unify(corparserStruct.CurModule, $3);
            else if (*($2) == corNoRewritingSymbol)
                COR_set_rewriting(corparserStruct.CurModule, !($3));
            else if (*($2) == corNoPreProcessingSymbol)
                COR_set_preprocessing(corparserStruct.CurModule, !($3));
            else if (*($2) == corNonGroundFactsSymbol)
                COR_set_non_ground_facts(corparserStruct.CurModule, $3);
            else if (*($2) == corDiskResidentSymbol)
                COR_set_disk_resident(corparserStruct.CurModule, $3);
            else if (*($2) == corNestedSccEvalSymbol)
                COR_set_nested_scc_eval(corparserStruct.CurModule, $3);
            else
                CORAL_error(COR_BAD_ANNO, ($2)->GetSymbolStr(),fname,
                    __FILE__, __LINE__);
	    delete $2;
        }

        |       PERIOD {} 

    |    error PERIOD
        {
            CORAL_error(COR_BAD_SYNTAX, NULL, fname, __FILE__, __LINE__);
            yyerrok ;
            corparserStruct.rule_var_names.length = 0;
            corparserStruct.rhspreds.length = 0;
            corparserStruct.nested_index = 0;
            //parserStruct.parameters = new parameter_chain;
            corparseEnv->print_prompt();
        }
        clause {}
;

plus_or_minus:
                PLUS  { $$ = 1;}
        |       MINUS { $$ = 0; }
        |       { $$ = 1;}
;
clause_rhs:
        MARK rhs_pred_list opt_annotes PERIOD
        {
        }
    |    PERIOD
        {
            corparseEnv->cur_rule->annotations = NULL;
            corparseEnv->cur_rule->sips    = NULL;
        }
;

opt_annotes:    ALIAS
        {
            corparseEnv->cur_rule->sips = NULL;
        }

        literal_alias sips_or_rule_anno
        { 
           $$ = NULL;
        }
    |   { 
            $$ = NULL; 
            corparseEnv->cur_rule->annotations = NULL; 
            corparseEnv->cur_rule->sips = NULL;
        }
;
rhs_pred_list:    rhs_pred_list COMMA rhs_pred
    |    rhs_pred
/*
    |
*/
;
rhs_pred:    pred
        {
            coradd_to_array(&corparserStruct.rhspreds, (int)$1);
            // LEAK?  delete $1;
        }
    |
;

pred:        const LPAREN optarglist RPAREN
        {
            $$ = new corLiteral(corparserStruct, $1,
                      corparserStruct.cur_arg_chain->count);
            corpop_arg_chain(corparserStruct, $$->args);
	    delete $1;
        }
        |       const
        {
            $$ = new corLiteral(corparserStruct, $1, 0);
	    delete $1;
        }
    |    arg relop arg
        {
            $$ = new corLiteral(corparserStruct, $2, 2);
            ($$->args)[0] = *($1);
            ($$->args)[1] = *($3);
	    delete $1;
	    delete $3;
        }
    |    NOT arg relop arg
        {
            $$ = new corLiteral(corparserStruct, $3, 2);
            $$->negated = 1;
            ($$->args)[0] = *$2;
            ($$->args)[1] = *$4;
	    delete $2;
	    delete $4;
        }
    |    NOT const LPAREN optarglist RPAREN
        {
            $$ = new corLiteral(corparserStruct, $2, 
                      corparserStruct.cur_arg_chain->count);
            $$->negated = 1;
            corpop_arg_chain(corparserStruct, $$->args);
	    delete $2;
        }
    |    NOT const
        {
            $$ = new corLiteral(corparserStruct, $2, 0);
            $$->negated = 1;
	    delete $2;
        }
        |       CUT
        {
            $$ = new corLiteral(corparserStruct, corCutSymbol, 0);
        }
;

relop:        EQUAL        { $$ =  (corEquOpSymbol); }
    |    NOTEQUAL    { $$ = corNeqOpSymbol; }
    |    LESS        { $$ = corLssOpSymbol; }
    |    GREATER        { $$ = corGrtOpSymbol; }
    |    LESSEQUAL    { $$ = corLeqOpSymbol; }
    |    GREATEREQUAL    { $$ = corGeqOpSymbol; }
;


arglist:
              arglist COMMA arg
        {
            if (corparserStruct.CurModule.ConvertFunctions &&
                    ($3)->kindof() == COR_NUM_CONST &&
                    ($3)->cNum_kindof() == COR_ARITH_EXPR) {

                // need to create the equality literal 
                corVarArg *temp_var = NULL;
                corpush_arg_chain(corparserStruct);
                corappend_arg(corparserStruct, ($3));
                temp_var = corpull_out_arith_expr(corparserStruct);
                // the above function also does a pop_arg_chain
    
                corappend_arg(corparserStruct, temp_var);
            } else {
                corappend_arg(corparserStruct, ($3));
            }
	    delete $3;
        }
    |    
                 arg
        {
            corpush_arg_chain(corparserStruct);
            if (corparserStruct.CurModule.ConvertFunctions &&
                    ($1)->kindof() == COR_NUM_CONST &&
                    ($1)->cNum_kindof() == COR_ARITH_EXPR) {
                // need to create the equality literal 
                corVarArg *temp_var = NULL;
                corpush_arg_chain(corparserStruct);
                corappend_arg(corparserStruct, $1);
                temp_var = corpull_out_arith_expr(corparserStruct);
                // the above function also does a pop_arg_chain

                corappend_arg(corparserStruct, temp_var);
            } else {
                corappend_arg(corparserStruct, $1);
            }
	    delete $1;
        }
;

/* grouping_list: var_arg */
grouping_list: arg 
        {
            corpush_arg_chain(corparserStruct);
            corappend_arg(corparserStruct, $1);
	    delete $1;
        }
;


optarglist:    arglist
    |
        {
            corpush_arg_chain(corparserStruct);
        }
;
arg:        arg_primary
    |    arg PLUS arg
        {
	 $$ = new corCArg (new corExprArg(corAddOpSymbol,$1, $3));	
	 delete $1; delete $3;
        }
    |    arg MINUS arg
        {
            $$ = new corCArg (new corExprArg(corSubOpSymbol, $1, $3));
	    delete $1; delete $3;
        }
    |    arg STAR arg
        {
            $$ = new corCArg( new corExprArg(corMulOpSymbol, $1, $3));
	    delete $1; delete $3;
        }
    |    arg SLASH arg
        {
            $$ = new corCArg( new corExprArg(corDivOpSymbol, $1, $3));
	    delete $1; delete $3;
        }
    |    arg SLASHSLASH arg
        {
            $$ = new corCArg( new corExprArg(corIDivOpSymbol, $1, $3));
	    delete $1; delete $3;
        }

        |       arg LSHIFT arg
        {
            $$ = new corCArg(new corExprArg(corLShiftOpSymbol, $1, $3));
	    delete $1; delete $3;
        }
        |       arg RSHIFT arg
        {
            $$ =  new corCArg(new corExprArg(corRShiftOpSymbol, $1, $3));
	    delete $1; delete $3;
        }
        |       arg BITXOR arg
        {
            $$ =  new corCArg(new corExprArg(corBitxorOpSymbol, $1, $3));
	    delete $1; delete $3;
        }
        |       arg BITAND arg
        {
            $$ =  new corCArg(new corExprArg(corBitandOpSymbol, $1, $3));
	 delete $1; delete $3;
        }

        |       arg BITOR arg
        {
            $$ =  new corCArg(new corExprArg(corBitorOpSymbol, $1, $3));
	 delete $1; delete $3;
        }

    |     BITNOT arg_primary 
        {
            $$ =  new corCArg(new corExprArg(corBitnotOpSymbol, $2));
	 delete $2;
        }

        |       arg MOD arg
        {
            $$ =  new corCArg(new corExprArg(corModOpSymbol, $1, $3));
	 delete $1; delete $3;
        }


    |     MINUS arg_primary    %prec BITNOT
        {
            if (($2)->kindof() == COR_NUM_CONST) {
              if (($2)->cNum_kindof() == COR_INTEGER)
                  $$ = new corCArg(-(($2)->NumIval()));
              else 
                  $$ = new corCArg(-(($2)->NumDval()));
            }
            else {
                $$ = new corCArg (new corExprArg(corSubOpSymbol, $2));
            }
	    delete $2;
        }
;
arg_primary:    arg_fact
    |    LPAREN arg RPAREN
        {
            $$ = $2;
        }

simple_arg:   const LPAREN optarglist RPAREN
        {
            $$ = new corCArg (corcreate_functor(corparserStruct, $1));
	    delete $1;
        }

arg_fact:    const
        |    NUM
        |    var_arg
        |    UNDERSCORE
        { 

            // NOTE : this is a kludge ! Unification with dont care terms
            // is tricky, so instead we force the DontCareSymbol to be
            // treated each time as a distinct variable. The parser
            // NEVER generates  TheDontCareArg !! The scanner NEVER
            // returns UNDERSCORE !! Instead it returns VAR for "_"
               $$ = new corCArg (&corTheDontCareArg);
               fputs("DONT_CARE_ARG!!\n",corexEnv->error_file); 
        }
        |       DOLLAR 
        {
            //$$ = &TheParameterArg;
            $$ = new corCArg (corTheParameterArg);
        }
    |    const LPAREN optarglist RPAREN
        {
            $$ = new corCArg (corcreate_functor(corparserStruct, $1)); 
	    delete $1;
        }
    |    LBRACKET arglist VBAR arg RBRACKET
        {
            $$  = corcreate_list(corparserStruct, $4); 
                 /* delete $4 can only be done inside create_list. */
        }
    |    LBRACKET optarglist RBRACKET
        {
            $$ = corcreate_list(corparserStruct, NilArg);
        }
    |    LESS grouping_list GREATER
        {
            if (corparserStruct.lhs_flag) {
                corparseEnv->complexRelation++ ;
                if (corexEnv->dbg_aggregates) 
                    fprintf(corexEnv->error_file,
                        "nested att found complex relation %d\n",
                         corparseEnv->complexRelation);
                $$ = new corCArg( corcreate_multi_grouping(corparserStruct)) ;
            } else {
                CORAL_error( COR_GROUPING_ON_RHS,NULL, fname, __FILE__,
                    __LINE__);
                YYERROR;
            }
        }
        |       { corpush_cur_set(corparserStruct, corcreate_nested_multiset()); }
                LBRACE member_list RBRACE
        {
            $$ = new corCArg (corparserStruct.cur_set->info);
            corparseEnv->complexRelation++ ;
            if (corexEnv->dbg_aggregates)
              fprintf(corexEnv->error_file,"nested att found complex relation %d\n",
                    corparseEnv->complexRelation);
            corpop_cur_set(corparserStruct);
        } 
;

var_arg: VAR { $$ = 
		 new corCArg (new corVarArg($1, corrule_var_enter(corparserStruct, $1))); 
	       delete $1;
	      }

member_list:    member_list COMMA arg_fact 
                { corinsert_cur_set(corparserStruct, $3); 
		 delete $3;
		 }
        |       arg_fact
                { corinsert_cur_set(corparserStruct, $1); 
		  delete $1; }
        |       /* added this to allow an empty set {}  */
;

const:        SYMBOL     
;

literal_alias:        LBRACKET { corparserStruct.sip_var_names.length=0; }
        p_idlist RBRACKET 
        {
            $$ = NULL;
        }
;

p_id:        id
        { 
	 corparserStruct.sip_var_names.add($1);
	 // can't  delete $1, see Array.C, add only copies ptr!
	}
;
p_idlist:    p_idlist COMMA p_id | p_id | ;


sips_or_rule_anno: rule_anno_list
        {
            $$ = NULL;
        }
;
    
rule_anno_list:        
         rule_anno_list COMMA rule_anno
        {
            if ($3 != NULL)
                $$ = corlink_rule_anno($3,$1);
            else $$ = $1;
        }
    |    rule_anno
        {
            $$ = corlink_rule_anno($1,NULL);
        }
    |
        {
            $$ = NULL;
        }
;

rule_anno:    anno_name anno_case_details
        {
            $$ = cormake_anno($1,$2);
	    delete $1;
        }
    |     sips
        {
            $$ = NULL;
        }
;

anno_name:    SYMBOL
;

anno_case:    anno_case_name ARROW anno_case_details
        {
            $$ = cormake_anno_case($1,$3);
        }
;

anno_case_list:    anno_case COMMA anno_case_list
        {
            $$ = corlink_anno_case_list($3, $1);
        }
    |     anno_case
        {
            $$ = corlink_anno_case_list($1,NULL);
        }
;

anno_case_name:      SYMBOL
        {
            $$ = cormake_anno_case_name($1);
	    delete $1;
        }
    |    NUM
        {
            $$ = cormake_anno_case_name($1);
	    delete $1;
        }
    |    VAR
        {
            corVarArg *v= new corVarArg($1, corrule_var_enter(corparserStruct, $1)) ;
            $$ = cormake_anno_case_name(v);
	    delete $1;
        }
    |     UNDERSCORE
        {
            $$ = cormake_anno_case_name(&corTheDontCareArg);
        }
    |    const LPAREN optarglist RPAREN
        {
            $$ = cormake_anno_case_name(corcreate_functor(corparserStruct, $1));
	    delete $1;
        }
    |     LBRACKET optarglist RBRACKET
        {
            $$ = cormake_anno_case_name(corcreate_list(corparserStruct, NilArg));
        }
;

anno_case_details:    SYMBOL
        {
            $$ = cormake_anno_case_details($1);
	    delete $1;
        }
        |    NUM
        {
            $$ = cormake_anno_case_details($1);
	    delete $1;
        }
    |    VAR 
        {
            corVarArg *v= new corVarArg($1, corrule_var_enter(corparserStruct, $1)) ; 
            $$ = cormake_anno_case_details(v);
	    delete $1;
        }
    |    UNDERSCORE 
        { 
            $$ = cormake_anno_case_details(&corTheDontCareArg); 
        }
    |    const LPAREN optarglist RPAREN
        {
            $$ = cormake_anno_case_details(corcreate_functor(corparserStruct, $1));
	    delete $1;
        }
    |     LBRACKET optarglist RBRACKET
        {
            $$ = cormake_anno_case_details(corcreate_list(corparserStruct, NilArg));
        }
    |    LBRACKET  anno_case_list RBRACKET
        {
            $$ = cormake_anno_case_details($2);
        }
;

sips:        SIPS LBRACKET sip_list RBRACKET
        {
            corparseEnv->cur_rule->sips = $3;
        }
;
        

sip_list_entry: LBRACKET adorn_string 
                /** LBRACKET ad_idlist RBRACKET  **/
        COLON sip RBRACKET
        {
            $$ = new corsips;
            {
                char *ap;
                int i;
                /* ??? Check for possible length errors */
                $$->adorn.setlen(($2)->GetSymbolLen());
                for(i=0,ap = ($2)->GetSymbolStr(); 
		    i < ($2)->GetSymbolLen(); i++, ap++) 
                    if ( *ap == 'b')
                        $$->adorn.set(i);
            }
            $$->sip_list = $4;
            $$->next = NULL;
	    delete $2;
        }
;
sip_list:    sip_list COMMA sip_list_entry
        {
            $$ = $3;
            $3->next = $1;

        }
    |    sip_list_entry
        {
            $$ = $1;
            $1->next = NULL;
        }
    |    LBRACKET RBRACKET { $$ = NULL; }
;

opt_adorn_string:    
        {$$ = NULL;}
    | LBRACKET id RBRACKET
        {$$ = $2;}
;

adorn_string:    id
        {
            $$ = $1;
        }
;


sip:        sip COMMA arc
        {
            $$ = $1;
            $$->next = $3;
        }
    |    arc
        {
            $$ = $1;
        }
;
arc:        LBRACKET
        arc_idlist RBRACKET ARROW id LPAREN bound_var_list RPAREN
        {
            $$ = new corsip();
            $$->tail = corparseEnv->CurTail;
            $$->bound_vars = corparseEnv->CurBoundVars;
            $$->head = corsip_var_lookup(corparserStruct, $5);
                    $$->next = NULL;
            corparseEnv->CurTail.clear();
            corparseEnv->CurBoundVars.clear();
	    delete $5;
        }
    |
        {
            $$ = NULL;
        }
;
bound_var:    VAR
        {
            corparseEnv->CurBoundVars.set(corrule_var_lookup(corparserStruct, $1));
	    delete $1 ;
        }
        |
;
bound_var_list:    bound_var_list COMMA bound_var
        |       bound_var
;

arc_idlist:    arc_idlist COMMA arc_id
    |    arc_id
;

arc_id:        id
        {
            corparseEnv->CurTail.set(corsip_var_lookup(corparserStruct, $1));
            delete $1;
        }
;

id: const
    |    VAR
;
%%

void yyerror(char *s)
{
 fprintf(corexEnv->error_file, "%s line %d:\n", s, corparseEnv->cur_line_num);
 if (corexEnv->C_interrupt_raised > 0) {
   fprintf(corexEnv->error_file,
       "CORAL:: Interrupt occurred during execution!:");
   fprintf(corexEnv->error_file, "resetting state !\n");
   fprintf(corexEnv->error_file, "Type a period to continue.\n");
   fflush(corexEnv->error_file);
   // abort a consult that is halfway thru
   corparseStack->abort_file();
   corexEnv->C_interrupt_raised = 0;
 }
}
