//***********************************************************************
//    ParserEnv.h
//***********************************************************************

#include "ParserEnv.h"
#include "Globals.h"
#include "ClientProtocol.h"
#include "Assert.h"

extern "C"     int yylex(YYSTYPE *yylval);

//***********************************************************************

void corParserEnv::init()
{
    cur_line_num = 0;
    prompt = coralConfig.prompt;
    strcpy(prompt_name,prompt);
    prompt = prompt_name;
    show_prompt = 1;

    history = new corHistory(MAX_HISTORY_DEPTH);
    command_count = 0;
    is_history_command = 0;

    linebuf[0] = '\0';
    line_offset = 0;
    commandbuf[0] = '\0';
    command_offset = 0;

    in_comment = 0;

    fp = (FILE *) NULL;

    cur_rule = (corClause *) NULL;
    C_num_head_arith_lits = 0;
    complexRelation = 0;

        // This line is NOT necessary, as lexer is always set,
        // but its better to be safe then sorry. 
    lexer = (LEXER) yylex;

    next = (corParserEnv *) NULL;
}

//***********************************************************************
corParserEnv::~corParserEnv()
{
    delete history;
}

//***********************************************************************

void corParserEnv::print_prompt()
{

    if (show_prompt){

        if (coralConfig.server_mode){
              // Modified to use Client protocol
            fprintf(corexEnv->print_file,"%s\n",CP_OKAY);
            fflush(corexEnv->print_file);

        } else if (corexEnv->C_print_prompt_default){

            fputs(prompt,corexEnv->print_file); 
            fflush(corexEnv->print_file);
        }
    }
}

//***********************************************************************

void corParserEnv::handle_history()
{
     if (incr_command_count && !is_history_command)
        (history)->addCommand(commandbuf, ++command_count);
     
     if (line_offset + command_offset >= MAX_RULE_LENGTH) {
        fprintf(corexEnv->error_file,
            "CORAL::Warning : Line buffer size exceeded !\n");
     }

     strncat(linebuf, commandbuf, command_offset);
     line_offset += command_offset;
     commandbuf[0] = '\0'; command_offset = 0;
     is_history_command = 0;
}

//***********************************************************************
