
%{
/*****************************************************************************/
/*    Scanner.l    */
/*****************************************************************************/

#undef yywrap
int yywrap();

#include "Globals.h"
#include "Scanner.h"
#include "Gram.tab.h"

#ifndef MAX_STR_LEN
#define MAX_STR_LEN    1024
#endif

static char  str_buff[MAX_STR_LEN];
static char *str_end = &str_buff[MAX_STR_LEN-1];
static char *str_ptr;


#ifdef FLEX_SCANNER

#undef YY_INPUT

#    ifdef STRING_PARSE

#define    min(X,Y)    ((X)<(Y) ? (X) : (Y))

    // Scan input from a string instead
#define YY_INPUT(buf,result,max_size) \
  { \
     int n = min(max_size, corparseEnv->parselimit - corparseEnv->parseptr);\
     if (n>0){ \
        memcpy(buf,corparseEnv->parseptr,(unsigned int) n); \
        corparseEnv->parseptr += n; \
        result = n; \
     } else \
        result = YY_NULL; \
  }

#    else    // not STRING_PARSE

    // Scan input from a file pointer (default)
#define YY_INPUT(buf,result,max_size) \
    { \
        if (fgets(buf,max_size,yyin) == NULL) { \
           result = YY_NULL; \
        } else { \
           result = strlen(buf); \
        } \
    }


#    endif    // not STRING_PARSE

#else    // not FLEX_SCANNER

static int input_char;

#undef input
#undef unput

#    ifdef    STRING_PARSE

    // Scan input from a string instead
#define input() \
     ((input_char=*parseptr++) ? input_char : (parseptr--, 0))

    // This unput trusts lex not to put back more than it got so far
#define unput(c)  ( *(--parseptr) = c ) 

#    else  // not STRING_PARSE

    // Scan input from a file pointer (default)
#define input() \
     ((((input_char=getc(yyin))==EOF) || (input_char == 0)) ? (0) : input_char)

#define unput(c) { ungetc(c, yyin); }

#    endif // not STRING_PARSE
#endif     // not FLEX_SCANNER



EXTERN int yylex ARGS((YYSTYPE *lvalp));

#undef YYLMAX
#define YYLMAX 2048

/*  To give some args to yylex */
#undef YY_DECL
#define YY_DECL int yylex YY_PROTO((YYSTYPE *lvalp))

/**** Kludge to support Dont Care terms.
<INITIAL>"_"    { C_STRCAT(corparseEnv->commandbuf,yytext); return UNDERSCORE; }
***/


#define C_STRCAT(A,B) \
        if (corexEnv->C_do_history_default) { \
            int tmp; \
            if ((tmp = corparseEnv->command_offset + yyleng) \
                       < MAX_RULE_LENGTH) { \
                strncat((A)+corparseEnv->command_offset,(B),yyleng); \
                corparseEnv->command_offset = tmp ; \
            } else {\
                corexEnv->C_do_history_default = 0;\
            }\
        }

%}
%Start C_COMMENT C_STR
%a 9000
%o 9000

dot             \.
digit           [0-9]
alpha           [a-zA-Z_]
capalpha        [A-Z]
alphanum        {alpha}|{digit}
optsign         [\+\-]?
litquote        \\\"

number          {digit}+
integer         {optsign}{number}
frac            {dot}{number}
mantissa        {number}|{number}{frac}
exponent        [Ee]{integer}
real            {mantissa}{exponent}?

%%

<INITIAL>\"[^\\\"]*\" {   /* Simple quoted strings */
                        C_STRCAT(corparseEnv->commandbuf,yytext);
                        yytext[yyleng-1] = '\0';
                        lvalp->carg = new corCArg (corEnterSymbol(yytext+1)) ;
                        /*  corparseEnv->command_offset -= 2;  */
                        return SYMBOL;
                      }

<INITIAL>\"         { /* For strings that contain escaped characters */
                      C_STRCAT(corparseEnv->commandbuf,yytext);
                      BEGIN(C_STR);
                      str_ptr = str_buff;
                    }

<INITIAL>"sips"     { C_STRCAT(corparseEnv->commandbuf,yytext); return SIPS; }
<INITIAL>"alias"    { C_STRCAT(corparseEnv->commandbuf,yytext); return ALIAS; }
<INITIAL>"mod"      { C_STRCAT(corparseEnv->commandbuf,yytext); return MOD; }
<INITIAL>"not"      { C_STRCAT(corparseEnv->commandbuf,yytext); return NOT; }
<INITIAL>"del"      { C_STRCAT(corparseEnv->commandbuf,yytext); return DEL; }

<INITIAL>"c_module" { C_STRCAT(corparseEnv->commandbuf,yytext);
                      return START_C_MODULE;
                    }
<INITIAL>"end_c_module" { C_STRCAT(corparseEnv->commandbuf,yytext);
                            return END_C_MODULE;
                        }
<INITIAL>"module"     { C_STRCAT(corparseEnv->commandbuf,yytext); return MODULE; }
<INITIAL>"export"     { C_STRCAT(corparseEnv->commandbuf,yytext); return EXPORT; }
<INITIAL>"end_module" {C_STRCAT(corparseEnv->commandbuf,yytext);return END_MODULE;}

<INITIAL>"<<"       { C_STRCAT(corparseEnv->commandbuf,yytext); return LSHIFT; }
<INITIAL>">>"       { C_STRCAT(corparseEnv->commandbuf,yytext); return RSHIFT; }
<INITIAL>"bitxor"   { C_STRCAT(corparseEnv->commandbuf,yytext); return BITXOR; }
<INITIAL>"bitand"   { C_STRCAT(corparseEnv->commandbuf,yytext); return BITAND; }
<INITIAL>"bitor"    { C_STRCAT(corparseEnv->commandbuf,yytext); return BITOR; }
<INITIAL>"bitnot"   { C_STRCAT(corparseEnv->commandbuf,yytext); return BITNOT; }  
<INITIAL>"is"       { C_STRCAT(corparseEnv->commandbuf,yytext); return EQUAL; }

<INITIAL>[A-Z]{alphanum}*  |
<INITIAL>[A-Z][A-Za-z0-9_.]*{alphanum}+    {
                            lvalp->carg = new corCArg(corEnterSymbol(yytext));
                            C_STRCAT(corparseEnv->commandbuf,yytext);
                            return VAR;
                    }

<INITIAL>"_"        {
                        lvalp->carg = new corCArg(corEnterSymbol(yytext));
                        C_STRCAT(corparseEnv->commandbuf,yytext);
                        return VAR;
                    }

<INITIAL>{number}   {
                        lvalp->carg = new corCArg( atoi(yytext)) ;
                        C_STRCAT(corparseEnv->commandbuf,yytext) ; 
                        return NUM;
                    }

<INITIAL>{real}     {
                        lvalp->carg = new corCArg(atof(yytext));
                        C_STRCAT(corparseEnv->commandbuf,yytext); return NUM;
                    }


<INITIAL>[a-z]([#A-Za-z0-9_.]*[#A-Za-z0-9_])? {
                        lvalp->carg = new corCArg(corEnterSymbol(yytext));
                        C_STRCAT(corparseEnv->commandbuf,yytext);
                        return SYMBOL;
                    }


<INITIAL>","      { C_STRCAT(corparseEnv->commandbuf,yytext); return COMMA; } 
<INITIAL>":"      { C_STRCAT(corparseEnv->commandbuf,yytext); return COLON; }
<INITIAL>"."      { C_STRCAT(corparseEnv->commandbuf,yytext); return PERIOD;}
<INITIAL>":-"     { C_STRCAT(corparseEnv->commandbuf,yytext); return MARK; }
<INITIAL>"<-"     { C_STRCAT(corparseEnv->commandbuf,yytext); return MARK; }
<INITIAL>"+="     { C_STRCAT(corparseEnv->commandbuf,yytext); return APPEND;}
<INITIAL>"-="     { C_STRCAT(corparseEnv->commandbuf,yytext); return DELETE;}
<INITIAL>"*="     { C_STRCAT(corparseEnv->commandbuf,yytext); return UPDATE;}
<INITIAL>":="     { C_STRCAT(corparseEnv->commandbuf,yytext); return ASSIGN;}
<INITIAL>"==>"    { C_STRCAT(corparseEnv->commandbuf,yytext); return ARROW; }
<INITIAL>"//"     { C_STRCAT(corparseEnv->commandbuf,yytext); return SLASHSLASH; }
<INITIAL>"("      { C_STRCAT(corparseEnv->commandbuf,yytext); return LPAREN; }
<INITIAL>")"      { C_STRCAT(corparseEnv->commandbuf,yytext); return RPAREN; }
<INITIAL>"{"      { C_STRCAT(corparseEnv->commandbuf,yytext); return LBRACE; }
<INITIAL>"}"      { C_STRCAT(corparseEnv->commandbuf,yytext); return RBRACE; }
<INITIAL>"["      { C_STRCAT(corparseEnv->commandbuf,yytext); return LBRACKET; }
<INITIAL>"]"      { C_STRCAT(corparseEnv->commandbuf,yytext); return RBRACKET; }
<INITIAL>"$"      { C_STRCAT(corparseEnv->commandbuf,yytext); return DOLLAR; }
<INITIAL>"|"      { C_STRCAT(corparseEnv->commandbuf,yytext); return VBAR; }
<INITIAL>"@"      { C_STRCAT(corparseEnv->commandbuf,yytext); return AT; }
<INITIAL>"+"      { C_STRCAT(corparseEnv->commandbuf,yytext); return PLUS; }
<INITIAL>"-"      { C_STRCAT(corparseEnv->commandbuf,yytext); return MINUS; }
<INITIAL>"*"      { C_STRCAT(corparseEnv->commandbuf,yytext); return STAR; }
<INITIAL>"/"      { C_STRCAT(corparseEnv->commandbuf,yytext); return SLASH; }
<INITIAL>"?"      { C_STRCAT(corparseEnv->commandbuf,yytext); return QUESTION; }
<INITIAL>"="      { C_STRCAT(corparseEnv->commandbuf,yytext); return EQUAL; }
<INITIAL>"\\="    { C_STRCAT(corparseEnv->commandbuf,yytext); return NOTEQUAL; }
<INITIAL>"~="     { C_STRCAT(corparseEnv->commandbuf,yytext); return NOTEQUAL; }
<INITIAL>"<"      { C_STRCAT(corparseEnv->commandbuf,yytext); return LESS; }
<INITIAL>">"      { C_STRCAT(corparseEnv->commandbuf,yytext); return GREATER; }
<INITIAL>"<>"     { C_STRCAT(corparseEnv->commandbuf,yytext); return NOTEQUAL; }
<INITIAL>"<="     { C_STRCAT(corparseEnv->commandbuf,yytext); return LESSEQUAL; }
<INITIAL>">="     { C_STRCAT(corparseEnv->commandbuf,yytext); return GREATEREQUAL;}
<INITIAL>"~"      { C_STRCAT(corparseEnv->commandbuf,yytext); return NOT; }
<INITIAL>"!"      { C_STRCAT(corparseEnv->commandbuf,yytext); return CUT; }


<INITIAL>"%".*$         ;    /* comment; ignore */
<INITIAL>"/*"           { BEGIN(C_COMMENT); corparseEnv->in_comment = 1;}
<C_COMMENT>[^*\n]*      ;      /* ignore anything that's not a '*' */
<C_COMMENT>"*"+[^*/\n]* ;     /* ignore '*'s not followed by '/'s */
<C_COMMENT>\n           { corparseEnv->cur_line_num++;
                          corparseEnv->linebuf[0] = '\0';
                          corparseEnv->line_offset = 0;
                        }

<C_COMMENT>"*"+"/"      { BEGIN(INITIAL); corparseEnv->in_comment = 0;}


<C_STR>\"           { /* Closing quote */
                        C_STRCAT(corparseEnv->commandbuf,yytext);
                        *str_ptr = '\0';
                        lvalp->carg =  new corCArg(corEnterSymbol(str_buff));
                        BEGIN(INITIAL);
                        return SYMBOL;
                    }

<C_STR>[^\\\"]+     { /* Normal chars */
                        C_STRCAT(corparseEnv->commandbuf,yytext);
                        if ((str_ptr + yyleng) < str_end){
                            strcpy(str_ptr,yytext);
                            str_ptr += yyleng;
                        }
                    }

<C_STR>\\n          { /* Newline char */
                        C_STRCAT(corparseEnv->commandbuf,yytext);
                        if (str_ptr < str_end)
                            *str_ptr++ = '\n';
                    }

<C_STR>\\t          { /* Tab char */
                        C_STRCAT(corparseEnv->commandbuf,yytext);
                        if (str_ptr < str_end)
                            *str_ptr++ = '\t';
                    }

<C_STR>\\b          { /* Backspace char */
                        C_STRCAT(corparseEnv->commandbuf,yytext);
                        if (str_ptr < str_end)
                            *str_ptr++ = '\b';
                    }

<C_STR>\\f          { /* Formfeed char */
                        C_STRCAT(corparseEnv->commandbuf,yytext);
                        if (str_ptr < str_end)
                            *str_ptr++ = '\f';
                    }

<C_STR>\\.          { /* Literal char (like " and \) */
                        C_STRCAT(corparseEnv->commandbuf,yytext);
                        if (str_ptr < str_end)
                            *str_ptr++ = yytext[1];
                    }

[ \t\f]             ;     /* space/tab/formfeed; ignore */

\n|\r               {
                      corparseEnv->cur_line_num++;
                      corparseEnv->linebuf[0] = '\0';
                      corparseEnv->line_offset = 0;
                    }

.                   {      /* any other single char; error */
                      fprintf(corexEnv->error_file,
                                "Error: Line %d: Illegal Character",
                                corparseEnv->cur_line_num);
                      if (yytext[0]>=' ' && yytext[0] <='~')
                        fprintf(corexEnv->error_file,": `%c'\n", yytext[0]);
                      else
                        fprintf(corexEnv->error_file,
                                ", ASCII: %03o (octal)\n", yytext[0]);
                    }

%%

//***********************************************************************
//
// yywrap() - This function is called by the lexer [yylex() or sslex()] when
//            the end-of-file (or end-of-string for sslex()) is reached.
//            It gets the next file/string ready (if any) and returns 1
//            which indicates to the lexer that there are no more files
//            remaining.  The lexer returns 0 to the parser, indicating
//            no more tokens are remaining. The parser function, yyparse(),
//            then returns 0, indicating that the parse is done.  This
//            behaviour allows a single item to be parsed to be placed
//            on the stack and parsed without the entire stack being consumed.
//
//***********************************************************************

int yywrap()
{
    if (corparseEnv->in_comment) {
        fprintf(corexEnv->error_file,
            "ERROR: End of file reached inside comment.\n" );
        corparseEnv->in_comment = 0;
    }

    // Prepare the next file to be read, if there is one, but return
    // so that yyparse() returns in any case.

    corparseStack->get_next_file();
    return(1);
}

//***********************************************************************
