// *************************************************************************
//    HashSimpleRelation.C
// *************************************************************************

#include "HashSimpleRelation.h"

#include "Globals.h"
#include "PrioritizeInfo.h"
#include "CoralError.h"
#include "Unify.h"

// *************************************************************************

int corC_NewHashRelation = 0;
int corC_DeleteHashRelation = 0;

// *************************************************************************
corHashSimpleRelation::corHashSimpleRelation( int arity, int delta, int buckets)
    : corStorageRelation(arity, delta), index_set(this)
{
    // NOTE: remember to update the code in empty_relation() and in 
    //        move_all_entries() when updating this code.

    // MEMORY LEAK : Praveen
    corBitVector *bv1 = new corBitVector(arity);

    for (int i = 0; i < arity; i++) {
        // all arguments bound
        bv1->set(i);
    }
    corHashIndex *h1 = new corHashIndex(this, bv1, buckets); 
    index_set.insert_index(h1);
    delete bv1;
    corC_NewHashRelation++;
}

// *************************************************************************
void corHashSimpleRelation::empty_relation(int deleteTuples)
{
    //    Empty all tuples out of relation.  

    if (deleteTuples) {
        HashIndexPtr index0 = (HashIndexPtr) index_set[0];
        ASSERT (index0 != NULL);
        index0->delete_all_tuples();
    }

    //    Empty out the indices

    corTupleLink* next;
    for (corTupleLink *link = chain; link != NULL; link = next) {
        if (deleteTuples) delete link->tuple;
        next = link->next;
        delete link;
    }
    chain = NULL;
    lastptr = &chain;

    //LinkedRelation::empty_relation(deleteTuples);

    for (int i = 0; i < index_set.num_indices(); i++) {
        index_set[i]->free_all_entries();
        index_set[i]->reset_index();
    }

    //    Recursively empty out the delta relations and add them to the 
    //    free_reln_chain

    corRelationLink *old;
    if (delta_indexed && delta_relations) {
        while(delta_relations->reln_chain) {
            corRelation *reln = delta_relations->reln_chain->reln;
            old = delta_relations->reln_chain;
            delta_relations->reln_chain = old->next;

            reln->empty_relation(deleteTuples);    
            old->next = free_reln_list;
            free_reln_list = old;
        }
        delta_relations->reln_lastptr = NULL;
        delta_relations = NULL;
    }

    headMark.next = &headMark;
    headMark.prev = &headMark;

    count  = 0;   
}

// *************************************************************************
int corHashSimpleRelation::move_all_entries(corStorageRelation *rel)
{
    // Moves all entries in the main relation of (this) to the relation rel.
    // This DOES NOT move entries that are in the delta_relations list or in 
    // the LinkedRelation list.  
    // In general it would be good to have this function
    // do that too, but right now this function is configured to only move
    // tuples from delta  relations into the main relation.

    headMark.next = &headMark;
    headMark.prev = &headMark;

    // NOTE:  If relation count is present, reset it here.
    rel->count += count;

    index_set.move_all_entries(rel);
    for (int i = 0; i < index_set.num_indices(); i++) {
        index_set[i]->reset_index();
    }
    count = 0;
    return 1;
}

// *************************************************************************
void corHashSimpleRelation::free_indices()
{
    corTupleLink* next;

    for (corTupleLink *link = chain; link != NULL; link = next) {
        next = link->next;
        // This doesn't free the components of the tuple.
        delete link->tuple;
        delete link;
    }
    chain = NULL;
    lastptr = &chain;

    //LinkedRelation::free_all();
    // when deleting relation, also delete the indices to the relation
    for (int i = 0; i < index_set.num_indices(); i++)
        delete ((corHashIndex *) index_set[i]); 
}

// *************************************************************************
void corHashSimpleRelation::free_delta_relations()
{
    corRelationLink *old;
    if (delta_indexed && delta_relations) {
        for (corRelationLink *tmp=delta_relations->reln_chain; tmp != NULL; ) {
            delete tmp->reln;    
            old = tmp;
            tmp = tmp->next;
            delete old;
        }
    }
    for (corRelationLink *tmp = free_reln_list; tmp != NULL;) {
        delete tmp->reln;
        old = tmp;
        tmp = tmp->next;
        delete old;
    }
}

// *************************************************************************
void corHashSimpleRelation::free_all()
{

    headMark.next = &headMark;
    headMark.prev = &headMark;
    
    HashIndexPtr index0 = (HashIndexPtr) index_set[0];
    ASSERT (index0 != NULL);
    index0->delete_all_tuples();

    free_indices();
    free_delta_relations();
    corC_DeleteHashRelation++;
}

// *************************************************************************
corHashSimpleRelation::~corHashSimpleRelation()
{
    free_all();
}


// *************************************************************************
void corHashSimpleRelation::freeMark(corRMark *mark, int hide_em)
{
    if (mark == NULL) return;
    corRMark *this_mark = mark;

    // Unlink mark
    this_mark->next->prev = this_mark->prev;
    this_mark->prev->next = this_mark->next;

    // add it to the free marks list
    this_mark->next = free_marks;
    free_marks = this_mark;

    // check if this was the first mark in the list of marks
    if (this_mark->prev != &headMark) return;

    // thus, if an "intermediate" mark is deleted, nothing is done!
    // this is fine for (!delta_indexed) relations, but with delta_indexed
    // relations, an alternative is to merge the delta relations 
    // corresponding to this mark with the "previous" relation.

    if (!delta_indexed) {
        corTupleLink *limit;

        // Move all linked tuples upto new first mark to hash table.
        if (there_are_marks()) {
            limit = *headMark.next->lastptr;
            for (corRMark *hmark = headMark.next; ; hmark = hmark->next) {
                if (hmark == &headMark) break;
                if (*hmark->lastptr != limit) break;
                hmark->lastptr = &chain;  
            }
        }
        else
            limit = *lastptr;
    
        corTupleLink *old;
        while (chain && chain != limit) {
            corTuple *tuple = chain->tuple;
            old = chain;
            chain = chain->next;
            if (!hide_em) {
                count++;
                insert_raw(tuple);
            }
            else {
                insert_in_hidden_relation(tuple);
            }
            delete old;
            // make sure this calls the index updates also.
        }
        if (chain == NULL)
            lastptr = &chain;
    }
    else { 
        corRelationLink *limit;

        ASSERT((delta_relations!= NULL));
        ASSERT((delta_relations->reln_lastptr != NULL));
        
        if (there_are_marks()) 
            limit = headMark.next->rel_p;
        else
            limit = NULL;

        corRelationLink *old;
        while (delta_relations->reln_chain!=limit) {
            corStorageRelation *reln = delta_relations->reln_chain->reln;
            old = delta_relations->reln_chain;
            delta_relations->reln_chain = old->next;

#            define DELTA_INDEX_HACK 
#            ifdef DELTA_INDEX_HACK
                // WARNING - we assume that the delta relation is a 
                // HashSimpleRelation.
                if (!hide_em) {
                    ((corHashSimpleRelation*)reln)->move_all_entries(this);
                }
                else {
                    reln->add_all_tuples(this,0,0,hide_em);
                    reln->empty_relation();
                }
#            else
                reln->add_all_tuples(this,0,0,hide_em);
                reln->empty_relation();
#            endif

            old->next = free_reln_list;
            free_reln_list = old;
        }
        if (delta_relations->reln_chain==NULL)
            delta_relations->reln_lastptr = NULL;
    }
}

// *************************************************************************
void corHashSimpleRelation::free_allMarks()
{
    corRMark *hmark;

    while (headMark.next != &headMark) {
        hmark = headMark.next;
        freeMark(hmark);
    }
    ASSERT((headMark.prev == headMark.next));
}

// *************************************************************************
void corHashSimpleRelation::initialize_iterator( corTupleIterator& iterator )
{
#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file,
                "\nHashSimpleRelation::initialize_iterator: ");
            iterator.relation->name->print(NULL, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, " ( ");
            iterator.arg_list.print(iterator.bindenv, corexEnv->trace_file);
            fprintf(corexEnv->trace_file, " )\n");
        }
#    endif
    iterator.stack_mark.get_mark();

    corTerm t1;

    scan_count++;

    if (!delta_indexed) {
        if (iterator.start_mark) {
            //    don't need to go to the hash table
            iterator.ipos = Scanning_New;
            iterator.ppos = (iterator.start_mark)->lastptr;
        }
        else {
            //    need to scan from hash table
            //    find the appropriate index to be used
            iterator.index = index_set.find_index(iterator);
        }
    } 
    else {
        if ( !iterator.in_delta()) {
            //    Starting on main relation
            if (iterator.start_mark) {
                //    go to the delta relations
                iterator.ipos = Scanning_Init;
                iterator.delta_rel_pos = (iterator.start_mark)->rel_p;
            }
            else {
                //    need to scan from hash table
                //    find the appropriate index to be used
                iterator.index = index_set.find_index(iterator);
            }
        }
        else {
            //    In a delta relation
            iterator.index = index_set.find_index(iterator);
        }
    }
}

// *************************************************************************
/*
    Tries to find tuple in relation that unifies with 
        <iterator.arg_list, iterator.bindenv>

    On success:  
        Resets iterator.no_match, sets iterator.tuple_env to
        the environment created for the tuple, and returns the fetched 
        tuple.  The tuple is NOT a copy, and should not be deleted or 
        destructively modified where it is called from.  The same holds
        for tuple.bindenv.  However, iterator.tuple_env (if it is non-null)
        is a local copy and may be modified by the calling routine.
        It may need to be disposed off by the calling routine once it has
        served its purpose.

        In general, DO NOT assume that a return value of NULL implies 
        that no tuple was found - evaluable relations do not create tuples,
        but merely modify iterator.bindenv.  Hence they always return null.

    On failure:  
        Sets iterator.no_match.
*/

#define FAIL_RETURN \
{ \
    iterator.ipos = Scanning_Done;     \
    iterator.set_no_match();    \
    return NULL;            \
}

corTuple* corHashSimpleRelation::get_next_tuple(corTupleIterator& iterator)
{
#    ifdef DO_PROFILE
        if (corT_Stack.count > 0) {
            (corT_Stack.top()->moduledata->
                module_info.corGlobalProfileInfo.getnexts)++;
        }
        else {
            (corGlobalProfileInfo.getnexts)++;
        }
#    endif

    iterator.tuple_env = NULL;

    corTuple *result = NULL;

    if (corexEnv->C_interrupt_raised) FAIL_RETURN;

    if ((iterator.ipos == Scanning_Init)) 
        initialize_iterator( iterator );
    else
        iterator.stack_mark.pop_to();

    if (!delta_indexed) {
        // do the old stuff with chain and everything
        if (iterator.ipos != Scanning_New) {
            // need to scan from hash table
            result = iterator.index->get_next_tuple(iterator);
            if (result != NULL) {
                // found something in hash table
                return result; 
            }
            else {
                iterator.ipos = Scanning_New;
                iterator.ppos = &chain;
                //    this is required because we don't want 
                //    HashIndex::get_next_tuple() to set iterator.ipos to 
                //    Scanning_New at the end of search
            }
        }
    
        corStackMark stack_mark;

        // need to scan from the chain now

        for (;;) {
            // Continue until we find a match

            corTupleLink** next = (corTupleLink**)iterator.ppos;
            corTupleLink* temp;
            if (*next == NULL) {
                iterator.ipos = Scanning_Done;
                iterator.set_no_match();
                result = NULL;
                stack_mark.pop_to();
                return result;
            }
            if (iterator.end_mark && (iterator.end_mark)->lastptr == next) {
                iterator.ipos = Scanning_Done;
                iterator.set_no_match();
                result = NULL;
                stack_mark.pop_to();
                return result;
            }
            corTuple *tuple = (*next)->tuple;

            if (tuple->is_deleted()) {
                // the tuple has been logically deleted
                temp = *next;
                *(next) = (*next)->next;
      
                /*
                The if check is needed because we might have a rule of the form
                try :- p(X), delete(p(X)), print(X), p(Y), print(Y).
                The delete() builtin merely marks the tuples as deleted.
                The later p iterator must not reclaim the space, because this 
                will corrupt the first iterator.
                */

                if (scan_count == 1) {
                    delete temp;
                    //    PRAVEEN:: I am doubtful about adding this here.
                    //    Will this break the system ?
                    /********
                    for (int ti = 0; ti < tuple->arity(); ti++)
                        delete tuple->arg(ti);
                    delete tuple->_args;
                    delete tuple->bindenv;
                    delete tuple;
                    ********/
                }

#                ifdef DO_PROFILE
                    if (corT_Stack.count > 0) {
                        (corT_Stack.top()->moduledata->
                            module_info.corGlobalProfileInfo.deleted_tuples)++; 
                    }
                    else {
                        (corGlobalProfileInfo.deleted_tuples)++; 
                    }
#                endif
                continue;
            }
            
            iterator.ppos = &(*next)->next;

            if (corunify_literal(iterator, tuple, corexEnv->dbg_get_next))
                return (tuple);
            stack_mark.pop_to();
        }
    }
    else {
        // relation is delta-indexed
        /*
            WARNING: the following hack has been done.
            iterator not copied when calling delta relations.  so it
            has the same "relation" field value, but each time the index field
            changes based on the main/delta relations.
            note that iterator.get_next_tuple is NOT called, which
            would be wrong because the "relation" field value is not correct!!

            the main/delta relations->get_next_tuple is called directly.
            the following ASSUMPTION is made:
            -- relation->get_next_tuple(iterator) does NOT use the 
               relation field of iterator.
            -- HashSimpleRelation::get_next_tuple satisfies this assumption.
        */

        if (!iterator.in_delta()) {
            // Scanning from main hash table
            result = iterator.index->get_next_tuple(iterator);
            if (result != NULL) return result;
            if (!delta_relations) FAIL_RETURN

            // now go into the delta relations
            iterator.ipos = Scanning_Init;
            iterator.delta_rel_pos = delta_relations->reln_chain;
        }

        // delta_rel_pos has to have been set by this point.
        if (delta_relations != NULL) {
            // iterator scanning deltas, this == main relation
            for(;;) {
                // Loop till successful delta is found
                if (iterator.ipos == Scanning_Init) {
                    if (iterator.delta_rel_pos == NULL) {
                        FAIL_RETURN
                    }
                    if (iterator.end_mark && (iterator.delta_rel_pos ==
                        (iterator.end_mark)->rel_p)) 
                    {
                        FAIL_RETURN
                    }
                }
                result = iterator.delta_rel_pos->reln->get_next_tuple(iterator);
                if (result != NULL) return result;

                // Done with current delta relation.  Got to go to next delta.
                iterator.delta_rel_pos = iterator.delta_rel_pos->next;
                iterator.ipos = Scanning_Init;      
            } 
        }
        else {
            // iterator scanning deltas, this == delta relation
            return iterator.index->get_next_tuple(iterator);
        }
    }
}

#undef FAIL_RETURN

// *************************************************************************
/*
    As above, except that it does not return the tuple, but instead
    returns iterator.bindenv on success, and NULL on failure.
    returns a fetched tuple
*/

corBindEnv* corHashSimpleRelation::get_next(corTupleIterator& iterator)
{
    corTuple *result = get_next_tuple(iterator);
    if (!result) return NULL;
    return iterator.bindenv;
}

// *************************************************************************
void corHashSimpleRelation::release(corTupleIterator& )
{
    //    Do nothing.  No state release is required
}


// *************************************************************************
int corHashSimpleRelation::insert_raw( corTuple *tuple )
{
    //    inserts tuple into each index constructed for the relation

    if (tuple->is_deleted()) {
        // tuple has been logically deleted
        return 0;
    }
        
    for (int i = 0; i < index_set.num_indices(); i++) {
        index_set[i]->insert_tuple(tuple);
    }
    return 1;
}

// *************************************************************************
int corHashSimpleRelation::is_subsumed(corTuple *newtuple, corRMark *start_mark,
    corRMark *end_mark)
{
    corStackMark stack_mark;
    int result;
    corTupleLink **next, *temp;
    corGenericIndex *index;

    if (! start_mark) {
        // need to go to the hash table
        // find the appropriate index to be used
        index = index_set.find_index(newtuple);
    }
    else
        index = NULL;

    if (!delta_indexed) {
        // do the old stuff with chain and everything
        if (index) {
            // need to scan from hash table
            result = index->is_subsumed(newtuple);
            if (result != 0) {
                // found something in hash table
                return result; 
            }
        }
    
        // need to scan from the chain now
        next = &chain;

        for (;;) {
            // Continue until we find a match
            if (*next == NULL) 
                return( is_agg_subsumed(newtuple, start_mark, end_mark) );
            if (end_mark && (end_mark)->lastptr == next) 
                return( is_agg_subsumed(newtuple, start_mark, end_mark) );
            corTuple *tuple = (*next)->tuple;

            if (tuple->is_deleted()) {
                // the tuple has been logically deleted
                temp = *next;
                *(next) = (*next)->next;
                delete temp;

#                ifdef DO_PROFILE
                    if (corT_Stack.count > 0) {
                        (corT_Stack.top()->moduledata->
                            module_info.corGlobalProfileInfo.deleted_tuples)++; 
                    }
                    else {
                        (corGlobalProfileInfo.deleted_tuples)++; 
                    }
#                endif
                continue;
            }

            next = &((*next)->next);

            if (corsubsumes_tuple(tuple, newtuple, (DISPLAY_INSERTIONS)))
                return 1;
        }
    }
    else {
        // relation is delta-indexed

        if (index) {
            // Need to look at index for this relation.
            result = index->is_subsumed(newtuple);
            if (result != 0) return result;
        }
        if (delta_relations != NULL) {
            corRelationLink* delta_rel_pos = delta_relations->reln_chain;

            if (start_mark) { /* Find first delta relation to be used */
            while((start_mark)->rel_p != delta_rel_pos &&
                delta_rel_pos != NULL)
                delta_rel_pos = delta_rel_pos->next;
            }

            for(;;) { // Loop till successful delta is found
                if (!delta_rel_pos) {
                    // End of delta_relations chain
                    return(is_agg_subsumed(newtuple, start_mark, end_mark));
                }
                if (end_mark && (delta_rel_pos == (end_mark)->rel_p)) {
                    return(is_agg_subsumed(newtuple, start_mark, end_mark));
                }
                result = delta_rel_pos->reln->is_subsumed(newtuple);
                if (result != 0) return result;
                // Done with current delta relation.  Got to go to next delta.
                delta_rel_pos = delta_rel_pos->next;
            } 
        }
        return(is_agg_subsumed(newtuple, start_mark, end_mark));
    }
}

// *************************************************************************
int corHashSimpleRelation::add_hidden_relation(corPrioritizeInfo *) 
{
    heap = NULL;
    // Nothing more to do.  The heap _is_ the hidden relation.
    return 0;
}

// *************************************************************************
int corHashSimpleRelation::insert_in_hidden_relation(corTuple *tuple) 
{
    if (tuple->is_deleted()) return 0;

    corTerm val_term(tuple->arg(priority_info->argnum), tuple->bindenv);
    corfull_dereference_term(val_term);
    if (val_term.expr.kindof() != COR_NUM_CONST) {
        CORAL_error(COR_BAD_ARG_TYPE, "argument to prioritize on not a number",
            "HashSimpleRelation::insert_in_heap");
        return 0;
    }
    corCArg val =  val_term.expr;

    corHeapNode *newnode = new corHeapNode(tuple, val);

    if (!heap)
        heap = newnode;
    else
        corheap_insert(heap, tuple, val, priority_info->order);

#    ifdef DO_TRACE
        if (DISPLAY_INSERTIONS)  {
            fprintf(corexEnv->error_file, "\tHiding fact: %s(", name->string());
            tuple->printon(corexEnv->error_file);
            fprintf(corexEnv->error_file, ")\n");
        }
#    endif

    return 1;
}

// *************************************************************************
int corHashSimpleRelation::insert_tuple(corTuple *tuple)
{
    if (there_are_marks()) {
        if (delta_indexed) {
            delta_relations->insert_tuple(tuple);
        }
        else {
            // PRAVEEN :: this here is a problem !!!
            this->tlist_insert_tuple(tuple);
            count++;
        }
    }
    else {
        count++;
        insert_raw(tuple);
    }

    //        _hash = unknownhashvalue; 
    // to ensure set is re-hashed, since it changed.
    // this fix doesn't work, consistency of hash-cons table may be violated!

    // return StorageRelation::insert_tuple(tuple);
    return 1;
}

// *************************************************************************
int corHashSimpleRelation::pop_from_hidden_relation() 
{
    for(;;) {
        // Until new tuple is found
        if (!heap)
            return 0;

        corTuple *tuple = corpop_heap(&heap,priority_info->order);

        if (tuple->is_deleted())
            continue;

#        ifdef DO_TRACE
            if (DISPLAY_INSERTIONS)  {
                fprintf(corexEnv->error_file, "\tUnhiding fact: %s(",
                        name->string());
                tuple->printon(corexEnv->error_file);
                fprintf(corexEnv->error_file, ")\n");
            }
#        endif
        if (check_subsum && is_subsumed(tuple)) {
            delete tuple;
            continue;
        }
        insert_tuple(tuple);  // This routine is typically called before
                              // the delta relation updates have been performed
                              // so the insert goes to the delta_new relation.
        return 1;
    }
}

// *************************************************************************
int corHashSimpleRelation::pop_all_from_hidden_relation() 
{
    for(;;) {
        // Until new tuple is found
        if (!heap)
            return 0;

        corTuple *tuple = corpop_heap(&heap,priority_info->order);
        if (tuple->is_deleted())
            continue;

#        ifdef DO_TRACE
            if (DISPLAY_INSERTIONS)    {
                fprintf(corexEnv->error_file, "\tUnhiding fact: %s(",
                        name->string());
                tuple->printon(corexEnv->error_file);
                fprintf(corexEnv->error_file, ")\n");
            }
#        endif
        if (check_subsum && is_subsumed(tuple)) {
            delete tuple;
            continue;
        }
        insert_tuple(tuple);    // This routine is typically called before
                            // the delta relation updates have been performed
                            // so the insert goes to the delta_new relation.
        // Don't return here -- move all the tuples
    }
}

// *************************************************************************
void corHashSimpleRelation::print_name(FILE *outf)
{
    fprintf(outf, "%15s/%d  : (base)    ", corSymbolString(name), arity());
    index_set.print_adornments(outf);
    fprintf(outf, ": count %d\n", count);
}

// *************************************************************************
