/*****************************************************************************/
//    RelationMisc.C
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "RelationMisc.h"

#include "Globals.h"
#include "CoralError.h"

#include "DontCareArg.h"
#include "Database.h"
#include "DatabaseStruct.h"
#include "QFModuleData.h"
#include "Literal.h"
#include "Relation.h"
#include "StorageRelation.h"
#include "AggSelInfo.h"
#include "Execute.h"
#include "HashSimpleRelation.h"

/*****************************************************************************/

#define MAX_STORED_TUPLEITS 1024

/*****************************************************************************/

int corRelCount = 0;

int corC_tupleiterConstructs = 0;
int corC_tupleiterDestructs = 0;

extern corDatabaseStruct corBuiltinDB;

/*****************************************************************************/
void corgroup_dump(corBindEnv * new_env, corArgList * new_arglist, corRelation *rel,
    char *old_arglist)
{
    if (corT_Stack.count <= 0 || corT_Stack.top() == NULL)
        return;
    
    corTable * F = corT_Stack.top();
    
    corRuleInfo * rInfo = F->moduledata->curr_rInfo;
   
    FILE *fp = corexEnv->explain_fp; 
    if (!fp) {
        return;
    }
    
    for (int j=0; j <= rInfo->num_literals; j++) {
        if (j==rInfo->num_literals) {
            //    First Argument
            fprintf(fp, "def_%s(",
                corSymbolString(F->moduledata->export_pred->name));
            rInfo->rule->head->predicate->name->print(new_env, fp);
            fputs("(",fp);
            ((corArgList *)new_arglist)->
                print_dump(new_env, fp);
            
            fputs(")",fp);
        }
        else {
            fprintf(fp, "use_%s(",
                corSymbolString(F->moduledata->export_pred->name));
            rInfo->rule->preds[j]->predicate->name->print(new_env,fp);
            fputs("(",fp);
            rInfo->arg_list[j].print_dump(new_env, fp);
            fputs(")",fp);
        }
            
        fputs(", [",fp);
        
        rInfo->rule->head->predicate->name->print(new_env, fp);
        fputs("(",fp);
        ((corArgList *)new_arglist)->print_dump(new_env, fp);
        fputs(")",fp);
        
        for (int i=0; i < rInfo->num_literals; i++) {
            fputs(",",fp);
            rInfo->rule->preds[i]->predicate->name->print(new_env,fp);
            fputs("(",fp);
            rInfo->arg_list[i].print_dump(new_env,fp);
            fputs(")",fp);
        }
        if (old_arglist) {
            fprintf(fp, ", %s_SUBSUMED__", corSymbolString(rel->name));
            fputs("(",fp);
            fputs(old_arglist,fp);
            fputs(")",fp);
        }
            
        fputs("]).\n",fp);
    }


    //    For two new relations: def_new and use_new
    for (int j1=0; j1 <= rInfo->num_literals; j1++) {
        if (j1==rInfo->num_literals) {
            // First Argument
            fprintf(fp, "def_new_%s(",
                corSymbolString(F->moduledata->export_pred->name));
            rInfo->rule->head->predicate->name->print(new_env, fp);
            fputs("(",fp);
            ((corArgList *)new_arglist)->print_dump(new_env, fp);
            fputs(")",fp);
        }
        else {
            fprintf(fp, "use_new_%s(",
                corSymbolString(F->moduledata->export_pred->name));
            rInfo->rule->preds[j1]->predicate->name->print(new_env, fp);
            fputs("(",fp);
            rInfo->arg_list[j1].print_dump(new_env,fp);
            fputs(")",fp);
        }

        fputs(", [",fp);

        rInfo->rule->head->predicate->name->print(new_env, fp);
        fputs("(",fp);
        ((corArgList *)new_arglist)->print_dump(new_env, fp);
        fputs(")",fp);

        for (int i1=0; i1 < rInfo->num_literals; i1++) {
            fputs(",",fp);
            rInfo->rule->preds[i1]->predicate->name->print(new_env, fp);
            fputs("(",fp);
            rInfo->arg_list[i1].print_dump(new_env,fp);
            fputs(")",fp);
        }
        if (old_arglist) {
            fprintf(fp, ", %s_SUBSUMED__", corSymbolString(rel->name));
            fputs("(",fp);
            fputs(old_arglist,fp);
            fputs(")",fp);
        }

        fputs("]).\n",fp);
    }
    fflush(fp);
}

/*****************************************************************************/
RelationKind corfind_r_kind(Name name)
{
    if (coris_magic(name)) return COR_R_MAGIC;
    else if (coris_supp(name)) return COR_R_SUPPLEMENTARY;
    else return COR_R_ANSWER;
}

/*****************************************************************************/
/*
    *::release(TupleIterator &iterator)
  
    Called when the tupleiterator is no longer needed, and the state
    stored in the tuple iterator can be released.  The get_next / get_next_tuple
    functions store data in the tuple iterator, so that the data is saved
    across calls to get_next / get_next_tuple.  When the external routine is
    done with getting new answers, it calls tuple_iterator.release(), 
    which in turn calls reln::release(tuple_iterator).  

    WARNING:: release may get called more than once with the 
    same tuple_iterator.  Care should be taken to avoid deleting space
    more than once.  This is typically done by resetting pointed fields 
    to NULL after deletion.
*/
/*****************************************************************************/

/*****************************************************************************/
int corhas_grouping_args(corArgList &args) 
{
    int i ;

    for (i=0;i<args.count();i++) {
        if (args[i].kindof()==COR_GROUPING)
            return 1;
    }
    return 0;
}

/*****************************************************************************/
/*
    checks the relation to see if the anno has already been specified
    temporary kludge !! - PRAVEEN
*/

int corduplicate_agg_sel_anno(corStorageRelation *rel, corAggSelInfo *newinfo)
{
    corAggSelInfo *aggsel;
         
    for (aggsel=rel->agg_sel_info; aggsel; aggsel=aggsel->next) {
        if ((newinfo->agg_kind == aggsel->agg_kind) &&
            (newinfo->agg_value == aggsel->agg_value) &&
            (newinfo->var_count == aggsel->var_count))
        {
            return 1;
        }
    }
    return 0;
}

/*****************************************************************************/
corRelation* corfind_local_or_external_relation(Name name, int arity) 
{
    corRelation *rel;
    rel = corfind_local_relation(name, arity);
    if (!rel) rel = cormake_relation(name, arity);
    return rel;
}

/*****************************************************************************/
corRelation* corfind_external_relation(Name name, int arity)
{
    Name new_name = name;

    Association *ptr = corSymbolLookup(corCurDB->RelationTable, new_name);
    if (HashNone(ptr->arg)) {
        Association *ptr1;
        ptr1 = corSymbolLookup(corBuiltinDB.RelationTable, name);
        if (HashNone(ptr1->arg)) return NULL;
        else return (corRelation *)ptr1->val;
    }
    else {
        return (corRelation *)(ptr->val);
    }
}

/*****************************************************************************/
corRelation* corFind_External_Relation(Name name, int arity)
{
    // This version creates a relation if none is found.
    // It gives a warning if it did not find an existing relation
    /* NOTE: remove FindRelation once all calls to it are replaced */

    corRelation *rel = corfind_external_relation(name, arity);
    if (rel) return rel;
    
    if (!corexEnv->C_quiet_mode_default) {
        CORAL_error(COR_REL_NOT_FOUND,name->string(),"Find_External_Relation");
        fprintf(corexEnv->trace_file, "Arity of relation = %d \n", arity);
        fprintf(corexEnv->trace_file, "Allocating new relation.\n");
    }
    
    return cormake_relation(name, arity);
}

/*****************************************************************************/
corRelation* find_relation(char * name, int arity) 
{
    Name n = corEnterSymbol(name);
    return find_relation(n, arity);
}

/*****************************************************************************/
corRelation* find_relation(Name name, int arity)
{
    corRelation *rel;

    rel = corfind_local_relation(name, arity);
    if (!rel) rel = corfind_external_relation(name, arity);
    if (!rel) {
        CORAL_error(COR_REL_NOT_FOUND,name->string(),"find_relation");
        fprintf(corexEnv->trace_file, "Arity of relation = %d\n", arity);
    }
    return rel;
}

//**************************************************************************

corRelation* find_relation(corCArg c, int arity)
{
    corRelation *rel;
    Name name = c.SymbolP();

    rel = corfind_local_relation(name, arity);
    if (!rel) rel = corfind_external_relation(name, arity);
    if (!rel) {
        CORAL_error(COR_REL_NOT_FOUND,name->string(),"find_relation");
        fprintf(corexEnv->trace_file, "Arity of relation = %d\n", arity);
    }
    return rel;
}

//**************************************************************************
corRelation* cormake_relation(Name name, int arity, int delta_indexed)
{
    char buf[400];

    if (coris_magic(name) || coris_supp(name) || coris_mp_done(name)) {
        fprintf(corexEnv->error_file,
          "Warning::relation name %s conflicts with system reserved names !\n",
          corSymbolString(name));
        sprintf(buf, "new%s%d", corSymbolString(name), arity);
        fprintf(corexEnv->error_file, "New name is %s.\n", buf);
    }
    else
        sprintf(buf, "%s", corSymbolString(name));

    Name new_name = corEnterSymbol(buf);

    Association *ptr = corSymbolLookup(corCurDB->RelationTable, new_name);
    if (HashNone(ptr->arg)) {
        Association *ptr1;
        ptr1 = corSymbolLookup(corBuiltinDB.RelationTable, name);
        if (HashNone(ptr1->arg)) {
            ptr->arg = new_name;
            ptr->val = corAllocateRelation(new_name, arity, delta_indexed);
            corTabInserted(corCurDB->RelationTable, ptr);
        }
        else return (corRelation *)ptr1->val;
    }
    return (corRelation*)ptr->val;
}

/*****************************************************************************/
corRelation* cormake_relation(char *name, int arity, int delta_indexed)
{
    Name n = corEnterSymbol(name);
    return cormake_relation(n, arity, delta_indexed);
}

/*****************************************************************************/
corStorageRelation* cormake_local_relation(Name name, int arity, int delta_indexed,
    int set_subsum)
{
    //    this stuff ensures that the relation constructor sees the desired
    //    value of exEnv->C_check_subsum_default    -- P

    if ((set_subsum == -1) || (corexEnv->C_check_subsum_default == set_subsum))
        return corAllocateRelation(name, arity, delta_indexed);


    int temp = corexEnv->C_check_subsum_default;
    corexEnv->C_check_subsum_default = set_subsum;
    corStorageRelation *rel = corAllocateRelation(name, arity, delta_indexed);
    corexEnv->C_check_subsum_default = temp;
    return rel;
}

/*****************************************************************************/
void correclaim_relation(corStorageRelation *rel)
{
/*
#    ifdef WITH_PERSISTENCE
    if (rel->r_kind == COR_R_EXODUS)
        delete ((PersistentRel *)rel);
    else
#    endif
        delete ((HashSimpleRelation *)rel);
*/
 // NEW_CHNG:wX
 delete rel;
}

/*****************************************************************************/
void correclaim_index(corGenericIndex *index, RelationKind r_kind)
{
    if (!index) return;
/*
#    ifdef WITH_PERSISTENCE
    if (r_kind == COR_R_EXODUS)
        delete ((PersistentIndex *)index);
    else
#    endif
    delete ((HashIndex *)index);
*/
 // NEW_CHNG:wX
 delete index;

}

/*****************************************************************************/


