/*****************************************************************************/
//    Collection.C
/*****************************************************************************/

#include "Collection.h"

#include "Literal.h"

/*****************************************************************************/
//    Initialize a collection of rules and provide it with a hash table

void corinit_collection(struct corcollection *corcollection, int hashable)
{
    corcollection->chain = NULL;
    corcollection->lastptr = &corcollection->chain;
    corcollection->count = 0;
    corcollection->hashable = hashable;
    if (hashable)
        corInitTab(corcollection->tab, 10, NULL);
}


/*****************************************************************************/
#ifdef DEBUG

static void check_collection(struct corcollection *corcollection)
{
    Link *link, **lptr;
    Name name = NULL;
    int count = 0, diff_count = 0;
    lptr = &corcollection->chain;
    for ( ; *lptr != NULL; lptr = &(*lptr)->next) {
        if (lptr != (*lptr)->prevptr) abort();
        count++;
        if ((*lptr)->name != name) diff_count++;
        name = (*lptr)->name;
        if (corcollection->hashable) {
            Association *hptr = corSymbolLookup(corcollection->tab, (*lptr)->name);
            if (HashNone(hptr->arg)) abort();
            for (link= (Link*)hptr->val; link != *lptr; link = link->next)
                if (link == NULL || link->name != (*lptr)->name) abort();
        }
    }
    if (lptr != corcollection->lastptr) abort();

    if (count != corcollection->count) abort();
    if (corcollection->hashable && corcollection->tab->cur_size != diff_count)
        abort();
}

#endif

/*****************************************************************************/
void coradd_member(Link *member, struct corcollection *corcollection)
{
    if (corcollection->hashable) {
        Association *ptr = corSymbolLookup(corcollection->tab, member->name);
        if (HashNone(ptr->arg)) {
            ptr->arg = (Pointer)member->name;
            ptr->val = (Pointer)member;
            corTabInserted(corcollection->tab, ptr);
            //     link into linked list
            member->next = NULL;
            member->prevptr = corcollection->lastptr;
            *corcollection->lastptr = member;
            corcollection->lastptr = &member->next;
        } else {
            Link *old_member = (Link*)ptr->val;
            Link *next_member = old_member->next;

            //    keep members ordered, for no real reason
            while ( next_member && next_member->name == member->name)
                old_member = next_member, next_member = old_member->next;
            member->next = old_member->next;
            member->prevptr = &old_member->next;
            if (old_member->next) old_member->next->prevptr = &member->next;
            else corcollection->lastptr = &member->next;
            old_member->next = member;
        }
    }
    else {
        //    link into linked list
        member->next = NULL;
        member->prevptr = corcollection->lastptr;
        *corcollection->lastptr = member;
        corcollection->lastptr = &member->next;
    }
    corcollection->count++;
#    ifdef DEBUG
        if (corcollection->hashable)
            check_collection(corcollection);
#    endif
}

/*****************************************************************************/
void corremove_member(Link *member, struct corcollection *corcollection)
{
    if (corcollection->hashable) abort();
    *member->prevptr = member->next;
    if (member->next) member->next->prevptr = member->prevptr;
    else corcollection->lastptr = member->prevptr;
    corcollection->count--;
}

/*****************************************************************************/
//    Select some member from 'collection', remove it from the 'collection',
//    and return the removed element.

Link* corpop_member(struct corcollection *corcollection)
{
    Link *member = corcollection->chain;
    if (member == NULL) return NULL;
    if (corcollection->hashable) {
        Association *ptr = corSymbolLookup(corcollection->tab, member->name);
        if (HashNone(ptr->arg) || (Link*)ptr->val != member) abort();
        if (member->next && member->next->name == member->name)
            ptr->val = (Pointer)member->next;
        else {
            corcollection->tab->cur_size--;
            ptr->arg = HashDeleted;
        }
    }
    *member->prevptr = member->next;
    if (member->next) member->next->prevptr = member->prevptr;
    else corcollection->lastptr = member->prevptr;
    corcollection->count--;

#    ifdef DEBUG
        check_collection(corcollection);
#    endif
    return member;
}

/*****************************************************************************/
void coradd_rule(corClause *rule, struct corcollection *corcollection)
{
    rule->link.name = corcollection->hashable ? rule->head->pred : NULL;
    coradd_member((Link*)rule, corcollection);
}

/*****************************************************************************/

