/*****************************************************************************/
//    Profile.C
/*****************************************************************************/

#include <string.h>
#include <stdio.h>

#include "Profile.h"

#include "Config.h"

/*****************************************************************************/

#define    COR_U_SEC 1000000

corProfileInfo    corGlobalProfileInfo;

/*****************************************************************************/
void corProfileInfo::clear()
{
    i_unsuccesful = 0;
    i_succesful = 0;
    i_newfacts = 0;
    getnexts = 0;
    deleted_tuples = 0;
    non_unifying_tuples = 0;
    unifies = 0;
    corsubsumes = 0;
    invocations = 0;
    usage_time = 0.0;
}

/*****************************************************************************/
void corProfileInfo::init_rule()
{
    invocations++;

    long t1 , t2;
    struct rusage tmp_usage;
     
#    ifndef UNIX_HP
    getrusage(RUSAGE_SELF, &tmp_usage);
#    endif

    t1 = tmp_usage.ru_utime.tv_sec;
    t2 = tmp_usage.ru_utime.tv_usec;
    usage_time_st = ((double) t1) + ((double) t2) / COR_U_SEC;
}

/*****************************************************************************/
void corProfileInfo::exit_rule()
{
    long t1 , t2;
    struct rusage tmp_usage;
     
#    ifndef UNIX_HP
    getrusage(RUSAGE_SELF, &tmp_usage);
#    endif

    t1 = tmp_usage.ru_utime.tv_sec;
    t2 = tmp_usage.ru_utime.tv_usec;
    usage_time = (((double) t1) + ((double) t2) / COR_U_SEC) - usage_time_st;
}

/*****************************************************************************/
void corProfileInfo::print(FILE *fd, char * str, int *predOrders,
    int numpreds) const
{
    fprintf(fd, "_____%s__PROFILE INFO_____\n", str);
    fprintf(fd, "%s\n", rule);
    if (numpreds && predOrders) {
        fprintf(fd, "Evaluation order : ");
        for (int i = 0; i < numpreds; i++)
            fprintf(fd, "%d ", predOrders[i]);
        fprintf(fd, "\n");
    }
     
    fprintf(fd, "Invocations : %4d,    Usage time :  %2.2f\n", invocations,
        usage_time);
    fprintf(fd, "Instantiations : (Successful) %4d,   (Unsuccessful) %4d\t|  ",
         i_succesful, i_unsuccesful);
    fprintf(fd, "New Facts %4d \n", i_newfacts);
    fprintf(fd, "Unifies %d\t| Subsumes %d\t| Non Unifying tuples %4d\n",
        unifies, corsubsumes, non_unifying_tuples);
    fprintf(fd, "GetNexts %4d\t| Deleted Tuples %4d\n\n",
        getnexts, deleted_tuples);
}

/*****************************************************************************/
void corAddProfileInfo(corProfileInfo *A, corProfileInfo *B) 
{
    B->deleted_tuples += A->deleted_tuples;
    B->non_unifying_tuples += A->non_unifying_tuples;
    B->i_succesful += A->i_succesful;
    B->i_unsuccesful += A->i_unsuccesful;
    B->i_newfacts += A->i_newfacts;
    B->getnexts    += A->getnexts;
    B->unifies += A->unifies;
    B->corsubsumes += A->corsubsumes;
    B->invocations += A->invocations;
    B->usage_time += A->usage_time;
}

/*****************************************************************************/

