// *************************************************************************
//    RuleEnv.C
// *************************************************************************

#include "RuleEnv.h"

#include "TupleIterator.h"

// *************************************************************************
corRuleEnv::corRuleEnv(corSemiNaive *parent) : bindenv(parent->rInfo->num_vars,4) 
{    
    max_literals = parent->rInfo->num_literals;

    // This is a hack to prevent memory overwrites for rules of the form p(X).
    // rInfo->num_literals == 0 for these rules, there are places in the code
    // where it is assumed that at least one rhs literal exists.  This should
    // actually be fixed by either changing that code, or by introducing a
    // true literal on the rhs.

    if (!max_literals) max_literals = 1;

    rmarks = new corStackMark [max_literals];
    tuple_its = new corTupleIterator * [max_literals];
    arglist_its = new corArgList * [max_literals];
    tuples = new corTuple * [max_literals];
    bindenvs = new corBindEnv *[max_literals];
    parent_sn = parent;

    // allocate tuple iterators if possible
    if (parent->rule_id >= 0) {
        for (int i = 0; i < max_literals; i++) {
            tuple_its[i] = new corTupleIterator(NULL, parent->rInfo->arg_list[i],
                NULL, NULL, NULL, NULL);
        }
    }
}

// *************************************************************************
corRuleEnv::~corRuleEnv()
{
    delete [] rmarks;
    // delete each of the tuple iterators !
    for (int i = 0; i < max_literals; i++)
        delete tuple_its[i];
    delete [] tuple_its;
    delete [] arglist_its;
    delete [] tuples;
    delete [] bindenvs;
}

// *************************************************************************
