// *************************************************************************
//    QFMaterializedModuleData.C
// *************************************************************************

#include "Globals.h"
#include "QFMaterializedModuleData.h"

#include "CoralBreak.h"

// *************************************************************************
corQFMaterializedModuleData::corQFMaterializedModuleData(corTable *table,
    corRuleInfoArray *rArray, corParserStruct& corparserStruct)
{
    F = table;
    ruleArray = rArray;
    numSccs = 0;
    sccArray = NULL;
    multiset_annotations = corparserStruct.MultisetAnnotations;
    agg_sel_annotations    = corparserStruct.agg_sel_annotations;
    prioritize_annotations = corparserStruct.prioritize_annotations;
    index_annotations = corparserStruct.make_index_annotations;
    diskrel_annotations = corparserStruct.diskrel_annotations;

    free_table_list = NULL;
    saved_state = 0;
    no_magic = 0;
    rel_options = 0;
    predicates = new corModulePred(NULL);

}

// *************************************************************************
corQFMaterializedModuleData::~corQFMaterializedModuleData()
{
    int i;
    for (i=0; i< numSccs; i++)
        delete sccArray[i];

    delete [] sccArray;
    delete predicates;
    delete ruleArray;
}

// *************************************************************************
void corQFMaterializedModuleData::AddScc(corScc *scc, corPredicate *ext_magicpred)
{
    int i;
    corSccInfo **temp = sccArray;

    sccArray = new corSccInfo*[++numSccs];
    for (i=0; i< numSccs-1; i++)
        sccArray[i] = temp[i];

    sccArray[i] = new corSccInfo(this, scc, ext_magicpred, numSccs-1);
    if (numSccs > 1) delete [] temp;
}

/*****************************************************************************/
void corQFMaterializedModuleData::clean_cache()
{
    for (int i = 0; i < numSccs; sccArray[i++]->clean_cache());
}

/*****************************************************************************/
/*
    Iterator is guaranteed to be either NULL, or have a state field allocated.
    If iterator->state->rm_Arr is NULL, it implies that this is the first
    time this module is being invoked using this iterator
*/

void corQFMaterializedModuleData::Solve(corTable *table, corTupleIterator *iterator)
{
    int outcome;

    outcome = strcmp(corb.module_name,
        (sccArray[0]->module->module_info.name)->string());

    if (!(iterator) || !(iterator->state->rm_Arr)) {
        clean_cache();
        // This is conservative.  In most cases it is not needed
        //fprintf(exEnv->error_file,"The number of Scc is %d\n", numSccs);
        for (int i = 0; i < numSccs-1; i++) {
            if (!outcome) {
                //    Break on module
                coral_break(corb.module_name,-1);
            }
            if (corb.scc_flags) {
                // Break on SCC
                fprintf(corexEnv->error_file,"scc_id = %d\n", sccArray[i]->scc_id);
                if ((!outcome) && (corb.scc_flags->test(sccArray[i]->scc_id+1))) {
                    coral_break(corScc_String, sccArray[i]->scc_id+1);
                }
                else if ((!outcome) && (corb.all)) {
                    coral_break(corScc_String, sccArray[i]->scc_id+1);
                }
            }
            sccArray[i]->Solve(table);
        }
    }

    if (corb.scc_flags) {
        //    Break on last scc
        fprintf(corexEnv->error_file,"scc_id = %d\n", sccArray[numSccs-1]->scc_id);
        if ((!outcome) && (corb.scc_flags->test(sccArray[numSccs-1]->scc_id+1))) {
            coral_break(corScc_String, sccArray[numSccs-1]->scc_id+1);
        }
        else if ((!outcome) && (corb.all)) {
            coral_break(corScc_String, sccArray[numSccs-1]->scc_id+1);
        }
    }
    sccArray[numSccs-1]->Solve(table, iterator);
    if (iterator) iterator->state->iter_number++;
}

/*****************************************************************************/
void corQFMaterializedModuleData::print(FILE *fd, int print_preds, 
                     int print_text, int print_snrules)
{ 
  for (int i=0; i<numSccs; i++)
    sccArray[i]->print(fd, print_preds, print_text, print_snrules);
}
