// *************************************************************************
//    SccInfo.h
// *************************************************************************

#ifndef _SccInfo_H_
#define _SccInfo_H_

//#include "Table.h"
#include "Predicate.h"
#include "QFMaterializedModuleData.h"
#include "Scc.h"
#include "SemiNaive.h"

// *************************************************************************
//    Codependent classes

class corSemiNaive;
class corQFMaterializedModuleData;
class corTable;

// *************************************************************************

struct corSccInfo {
    int  allocateRuleStructures(corPredicate *);
        // allocates memory for seminaive rules, and returns the number
        // of rules in the scc. This method also sets the value of numSNRules
public:
    corQFMaterializedModuleData* module;  // Module to which it belongs.
    corScc* scc;
    corSemiNaive* sn_rules;               // Array of seminaive rules
    int numSNRules;
    corPredicate** pred_ptrs;             // Array of preds corresponding to
                                       // the scc
    int numPreds;                      // number of preds in this SCC
    int num_lowerPreds;                // number of preds in all previous SCCs
    int*offset_arr;                    // Offsets of preds in RMarkArray
                                       // Size of this is numPreds
    int*sn_offset_arr;                 // Positions of heads of sn rules
                                       // in pred_ptrs
    int*pred_success_arr;              // Number of newtuples generated for
                                       // each predicate in the previous
                                       // iteration size is : numPreds
    int rec_rule_start;                // Starting position of recursive rules
    int has_hidden_relations;          // True if any relation has a hidden
                                       // part due to prioritize
    int scc_id;                        // number identifying the position of
                                       // this scc in the sccArray of the
                                       // module

    corSccInfo(corQFMaterializedModuleData *md, corScc *scc, corPredicate *, int id);
   ~corSccInfo();
    void Solve(corTable *F, corTupleIterator *iterator = NULL);
    void clean_cache() {
        if (sn_rules) sn_rules->clean_cache();
    }
    void clean_hidden_relations(corTable *F);
    int iterateRecursiveRules(corTable *F, corRMarkArray *rm_Arr, int, int);
    void print(FILE *fd, int print_preds = 1, 
         int print_text = 1, int print_snrules = 1); // added by Ken

private:
    void createSNStructs(corPredicate *);
    corRMarkArray *setupSolve(corTable *F);
    void display_annotation_info(corTable *F);
};

typedef struct corSccInfo corSccInfo;

// *************************************************************************

#endif _SccInfo_H_

