//**************************************************************************
//    SymbolTable.C
//**************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SymbolTable.h"

#include "Arg.h"
#include "SymbolArg.h"
#include "Util.h"

static Association clear_symbol(Pointer , Association* ass);

//**************************************************************************

corHashFuncs corStdHashFuncs = {NULL, corSymbolLookup};

//**************************************************************************

extern void corSymbolTabInit();
typedef Association (*Func)(Pointer, Association*);
extern void corForEachTabElement(HashTable*, Func, Pointer) ;
int corenter_symbol = 1;

struct SymbolPackage {
   SymbolPackage() { corSymbolTabInit(); }
};

SymbolPackage KeywordPackage();

int corSymbolTabInitDone = 0;

#define EQwith(el) (SymbolHash(el)==hash && SymbolLength(el) == len \
  && memcmp(corSymbolString(el), s, len) == 0)

HashTable corSymbolTab[1] =
  {{ 4, 2, 8 /* half log of initial size*/, 0, &corStdHashFuncs, 0, 0 }};

static Name *
FindSymbol(const char *s, int len, HashVal hash) /* patterned after SymbolLookup */
{
#define tab corSymbolTab
#undef    HASH_ELEMENT_TYPE
#define HASH_ELEMENT_TYPE Name
#define HASH_LENGTH_LOG tab->tab_len_log
#define HASH_ELEMENT_LOG tab->element_len_log
#undef    HASH_TO_INDEX
#define HASH_TO_INDEX(hash) (hash >> (HASH_WORD_LEN - tab->tab_len_log)) << tab->element_len_log
#undef    HASH_EQUAL
#define HASH_EQUAL(element,arg) EQwith(*element)
#undef    HASH_NULL
#define HASH_NULL(element) (*element==(Name)HashNull)
#define HASH_DATA tab->data
#undef    HASH_TO_STEP
#define HASH_TO_STEP(hash) (((hash<<1)+1) & ((1<<tab->tab_len_log)-1)) << tab->element_len_log

#include "HashFunc.h"
#undef tab
}

//**************************************************************************
corSymbolTable::~corSymbolTable()
{
    if (f != &corStdHashFuncs)
        delete f;
    delete data;
}

//**************************************************************************
void corPrintSymbol(Name sym, FILE *ff)
{
    fputs(sym->string(), ff);
}

//**************************************************************************
void corSymbolTabInit()
{
    if (corSymbolTabInitDone) return;
    corSymbolTabInitDone = 1;
    corRehashBigger(corSymbolTab);
}

//**************************************************************************
Name corEnterSymbol(const char *str)
{
    int len;
    HashVal h; register Name *p, s;

    len = strlen(str);
    h = corIntToHash(corStrHash(str, len));

    if (!corenter_symbol) {
        s = new corSymbol(str, len, h);
        return s;
    }

    if (!corSymbolTabInitDone)
        corSymbolTabInit();
 
    p = (Name*)FindSymbol(str, len, h);
    if (*p != NULL) {
        return *p;
    }

    s = new corSymbol(str, len, h);
    *p = s;
    corTabInserted(corSymbolTab, (Association*)s);

    return s;
}

//**************************************************************************
static Association clear_symbol(Pointer , Association* ass)
{
    delete (corSymbol *)(ass->val) ;
    return *ass ;
}

//**************************************************************************
void corclearSymbols()
{
    // This causes some problems, since the symbols are needed by
    // some of the other destructors of global variables.

    //if (SymbolTab) {
    //    ForEachTabElement(SymbolTab, (Func)clear_symbol, 0) ;
    //}

    // Expanded here to try to avoid segmentation fault
    // (Note the limit on i has changed)

    HashTable *tab = corSymbolTab;
    Pointer extra = 0;

    if (corSymbolTab) {

        register i;
        register Association *element;

        i = 1<<tab->tab_len_log; element = tab->data;
        for (; i > 1; i--) {
            if (!HashNone(element->arg))
                clear_symbol(extra, element);
            element = (Association*)((char*)element + tab->element_len);
        }
    }
}

//**************************************************************************
Association * corSymbolLookup(HashTable *tab, ConstPointer arg)
{
    unsigned short hash = (unsigned short) SymbolHash(arg);
#undef  HASH_ELEMENT_TYPE
#define HASH_ELEMENT_TYPE Association
#define HASH_LENGTH_LOG tab->tab_len_log
#define HASH_ELEMENT_LOG tab->element_len_log
#undef  HASH_TO_INDEX
#define HASH_TO_INDEX(hash) \
    (hash >> (HASH_WORD_LEN - HASH_LENGTH_LOG)) << HASH_ELEMENT_LOG
#undef  HASH_EQUAL
#define HASH_EQUAL(element,arg) (element->arg==(Pointer)arg)
//#define HASH_EQUAL(element,arg) (((Symbol *)(element->arg))->equals((Symbol *)arg))
#undef  HASH_NULL
#define HASH_NULL(element) (element->arg==HashNull)
#define HASH_DATA tab->data
#define HASH_DELETED(element)  (element->arg == HashDeleted)
/* the mask is redundant if masking is done before indexing */
#undef  HASH_TO_STEP
#define HASH_TO_STEP(hash) \
    (((hash<<1)+1) & ((1<<HASH_LENGTH_LOG)-1)) << HASH_ELEMENT_LOG
#include "HashFunc.h"
}

//**************************************************************************
Association * corDeleteInTab(HashTable *tab, ConstPointer arg)
{
    Association *p = corSymbolLookup(tab, arg);
    if (HashNone(p->arg)) return NULL;
    tab->cur_size--;
    tab->cur_deleted++;
    p->arg = HashDeleted;
    return p;
}

//**************************************************************************
void corForEachTabElement(register HashTable *tab, Func f, Pointer extra)
{
    register i;
    register Association *element;

    i = 1<<tab->tab_len_log; element = tab->data;
    for (; i > 0; i--) {
        if (!HashNone(element->arg)) (*f)(extra, element);
        element = (Association*)((char*)element + tab->element_len);
    }
}

//**************************************************************************
void corTabInserted(register HashTable *tab, Association *)
{
    int max = 1<<tab->tab_len_log;
    max -= max>>3; /* max is ceiling(max_space*7/8) */
    tab->cur_size++;
    if (tab->cur_size >= max)
    corRehashBigger(tab);
}

//**************************************************************************
void corTabInsert(register HashTable *tab, Association *element)
{
    register Pointer *dst, *src = (Pointer*)element;
    register len = tab->element_len >> 2;

    dst = (Pointer*)(*tab->f->lookup)(tab, element->arg);
    if (HashNone(((Association*)dst)->arg)) {
        while (--len >= 0)
            *dst++ = *src++;
        corTabInserted(tab, element);
    }
    else {
        while (--len >= 0) *dst++ = *src++;
    }
}

//**************************************************************************
void corInsert(register HashTable *tab, ConstPointer arg, Pointer val)
{
    Association element;

    element.arg = arg;
    element.val = val;
    corTabInsert(tab, &element);
}

//**************************************************************************
void corRehashBigger(register HashTable *tab)
{
    Association *old_data = tab->data;
    register int i; register char *element;
    register size;

    i = 1<<tab->tab_len_log;
    element = (char*)old_data;
    tab->tab_len_log++;
    size = 1<<(tab->element_len_log + tab->tab_len_log);
    tab->data = (Association*) new char[size];
    tab->cur_size = 0;
    memset(tab->data,0,size);
    if (old_data) {
        size = tab->element_len;
        for (; i > 0; i--, element = element + size)
            if (!HashNone(((Association*)element)->arg))
                corTabInsert(tab, (Association*)element);
        delete old_data;
    }
}

//**************************************************************************
HashTable* corInitTab(register HashTable *tab, int n,
    Pointer defaultVal)
{
    register int i;
    int size;

    tab->element_len = 8;
    tab->element_len_log = 3;
    for (i = 3; (1 << i) < n; i++) ;
    tab->tab_len_log = i;
    if (defaultVal == NULL)
        tab->f = &corStdHashFuncs;
    else {
        tab->f = new corHashFuncs;
        *tab->f = corStdHashFuncs;
        tab->f->defaultVal = defaultVal;
    }
    tab->cur_size = 0;
    tab->cur_deleted = 0;
    tab->flags = 0;
    size = 1 << (tab->element_len_log + tab->tab_len_log);
    if (tab->data)
        delete tab->data;
    tab->data = (Association *) new char[size];
    memset(tab->data, 0, size);
    return tab;
}

//**************************************************************************
HashTable* corAllocTab(int n, Pointer defaultVal)
{
    HashTable *tab;

    tab = new HashTable;
    corInitTab(tab, n, defaultVal);
    return tab;
}

//**************************************************************************

#define CALL_f(name, value) (*f)(extra, name, value)

//**************************************************************************
void corForEachInTab(register corSymbolTable *tab, TabIterator f, Pointer extra)
{
    DoForEachInTab(tab, CALL_f);
}

//**************************************************************************
int corTableSize(HashTable *tab)
{
#    if 1
        return tab->cur_size;
#    else
        register i; register MapPair *pair; register n = 0;
        for (i = tab->tabLen, pair = tab->tab; i > 0; i--, pair++)
            if (pair->arg.O != NULLARG) n++;
        return n;
#    endif
}

//**************************************************************************
