//**************************************************************************
//    SymbolTable.h
//**************************************************************************

#ifndef _SymbolTable_H_
#define _SymbolTable_H_

#include "Config.h"
#include "SymbolArg.h"

//**************************************************************************

struct Association {
    ConstPointer arg;
    Pointer val;     // Maybe be missing in ObArray tables for Symbols.
};

typedef struct corSymbolTable HashTable;

typedef Association *(*LookupFunc)(HashTable *, ConstPointer);

//**************************************************************************

typedef struct corHashFuncs {
    void *defaultVal;
    LookupFunc lookup;
} corHashFuncs;

//**************************************************************************

struct corSymbolTable {
    unsigned char element_len;
    unsigned char element_len_log;
    unsigned char tab_len_log;
    unsigned char flags;
    corHashFuncs *f;
    int cur_size;
    int cur_deleted;
    Association *data;
/***
    SymbolTable(unsigned char arg1,
        unsigned char arg2,
        unsigned char arg3,
        unsigned char arg4,
        HashFuncs *arg5,
        int arg6,
        int arg7
           ) {
      element_len = arg1; element_len_log = arg2; tab_len_log = arg3;
      flags = arg4; f = arg5; cur_size = arg6; cur_deleted = arg7; }
***/
     corSymbolTable(){ data = NULL; f = NULL; }
    ~corSymbolTable();
};

//**************************************************************************

#define HASH_WORD_LEN 16

#define NULLARG (ConstPointer)HashNull

#define HashSymbolsOnly        1
#define HashDefaultIsProc    2    // get default as Apply(arg, default)

//**************************************************************************

#define DoForEachInTab(tab, action)            \
  { register struct corSymbolTable *__tab__ = tab;        \
    register Association *__element__ = __tab__->data;    \
    register __i__ = 1<<__tab__->tab_len_log;        \
    for (; __i__ > 0; __i__--,                \
       __element__ = (Association*)((char*)__element__+__tab__->element_len))\
        if (!HashNone(__element__->arg))            \
      { action(__element__->arg, __element__->val); } }

#define HashNull NULL
#define HashDeleted ((ConstPointer)1)
#define HashNone(x) ((unsigned)(x) <= 1) // i.e. either HashNull/HashDeleted
#define FailedSearch(arg) (arg == 0 || arg == NoValue || False(arg))
#define SymbInTab(s) (_RefCount(s) == RefsInfinity)

//**************************************************************************

typedef Association (*Func)(Pointer, Association*);

typedef void (*TabIterator)(Pointer extra, ConstPointer arg, Pointer val);

//**************************************************************************

void corPrintSymbol(Name sym, FILE *ff);
void corSymbolTabInit();
Name corEnterSymbol(const char *str);
void corclearSymbols();
Association * corSymbolLookup(HashTable *tab, ConstPointer arg);
Association * corDeleteInTab(struct corSymbolTable *tab, ConstPointer arg);
void corForEachTabElement(struct corSymbolTable *tab, Func f, Pointer extra);
void corTabInserted(register struct corSymbolTable *tab, Association *);
void corTabInsert(register struct corSymbolTable *tab, Association *element);
void corInsert(register HashTable *tab, ConstPointer arg, Pointer val);
void corRehashBigger(register struct corSymbolTable *tab);
struct corSymbolTable* corInitTab(register struct corSymbolTable *tab, int n,
    Pointer defaultVal);
struct corSymbolTable* corAllocTab(int n, Pointer defaultVal);

void corForEachInTab(register corSymbolTable *tab, TabIterator f, Pointer extra);
int corTableSize(struct corSymbolTable *tab);

//**************************************************************************

#endif _SymbolTable_H_

