// *************************************************************************
//    TupleIterator.h
// *************************************************************************

#ifndef _TupleIterator_H_
#define _TupleIterator_H_

#include "ArgList.h"
#include "BindEnv.h"
//#include "Relation.h"
#include "RMark.h"
//#include "QFModuleData.h"
#include "StackMark.h"
#include "PipelinedExecInfo.h"
//#include "RelationList.h"
#include "GenericIndex.h"
#include "ModuleInfo.h"
#include "ModuleExecInfo.h"

// *************************************************************************
//    Codependent classes

class corPipelinedExecInfo;
class corRelation;
class corQFModuleData;
class corRelationLink;

// *************************************************************************

class corTupleIterator {
public:
    //    Fields that are (usually) invariant within a rule application
    corArgList& arg_list;
    corArgList& temp_arglist;            // Is this causing problems ??
    corBindEnv* bindenv;
    corRelation* relation;

    corRMark* start_mark;
    union {
        corRMark *end_mark;
        corQFModuleData *moduledata;    // For derived relations
    };

    corStackMark stack_mark;
    corBindEnv *tuple_env;
    union {
        corRMark *mpos;
        void *ppos;                // position of next tuple to be considered
        void *scanDesc;            // for persistent relations
        void *cursor;              // for persistent relations
        corTupleIterator *next;    // For free store handling
        corTupleIterator *answer_iterator;
        int SocketRel_next; // next tuple in cache for socket relation
    };
    union {
        int ipos;
        corPipelinedExecInfo *pipe;  // for pipelined relations only. 
    };
    int pindex;                     // index of current bucket, if the
                                    // relation is a hash relation

    int _no_match;                  // To return status on calls to
                                    // get_next_tuple

    corRelationLink *delta_rel_pos; // Used to handle semi-naive delta relns

    // The following three fields are for dealing with indices
    HashVal hash;
    corGenericIndex *index;
    int *indexnump;

    // The following field is used by return unify.  It is set by
    // unify_literal, to signal that apply-rule should do the actual 
    // unification, since extra information about the supplementary and 
    // answer facts is required.
    int unification_number;  

    union {
        corPipelinedExecInfo* calling_pei;    // used by pipelining for Cut
        corTuple *prev_tuple;
    };

    corBindEnv *prev_bindenv;

    // the following field is used to store intermediate state of
    // bottom-up computation, when answers are being returned at
    // the end of each iteration
    union {
        corModuleExecInfo *state;
        corModuleInfo *prev_module_info;   // for pipelined relations only
    };
    
    // userData is meant to be used by users to store data for
    // builtins between calls  ---- wgr 5/12/93
 
    void* userData;

    //void * operator new(size_t size);
    //void operator delete(void *, size_t size);

    void reset_fields(corRelation *rel, corBindEnv *env,
            corRMark *start, corRMark *end, int *tryindexp)
    {
        relation = rel;
        bindenv = env;
        start_mark = start;
        end_mark = end;
        indexnump = tryindexp;
    }

    corTupleIterator(corRelation *rel, corArgList& args, corBindEnv* env,
        corRMark *start = NULL, corRMark *end = NULL, int *tryindexp = NULL);

/* corTupleIterator(Relation *rel, Arg** args, BindEnv* env, RMark *start=NULL,
     RMark *end = NULL, int *tryindexp = NULL);
*/
    corTupleIterator(corRelation *rel, corTuple *tuple, corRMark *start = NULL,
        corRMark *end = NULL);
   ~corTupleIterator();
    inline void reset() {
        // Reset only those fields that change within a rule application
        tuple_env = NULL;
        ppos = NULL;
        ipos = 0;
        _no_match = 0;
        delta_rel_pos = NULL;
        hash = VarHashValue;
        index = NULL;
        unification_number = -1;
        prev_tuple = NULL;
        prev_bindenv = NULL;
        state = NULL;
    }

    inline int in_delta()        { return (delta_rel_pos != NULL); }
    inline void release() {
        if (relation && relation->has_release) {
            relation->release(*this);
            relation->scan_count--;
        }
    }

    inline int no_match()        {return _no_match;}
    inline void set_no_match()    {_no_match = 1;}
    inline void reset_no_match(){_no_match = 0;}
    corBindEnv *get_next();

    //    get_next_tuple(): returns tuple if there is one (Builtin rels etc
    //    return NULL). 
    inline corTuple *get_next_tuple() {
        return relation->get_next_tuple(*this);
    }
};

// *************************************************************************

#endif _TupleIterator_H_

