/*****************************************************************************/
//	FunctArgs.C
/*****************************************************************************/

#include "FunctArgs.h"

#include "Rule.h"
#include "TermList.h"

/*****************************************************************************/
//	Returns true if the argument 'a' is already in the termlist 'terms'.

TermList* find_arg(TermList *terms, corArg *a)
{
	for (terms = terms->next; terms; terms = terms->next) {
		if (terms->func_arg->equals(a)) return terms;
	}
	return NULL;
}

/*****************************************************************************/
//	Adds argument 'a' to the end of termlist 'terms'

void add_arg(TermList *terms, corArg *a, corArgList* args, int pos, TermList *t)
{
	TermList *tmp =  new TermList(args, pos);
	tmp->set_funcarg((corFuncArg *)a);
	tmp->set_similar(t);

	//	If t is NULL, add arg to the end of the list, so that it will be
	//	processed eventually.  Else, it is similar to another such arg seen
	//	already, and can be placed at the head of the list.

	if (t)
		tmp->next = terms->next;
	else
		while (terms->next) terms = terms->next;

	terms->next = tmp;
}

/*****************************************************************************/
/*
	Gets only terms that are arguments of pred and are functors.  Doesn't
	return subterms separately.
*/

void get_terms(corLiteral * pred, TermList *terms)
{
	int j = 0;

	ASSERT(pred != NULL);

	FOR_EACH_ARG(a, pred->args) {
		if ( a->kindof() == COR_FUNCTOR ) {
			//	Put this in termlist
			// Should these lines be commented out ?? -- PRAVEEN
			TermList *tmp = find_arg( terms, a->FuncArgP() );
			add_arg(terms, a->FuncArgP(), &(pred->args), j, tmp);
		}
		j++;
	} END_EACH_ARG
}

/*****************************************************************************/

